;;; sidebar.el --- browse files and tags in a separate window

;; Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010 Martin Rudalics

;; Time-stamp: "2011-09-03 10:15:15 martin"
;; Author: Martin Rudalics <rudalics@gmx.at>
;; Keywords: tools, files, tags, bookmarks, buffers, matches

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software Foundation,
;; Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;; Bugs:

;; grepping for get this-command in lisp directory doesn't list hits in
;; emulation sub-dir!

;;; TODO

;; (1) For c-mode list two lines or so for defun-beginning in
;;     match-buffers, also check why *two* calls are needed to find
;;     defun-starts (in other modes as well!!).  Finally, make the
;;     c-begin-/end-of stuff work.

;; (2) Find out why sidebar-down is slow with invisible text at point.

;; (3) When I'm in buffer help.el and the buffer-list has the buffer
;;     calc-help.el, switching to the sidebar jumps to calc-help.el.  This
;;     must be anchored at bol somehow.

;; (4) We currently use sidebar-client-buffer/-window in the
;;     client-buffer too, seems unpretty.

;;; Commentary: 

;; sidebar provides a separate window to list and traverse directories, files,
;; tags, bookmarks, buffers, matches, ...

;;; Invariants:

;;  - every sidebar has an associated window
;;  - a window may have at most one associated sidebar
;;  - sidebars don't have an associated sidebar

;;; Code:

;; _____________________________________________________________________________
;;                                                                              
;;;				customization					
;; _____________________________________________________________________________
;;                                                                              

;;				 sidebar group					
(defgroup sidebar nil
  "Sidebar."
  :version "22.1"
  :group 'environment)

(defcustom sidebar-wrap nil
  "Non-nil means wrap movement at beginning and end of sidebars.

Applies to commands like `sidebar-up', `sidebar-down', `sidebar-down-jump' and
`sidebar-up-jump'."
  :type 'boolean
  :group 'sidebar)

(defcustom sidebar-directory-brackets nil
  "Non-nil means enclose directory names in square brackets.

On Emacsen before 22.1 turning this option on usually breaks `mouse-face'
overlay.  On Emacs 22.1 turning this option on will cause highlighting the
closing bracket when the mouse pointer is over the line.  This is an Emacs
redisplay bug which should eventually get fixed.

Emacs 23.1 breaks this again, hence don't use it.  Let's see how the
before-/after-string story develops."
  :type 'boolean
  :group 'sidebar)

(defcustom sidebar-even-heights t
  "When non-nil even heights of sidebar and associated window."
  :type 'boolean
  :set (lambda (symbol value)
         (set-default symbol value)
         (when (fboundp 'sidebar-reconfigure)
           (sidebar-reconfigure)))
  :group 'sidebar)

(defcustom sidebar-minimum-width 12
  "Minimum number of columns of sidebar.
Selected sidebars always have at least that many columns."
  :type 'integer
  :group 'sidebar)

(defcustom sidebar-maximum-width 36
  "Maximum number of columns of selected sidebars.
This value applies if at least one object displayed in a sidebar contains
that many characters.  Otherwise the actual width of a selected sidebar is
chosen such to accomodate the largest object it lists but may be larger
by `sidebar-width-tolerance' columns."
  :type 'integer
  :group 'sidebar)

(defcustom sidebar-width-tolerance 6
  "Changing width of sidebar may leave that many columns blank.
Setting this to a larger value will reduce the number of sidebar width
changes.  Compare `sidebar-minimum-width' and `sidebar-maximum-width'."
  :type 'integer
  :group 'sidebar)

(defcustom sidebar-indent 1
  "Standard indentation for tags and bookmarks."
  :type 'integer
  :group 'sidebar)

(defcustom sidebar-jump-delay 1.0
  "Time current `sidebar-jump-string' is retained.
The value of this variable applies when executing `sidebar-down-jump' and
`sidebar-up-jump' commands.  Default is one second.

You may jump to a sidebar line by typing the first character(s) of the
string on that line.  The characters you type are accumulated in the
variable `sidebar-jump-string'.  If you don't enter input for
`sidebar-jump-delay' seconds, `sidebar-jump-string' is cleared and the very
next character you enter reinitializes `sidebar-jump-string'."
  :type 'number
  :group 'sidebar)

(defcustom sidebar-attributes-delay 0.25
  "Delay for displaying attributes and other informative messages.
The value of this variable applies when executing `sidebar-down',
`sidebar-up', or related commands.  Default is 0.25 seconds.

Setting this variable to a higher value will increase the time until
attributes are displayed in the echo area.  Setting this variable to a
value less than that of your \"auto repeat delay\" \(the time the
operating system waits until it reports a second event whenever you hold
down a key\) may cause the display of an extra message after the first
movement.  You may have to experiment with the value of this variable
until you get satisfactory settings for your system."
  :type 'number
  :group 'sidebar)

;; here we might want nil for no-synch and t for immediate synch
(defcustom sidebar-synchronize-delay 0.25
  "Delay for synchronizing tags/bookmarks and associated window.
The value of this variable applies when executing `sidebar-down',
`sidebar-up' and related commands in tag or bookmark lists.  Default is
0.25 seconds.  Windows are synchronized when this time period has
elapsed and no user input has arrived.

Setting this variable to a higher value will \(annoyingly\) increase the
time until windows are synchronized during vertical movement.  Setting
this variable to a value less than that of your \"auto repeat delay\"
\(the time the operating system waits until it reports a second event
whenever you hold down a key\) may cause one \(annoying\) extra
synchronization after the first movement.  You may have to experiment
with the value of this variable until you get satisfactory settings for
your system."
  :type 'number
  :group 'sidebar)

(defcustom sidebar-redisplay-discard-input nil
  "Discard input after executing a sidebar command.
On most systems setting this variable should not have any noticeable
effect.  Set it non-nil when redisplay doesn't occur after a `sidebar-up'
or `sidebar-down' command."
  :type 'boolean
  :group 'sidebar)

(defgroup sidebar-directory nil
  "Sidebar files group."
  :version "22.1"
  :group 'sidebar)

(defcustom sidebar-files-sort 'name
  "How to sort file names in file lists.
Choices are:

 `name': alphabetically (default)

 `time': by modification time (latest first), then by name

 `size': by size (largest first), then by name

 `extension': by extension first, then by name

 `unsorted': unsorted (nil)

See also `sidebar-reverse'."
  :type '(choice (const :tag "by name" name)
                 (const :tag "by extension" extension)
                 (const :tag "by time" time)
                 (const :tag "by size" size)
                 (const :tag "unsorted" unsorted))
  :group 'sidebar-directory)
(make-variable-buffer-local 'sidebar-files-sort)

(defcustom sidebar-reverse nil
  "When non-nil reverse order of listings.
See also `sidebar-files-sort'."
  :type 'boolean
  :group 'sidebar-directory)
(make-variable-buffer-local 'sidebar-reverse)

(defcustom sidebar-files-use-ls nil
  "Non-nil means `sidebar-file-list' should use \"ls\" (faster)."
  :type 'boolean
  :group 'sidebar-directory)

(defcustom sidebar-grep-recurse t
  "Non-nil means grep should recurse into directories."
  :type 'boolean
  :group 'sidebar-directory)

(defgroup sidebar-tags nil
  "Sidebar tags group."
  :version "22.1"
  :group 'sidebar)

;; *** TODO: The following customizations are likely ignored ***
(defcustom sidebar-tags-sort 'unless-nosort
  "Non-nil means sort tags alphabetically.  Default is `unless-nosort'.
The choices are

 `always': always sort tags

 `unless-nosort': sort tags unless associated file's extension is in
   `sidebar-tag-extension-nosort' or `imenu-sort-function' has the
   buffer-local value nil

 `when-sort': sort tags iff associated file's extension is in
   `sidebar-tag-extension-sort'

 `never': never sort tags"
  :type '(choice (const always)
                 (const unless-nosort)
                 (const when-sort)
                 (const never))
  :group 'sidebar-tags)

(defvar sidebar-tag-sort-regexp nil)

(defcustom sidebar-tag-extension-sort '()
  "List of extensions that require alphabetic sorting of tags."
  :type '(repeat (regexp :tag "Extension"))
  :set (lambda (sym val)
	 (set-default sym val)
         (when val
           (setq sidebar-tag-sort-regexp
                 (concat "." "\\(" (mapconcat 'regexp-quote val "\\|") "\\)$"))))
  :group 'sidebar-tags)

(defvar sidebar-tag-nosort-regexp nil)

(defcustom sidebar-tag-extension-nosort
  '(".tex" ".sty" ".texi" ".texinfo" ".html" ".htm" ".letter" ".article")
  "List of extensions that forbid alphabetic sorting of tags."
  :type '(repeat (regexp :tag "Extension"))
  :set (lambda (sym val)
	 (set-default sym val)
         (when val
           (setq sidebar-tag-nosort-regexp
                 (concat "." "\\(" (mapconcat 'regexp-quote val "\\|") "\\)$"))))
  :group 'sidebar-tags)

(defcustom sidebar-tag-group 'when-sort
  "Whether tags shall be grouped.  Default is `when-sort'.
Grouping tags means stripping a common prefix from neighboring tags.  In
general, grouping makes sense iff tags have been sorted before.

The choices are

 `always': always group tags

 `when-sort': group tags iff they are sorted too

 `never': never group tags"
  :type '(choice (const always)
                 (const when-sort)
                 (const never))
  :group 'sidebar-tags)

(defcustom sidebar-tag-group-minimum-size 4
  "Minimum size of a tag-group.  Default is 4.
Minimum value of this variable is 2."
  :type 'integer
  :group 'sidebar-tags)

(defgroup sidebar-bookmarks nil
  "Sidebar bookmarks group."
  :version "22.1"
  :group 'sidebar)

(defcustom sidebar-bookmarks-sort 'name
  "How bookmark names are sorted in bookmark lists.
Choices are:

 `name': alphabetically (default)

 `position': by position (within file)

 `time': by creation time (latest first)

See also `sidebar-bookmarks-reverse'."
  :type '(choice (const :tag "by name" name)
                 (const :tag "by position" position)
                 (const :tag "by time" time))
  :group 'sidebar-bookmarks)
(make-variable-buffer-local 'sidebar-bookmarks-sort)

(defcustom sidebar-bookmarks-reverse nil
  "When non-nil reverse order of bookmark names in bookmark lists.
Default is nil.  See also `sidebar-bookmarks-sort'."
  :type 'boolean
  :group 'sidebar-bookmarks)
(make-variable-buffer-local 'sidebar-bookmarks-reverse)

(defcustom sidebar-bookmarks-update 'when-write
  "When non-nil add a file visited or written to bookmarks.
Choices are:

 `when-write': add file only when it is written (default)

 `when-find': add file when it is written or found

 `never': don't add file written or found (nil)"
  :type '(choice (const when-write)
                 (const when-find)
                 (const :tag "never" nil))
  :group 'sidebar-bookmarks)

(defcustom sidebar-bookmarks-norecord nil
  "List of file or directories whose bookmarks are not recorded on saving."
  :type '(repeat string)
  :group 'sidebar-bookmarks)

(defgroup sidebar-buffers nil
  "Sidebar buffer group."
  :version "22.1"
  :group 'sidebar)

(defcustom sidebar-buffers-sort t
  "Non-nil means sort buffers alphabetically.
If this variable is non-nil buffers appear in alphabetical order.
Otherwise buffers are listed in the order returned by `buffer-list'."
  :type 'boolean
  :group 'sidebar-buffers)
(make-variable-buffer-local 'sidebar-buffers-sort)

(defcustom sidebar-buffer-ignore-space t
  "Non-nil means do not list buffers whose name starts with a space.
Default is t.  Automatically becomes buffer-local when set in any fashion."
  :type 'boolean
  :group 'sidebar-buffers)
(make-variable-buffer-local 'sidebar-buffer-ignore-space)

(defcustom sidebar-buffer-ignore-asterisk nil
  "Non-nil means do not list buffers whose name starts with an asterisk.
Default is nil.  Automatically becomes buffer-local when set in any fashion."
  :type 'boolean
  :group 'sidebar-buffers)
(make-variable-buffer-local 'sidebar-buffer-ignore-asterisk)

(defgroup sidebar-match nil
  "Sidebar match group."
  :version "22.1"
  :group 'sidebar)

(defcustom sidebar-match-lines-before 2
  "Number of lines displayed before each match."
  :type 'integer
  :group 'sidebar-match)
(make-variable-buffer-local 'sidebar-match-lines-before)

(defcustom sidebar-match-lines-after 2
  "Number of lines displayed after each match."
  :type 'integer
  :group 'sidebar-match)
(make-variable-buffer-local 'sidebar-match-lines-after)

(defcustom sidebar-match-symbol-inhibit "-_"
  "Characters before of after symbol that inhibit symbol matches.
A character appearing in this string inhibits symbol matches preceded or
followed by this character.  Compare `sidebar-match-object'.  Default is
\"-_\".  Automatically becomes buffer-local when set in any fashion."
  :type 'string
  :group 'sidebar-match)
(make-variable-buffer-local 'sidebar-match-symbol-inhibit)

(defcustom sidebar-match-context t
  "When non-nil visualize match context in sidebar and buffer.
When this variable is non-nil

 1. visible lines before and after each match within a defun are confined to the
    containing defun

 2. visible lines before and after a match not contained in a defun are trimmed
    by preceding of following defuns

 3. if the current match is within a defun, that defun is highlighted with
    `sidebar-match-context', all matches contained within that defun have
    their corresponding sidebar lines highlighted with that face too

Default is t (non-nil).  Automatically becomes buffer-local when set in any
fashion.

Observe: Setting this variable non-nil makes sense iff a suitable function for
finding the begin and end of a defun exists.  You may have to define your own
`beginning-of-defun-function' and `end-of-defun-function' if the standard
functions are insufficient for this purpose.  Moreover, visualizing match
contexts may considerably slow down redisplay.  It might be a good idea to make
the value of this variable depend on the major mode of the associated buffer."
  :type 'boolean
  :group 'sidebar-match)
(make-variable-buffer-local 'sidebar-match-context)

;; _____________________________________________________________________________
;; 										
;;;				  variables					
;; _____________________________________________________________________________
;; 										
(defvar sidebar-list nil
  "List of \(buffers of\) all sidebars.")

(defvar sidebar-mode nil
  "Mode of this sidebar.
Must be one of `directory', `tag', `bookmark', `match', or `grep'.")
(make-variable-buffer-local 'sidebar-mode)

(defvar sidebar-client-buffer nil
  "Buffer associated with this sidebar.")
(make-variable-buffer-local 'sidebar-client-buffer)

(defvar sidebar-client-buffer-chars-modified-tick nil
  "Previous value of modified tick of `sidebar-client-buffer'.
Used to detect whether match overlays must be updated.")
(make-variable-buffer-local 'sidebar-client-buffer-chars-modified-tick)

(defvar sidebar-file nil
  "File associated with this sidebar.
For each sidebar records the file visited by `sidebar-client-buffer'.  nil when
`sidebar-client-buffer' doesn't visit a file.

Should be identic to `(buffer-file-name sidebar-client-buffer)'.")
(make-variable-buffer-local 'sidebar-file)

(defvar sidebar-old nil
  "A list saving some values when a sidebar is updated.
Currently saves `sidebar-client-buffer', `sidebar-file', and `default-directory'.")

(defvar sidebar-client-window nil
  "Window associated with this sidebar.
For each sidebar the window to the right with the same top coordinates.
When that window is deleted the sidebar eventually gets deleted too.")
(make-variable-buffer-local 'sidebar-client-window)

(defvar sidebar-buffer-window-list nil
  "List of buffer/window pairs associated with sidebars in match mode.
Each entry to this list is a cons pair whose `car' is the `sidebar-client-buffer'
of some sidebar in match mode and whose `cdr' is the `sidebar-client-window' of
that sidebar.")

(defvar sidebar-unfolded nil
  "Non-nil means entries of current sidebar have been unfolded.")
(make-variable-buffer-local 'sidebar-unfolded)

(defvar sidebar-jump-string ""
  "Accumulated jump string.")
(make-variable-buffer-local 'sidebar-jump-string)

(defvar sidebar-jump-time 0
  "Time of last sidebar jump event.")
(make-variable-buffer-local 'sidebar-jump-time)

(defvar sidebar-inhibit-reconfigure nil
  "When non-nil inhibit `sidebar-reconfigure' after current command finishes.
Temporarily set by functions that change the window configuration in a
controlled way and do not want `sidebar-reconfigure' to readjust things.")

(defvar sidebar-width sidebar-minimum-width
  "Width of \(window of\) sidebar.")
(make-variable-buffer-local 'sidebar-width)

(defvar sidebar-title ""
  "Title of sidebar.")
(make-variable-buffer-local 'sidebar-title)

(defvar sidebar-line-overlay nil
  "Overlay for current sidebar line.")
(make-variable-buffer-local 'sidebar-line-overlay)

(defvar sidebar-line-echo-overlay nil
  "Overlay used to echo current-sidebar line within `sidebar-client-window'.")
(make-variable-buffer-local 'sidebar-line-echo-overlay)

(defvar sidebar-overlay nil
  "Sidebar overlay.
Overlay used to highlight entire text of sidebar.  Installed by
`sidebar-overlays'.")
(make-variable-buffer-local 'sidebar-overlay)

;; *** TODO: Doc of following appears incorrect ***
(defvar sidebar-before-length 0
  "Length of before overlay.")
(make-variable-buffer-local 'sidebar-before-length)

(defvar sidebar-before-overlay nil
  "Overlay preceding text of sidebar.
Installed by `sidebar-overlays'.")
(make-variable-buffer-local 'sidebar-before-overlay)

(defvar sidebar-after-overlay nil
  "Overlay following text of sidebar.
Installed by `sidebar-overlays'.")
(make-variable-buffer-local 'sidebar-after-overlay)

(defvar sidebar-after-string nil
  "Sidebar after string.
A sequence of newline characters calculated by `sidebar-mode' from display
parameters and inserted by `sidebar-overlays' after the text of the
sidebar.")
(make-variable-buffer-local 'sidebar-after-string)

(defvar sidebar-attributes-timer nil
  "Idle timer for showing attributes and other informative messages.")

(defvar sidebar-synchronize-timer nil
  "Idle timer for synchronization.")

(defvar sidebar-scroll-delay 0.1
  "Delay for `sidebar-window-scroll-timer'.
Sidebars may appear \"unclean\" for this many seconds.")

(defvar sidebar-window-scroll-timer nil
  "Idle timer for `sidebar-scroll'.")

(defvar sidebar-mode-hook nil
  "List of functions to call when sidebar mode is invoked.
This hook is automatically executed after sidebar mode has been loaded.")

;			   directory mode variables				
(defvar sidebar-file-list nil
  "File list produced by `sidebar-file-list-dfa' or `sidebar-file-list-ls'.")

(defvar sidebar-ls-buffer nil
  "Buffer used for ls output.")

(defvar sidebar-file-list-ls nil
  "Non-nil iff `sidebar-file-list' was created by `sidebar-file-list-ls'.")

(defvar sidebar-file-list-history nil
  "History variable for `sidebar-file-list-query' and
`sidebar-file-list-query-non-recursive'.")

;;			      tag mode variables				
;; *** TODO: Clarify doc-string in the following three ***
(defvar sidebar-window-point nil
  "`window-point' of `sidebar-client-window'.")
(make-variable-buffer-local 'sidebar-window-point)

(defvar sidebar-tag-point nil
  "Sidebar position corresponding to `window-point' of `sidebar-client-window'.")

(defvar sidebar-tag-best-point nil
  "Position in `sidebar-client-buffer' corresponding to `sidebar-tag-point'.")

;			     bookmark mode variables				
(defvar sidebar-bookmark-default-file-modtime nil
  "Saved value of modification time of `bookmark-default-file'.")

(defvar sidebar-bookmark-alist-modification-count nil
  "Saved value of `bookmark-alist-modification-count'.")

;;			     match mode variables				
(defvar sidebar-match-history ()
  "sidebar match history.")

(defvar sidebar-match-string nil
  "Last matching exp.")
(make-variable-buffer-local 'sidebar-match-string)

(defvar sidebar-match-count nil
  "Number of matches.")
(make-variable-buffer-local 'sidebar-match-count)

(defvar sidebar-replace-string nil
  "Replace expression.")
(make-variable-buffer-local 'sidebar-replace-string)

(defvar sidebar-replace-length nil
  "Length of replace expression.")
(make-variable-buffer-local 'sidebar-replace-length)

(defvar sidebar-match-yank-at nil
  "Yank position.")

(defvar sidebar-match-yank-buffer nil
  "Yank buffer.")

(defvar sidebar-context-echo-overlay nil
  "Sidebar match context overlay.")
(make-variable-buffer-local 'sidebar-context-echo-overlay)

(defvar sidebar-context-overlay nil
  "Sidebar context overlay.")
(make-variable-buffer-local 'sidebar-context-overlay)

(defvar sidebar-match-minibuffer-local-map
  (let ((map (make-sparse-keymap)))
    (set-keymap-parent map minibuffer-local-map)
    (define-key map "\M-r" 'sidebar-match-regexp-exit-minibuffer)
    (define-key map "\M-s" 'sidebar-match-symbol-exit-minibuffer)
    (define-key map "\M-w" 'sidebar-match-word-exit-minibuffer)
    (define-key map "\C-n" 'sidebar-match-yank-next-character)
    (define-key map "\C-w" 'sidebar-match-yank-word)
    map)
  "Keymap for editing match strings in the minibuffer.")

(defvar sidebar-match-type nil
  "Type of last match.  May be `regexp', `symbol', `word', or nil.")
(make-variable-buffer-local 'sidebar-match-type)

(defvar sidebar-match-exit nil
  "Global.")

;;			     grep mode variables				
(defvar sidebar-grep-buffer nil
  "Buffer receiving grep output.")

(defvar sidebar-grep-buffer-point 1
  "Position in buffer receiving grep output.")

(defvar sidebar-grep-buffer-of-sidebar nil
  "Sidebar where grep is eventually placed.")

(defvar sidebar-grep-process nil
  "Grep process.")

(defvar sidebar-grep-transact nil
  "Grep transacting.")

(defvar sidebar-grep-string nil
  "Grep string.")

(defvar sidebar-grep-type nil
  "Grep type.")

(defvar sidebar-grep-last nil
  "Last grep buffer.")

(defvar sidebar-grep-last-point nil
  "Last grep point.")

(defvar sidebar-grep-last-width nil
  "Last grep buffer width.")

(defvar sidebar-rename-map nil
  "Local keymap for sidebar-renaming.")

(unless sidebar-rename-map
  (setq sidebar-rename-map (make-sparse-keymap))
  (set-keymap-parent sidebar-rename-map global-map)
  ;; ..... RETURN, F2
  (define-key sidebar-rename-map [return] 'sidebar-rename-return)
  (define-key sidebar-rename-map [(control g)] 'sidebar-rename-restore)
  (define-key sidebar-rename-map [f2] 'sidebar-rename))

(defvar sidebar-rename nil
  "Non-nil when rename in progress.")
(make-variable-buffer-local 'sidebar-rename)

(defvar sidebar-rename-buffer nil
  "Buffer of sidebar with rename operation in progress.
Must be nil when there is no rename in progress.")

(defvar sidebar-rename-type nil
  "Type of renaming operation in progress.")

(defvar sidebar-rename-from (make-marker)
  "Start position of rename string.")

(defvar sidebar-rename-mask-from nil
  "Start position of rename mask.")

(defvar sidebar-rename-mask-to nil
  "End position of rename mask.")

(defvar sidebar-rename-to (make-marker)
  "End position of rename string.")

(defvar sidebar-rename-string nil
  "String before rename operation.")

(defvar sidebar-rename-string-no-properties nil
  "String before rename operation without properties.")

(defvar sidebar-rename-before-overlay nil
  "Overlay masking head of rename string.")

(defvar sidebar-rename-after-overlay nil
  "Overlay masking tail of rename string.")

(defvar sidebar-rename-overlay nil
  "Rename overlay.")

(defvar sidebar-rename-highlight-overlay nil)

;; _____________________________________________________________________________
;;                                                                              
;;;                            timer and delay group                            
;; _____________________________________________________________________________
;;                                                                              

;; _____________________________________________________________________________
;;                                                                              
;;;                                  utilities                                  
;; _____________________________________________________________________________
;;                                                                              
(defsubst sidebar-addq (elt list)
  "Add ELT to LIST unless ELT is a `memq' of LIST and return LIST."
  (if (memq elt list) list (cons elt list)))

(defsubst sidebar-add (elt list)
  "Add ELT to LIST unless ELT is a `member' of LIST and return LIST."
  (if (member elt list) list (cons elt list)))

;; I'm afraid but the following two are likely needed:

;;; (defsubst set-buffer-when-live (buffer)
;;;   "Do `set-buffer' iff BUFFER is live.  Return nil otherwise."
;;;   (when (buffer-live-p buffer) (set-buffer buffer)))

;;; (defsubst set-window-buffer-only (window buffer &optional keep-margins)
;;;   "As `set-window-buffer' but don't change current buffer."
;;;   (let ((current-buffer (current-buffer)))
;;;     (set-window-buffer window buffer keep-margins)
;;;     (set-buffer current-buffer)))

(defsubst set-window-and-current-buffer (window buffer &optional keep-margins)
  "As `set-window-buffer' but when window is selected make BUFFER current."
  (set-window-buffer window buffer keep-margins)
  (when (eq window (selected-window))
    (set-buffer buffer)))

;; Observe: When we have two match windows on the same buffer and we set
;; buffer-invisibility-spec on one window we get invisibility on the other
;; window as well (consider that as a feature).
(defun sidebar-add-to-invisibility-spec (arg)
  "Add ARG to `buffer-invisibility-spec'.
See documentation of `buffer-invisibility-spec' for the kind of elements
that can be added.  This avoids that the same element is added twice as
with the original `add-to-invisibility-spec'."
  (cond
   ((eq buffer-invisibility-spec t)
    (setq buffer-invisibility-spec (list t arg)))
   ((member arg buffer-invisibility-spec))
   (t (setq buffer-invisibility-spec
            (cons arg buffer-invisibility-spec)))))

;; *** TODO: Replace that when doing bookmarks ***
(unless (fboundp 'cdaadr)
  (defsubst cdaadr (x)
    "Return the `cdr' of the `car' of the `car' of the `cdr' of X."
    (cdr (car (car (cdr x))))))

;; _____________________________________________________________________________
;;                                                                              
;;;                           string predicates group                           
;; _____________________________________________________________________________
;;                                                                              
(defsubst sidebar-string-nocase-equal (string1 string2)
  "Return t if STRING1 equals STRING2 ignoring case."
  (eq (compare-strings string1 0 nil string2 0 nil t) t))

(defsubst sidebar-string-equal (string1 string2)
  "Return t if STRING1 equals STRING2 respecting `case-fold-search'."
  (if case-fold-search
      (sidebar-string-nocase-equal string1 string2)
    (string-equal string1 string2)))

(defsubst sidebar-string-nocase-lessp (string1 string2)
  "Return t if STRING1 is less than STRING2 ignoring case."
  (let ((v (compare-strings string1 0 nil string2 0 nil t)))
    (and (numberp v) (< v 0))))

(defsubst sidebar-string-case-car-lessp (item1 item2)
  "Return t if string in car of ITEM1 is less than string in car of ITEM2."
  (string-lessp (car item1) (car item2)))

(defsubst sidebar-string-nocase-car-lessp (item1 item2)
  "Return t if string in car of ITEM1 is less than string in car of ITEM2
ignoring case."
  (let ((v (compare-strings (car item1) 0 nil (car item2) 0 nil t)))
    (and (numberp v) (< v 0))))

(defsubst sidebar-string-case-caar-lessp (item1 item2)
  "Return t if string in caar of ITEM1 is less than string in caar of ITEM2."
  (string-lessp (caar item1) (caar item2)))

(defsubst sidebar-string-nocase-caar-lessp (item1 item2)
  "Return t if string in caar of ITEM1 is less than string in caar of ITEM2
ignoring case."
  (let ((v (compare-strings (caar item1) 0 nil (caar item2) 0 nil t)))
    (and (numberp v) (< v 0))))

(defsubst sidebar-string-lessp (string1 string2)
  "Return t if STRING1 is less than STRING2 respecting `case-fold-search'."
  (if case-fold-search
      (sidebar-string-nocase-lessp string1 string2)
    (string-lessp string1 string2)))

(defsubst sidebar-string-car-lessp (item1 item2)
  "Return t if string in car of ITEM11 is less than string in car of ITEM2.
Respect current value of `case-fold-search'."
  (if case-fold-search
      (sidebar-string-nocase-car-lessp item1 item2)
    (sidebar-string-case-car-lessp item1 item2)))

(defsubst sidebar-string-caar-lessp (item1 item2)
  "Return t if string in caar of ITEM11 is less than string in caar of ITEM2.
Respect current value of `case-fold-search'."
  (if case-fold-search
      (sidebar-string-nocase-caar-lessp item1 item2)
    (sidebar-string-case-caar-lessp item1 item2)))

(defvar sidebar-string-lessp-function
  (if case-fold-search 'sidebar-string-nocase-lessp 'string-lessp))

(defvar sidebar-string-car-lessp-function
  (if case-fold-search
      'sidebar-string-nocase-car-lessp
    'sidebar-string-case-car-lessp))

(defvar sidebar-string-caar-lessp-function
  (if case-fold-search
      'sidebar-string-nocase-caar-lessp
    'sidebar-string-case-caar-lessp))

;; _____________________________________________________________________________
;;                                                                              
;;;                                filter group                                 
;; _____________________________________________________________________________
;;                                                                              
(defsubst sidebar-filter-state (filter) (nth 0 filter))

(defsubst sidebar-filter-domain (filter) (nth 1 filter))

(defsubst sidebar-filter-name (filter) (nth 2 filter))

(defsubst sidebar-filter-type (filter) (nth 3 filter))

(defsubst sidebar-filter-invert (filter) (nth 4 filter))

(defsubst sidebar-filter-phase (filter) (nth 5 filter))

(defsubst sidebar-filter-interpret (filter) (car (nth 6 filter)))

(defsubst sidebar-filter-string (filter) (cdr (nth 6 filter)))

(defgroup sidebar-filter nil
  "Sidebar filters group."
  :version "22.1"
  :group 'sidebar)

(defcustom sidebar-filter-apply t
  "If non-nil apply filters.
Default is t.  Automatically becomes buffer-local when set in any fashion.  See
also `sidebar-filter-list'."
  :type 'boolean
  :group 'sidebar-filter)
(make-variable-buffer-local 'sidebar-filter-apply)

(defcustom sidebar-filter-wildcard-expand-braces t
  "When non-nil expand braces when defining and using filters in filter-lists."
  :type 'boolean
  :set (lambda (sym val)
         (setq sidebar-filter-wildcard-expand-braces val)
         (when (fboundp 'sidebar-filter-list-regexp)
           (sidebar-filter-list-regexp)))
  :group 'sidebar-filter)

(defcustom sidebar-filter-list-sort nil
  "If non-nil sort `sidebar-filter-list' in menus and customization buffers.
This variable _only_ affects the way filters are sorted in menus and
customization buffers.  Filters are _always_ sorted by `sidebar-filter-lessp'
before they are applied to lists of files or directories."
  :type '(choice (const :tag "by name" name)
                 (const :tag "by phase" phase)
                 (const :tag "by phase and name" phase-name)
                 (const :tag "unsorted" nil))
  :group 'sidebar-filter)

(defun sidebar-filter-lessp (filter1 filter2)
  "Non-nil if FILTER1 is less than FILTER2.
Filters are ordered

 1. by `state' where `on' < `off',

 2. by `phase',

 3. by `type', where `pass' < `block' < `stop'

 4. by `invert', where `off' < `on'

 5. by `string', where `regexp' < `wildcard' < `extension'"
  (or (and (sidebar-filter-state filter1) (not (sidebar-filter-state filter2)))
      (and (eq (sidebar-filter-state filter1) (sidebar-filter-state filter2))
           (or (< (sidebar-filter-phase filter1) (sidebar-filter-phase filter2))
               (and (= (sidebar-filter-phase filter1) (sidebar-filter-phase filter2))
                    ;; pass < block < stop
                    (or (and (eq (sidebar-filter-type filter1) 'pass)
                             (not (eq (sidebar-filter-type filter2) 'pass)))
                        (and (eq (sidebar-filter-type filter1) 'block)
                             (eq (sidebar-filter-type filter2) 'stop))
                        (and (eq (sidebar-filter-type filter1)
                                 (sidebar-filter-type filter2))
                             ;; (not invert) < invert
                             (or (and (not (sidebar-filter-invert filter1))
                                      (sidebar-filter-invert filter2))
                                 (and (eq (sidebar-filter-invert filter1)
                                          (sidebar-filter-invert filter2))
                                      ;; regexp < wildcard < extension
                                      (or (and (eq (sidebar-filter-interpret filter1)
                                                   'regexp)
                                               (memq (sidebar-filter-interpret filter2)
                                                     '(wildcard extension)))
                                          (and (eq (sidebar-filter-interpret filter1)
                                                   'wildcard)
                                               (eq (sidebar-filter-interpret filter2)
                                                   'extension))))))))))))

(defun sidebar-wildcard-list-to-regexp (wildcard-list)
  "Convert list of wildcard expressions WILDCARD-LIST to a regular expression in
\"\\(?:elt1\\|elt2\\|...\\|eltn\\)\" style.

Currently not used and never tested."
  (when wildcard-list
    (let (done)
      (concat
       "^\\(?:" (sidebar-filter-wildcard-to-regexp (car wildcard-list))
       (dolist (filter (cdr wildcard-list) done)
         (setq done
               (concat
                done "\\|" (sidebar-filter-wildcard-to-regexp filter))))
       "\\)$"))))

(defun sidebar-string-list-to-wildcard (string-list &optional sort prepend)
  "Convert list of strings STRING-LIST to a wildcard expression in
\"{elt1\\|elt2\\|...\\|eltn}\" style.  When optional argument SORT is non-nil
sort list before.  When optional argument PREPEND is non-nil prepend PREPEND to
each element."
  (when string-list
    (when sort
      (setq string-list (sort string-list sidebar-string-lessp-function)))
    (let (done)
      (concat
       "{" prepend (car string-list)
       (dolist (string (cdr string-list) done)
         (setq done
               (concat done "," prepend string)))
       "}"))))

(defun sidebar-string-list-to-extension (string-list &optional sort)
  "Convert list of strings STRING-LIST to a string of extensions.  When SORT is
non-nil sort list before."
  (when string-list
    (when sort
      (setq string-list (sort string-list sidebar-string-lessp-function)))
    (mapconcat 'identity string-list " ")))

(defconst sidebar-filter-match-string
  (concat "\\(\\\\.?\\)\\|"             ; 1: \ followed by zero or one character
          "\\([.+^$]\\)\\|"             ; 2: .+^$
          "\\(\\*\\)\\|"                ; 3: *
          "\\(\\?\\)\\|"                ; 4: ?
          "\\({\\)\\|"                  ; 5: {
          "\\(\\[!.*\\]\\)\\|"          ; 6: [!...]
          "\\(\\[\\^.*\\]\\)\\|"        ; 7: [^...]
          "\\(\\[.*\\]\\)"              ; 8: [...]
          ))

(defun sidebar-filter-extension-to-regexp (extension-string)
  "Convert a string of extensions to a regular expression"
  (when extension-string
    (concat ".*\\(?:"
            (regexp-opt (split-string extension-string))
            "\\)")))

(defun sidebar-filter-wildcard-to-regexp (filter)
  "Expand FILTER to regexp according to shell wildcard rules.
If `sidebar-filter-wildcard-expand-braces' is non-nil expand `{a,b,...}'  like
bash, allowing arbitrary nesting.  In this case use of `{', `,' and `}' for any
other purposes must be escaped by preceding `\\'.  Do not signal errors for
malformed filters with, for example, unclosed brackets or braces."
  (when filter
    (let ((index (string-match sidebar-filter-match-string filter)))
      (if index
          (concat
           (substring filter 0 (match-beginning 0))
           (cond
            ((match-beginning 1)        ; \\
             (match-string 0 filter))
            ((match-beginning 2)        ; .+$ -> preprend \\
             (concat "\\" (match-string 0 filter)))
            ((match-beginning 3) ".*")  ; * -> .*
            ((match-beginning 4) ".")   ; ? -> .
            ((match-beginning 5)        ; {
             (if sidebar-filter-wildcard-expand-braces
                 (sidebar-filter-wildcard-expand-brace filter (match-beginning 0))
               "{"))
            ((match-end 6)              ; [!...] -> [^...]
             (concat "[^" 
                     (substring filter (+ (match-beginning 0) 2) (match-end 0))))
            ((match-end 7)              ; [^...] -> [\0...]
             ;; use trick from files.el, don't like it, however
             (concat "[\000^"
                     (substring filter (+ (match-beginning 0) 2) (match-end 0))))
            ((match-end 8)              ; [...] -> [...]
             (match-string 0 filter)))
           (sidebar-filter-wildcard-to-regexp (substring filter (match-end 0))))
        filter))))

(defun sidebar-filter-wildcard-expand-brace (filter index)
  "Expand string FILTER which _must_ start with a left brace.
Adapted from F.J.Wright's `ls-lisp-wildcard-to-regexp-{}'."
  (let ((match-data (match-data))       ; match-data is restored after a failure
        (index-1 (1+ index))
        (index (sidebar-filter-find-comma-or-brace filter index)))
    ;; note: uses "1" and "(1+ index)" since "0" and "index" are reserved for
    ;; the "preceding" character (to handle backslashes in string-match)
    (if (and index (eq (aref filter index) ?,))
        ;; found first comma
        (let ((regexp
               (concat "\\(?:"          ; use shy grouping here
                       (sidebar-filter-wildcard-to-regexp
                        (substring filter index-1 index)))))
          ;; find next comma or closing brace
          (while (and (setq index-1 (1+ index))
                      (setq index
                            (sidebar-filter-find-comma-or-brace filter index))
                      (eq (aref filter index) ?,))
            ;; found next comma
            (setq regexp
                  (concat regexp "\\|"
                          (sidebar-filter-wildcard-to-regexp
                           (substring filter index-1 index)))))
          (if index
              ;; found closing brace
              (save-match-data
                (concat regexp "\\|"
                        (sidebar-filter-wildcard-to-regexp
                         (substring filter index-1 index))
                        "\\)"))
            ;; no closing brace, dont't report error but restore match-data
            (set-match-data match-data)
            "{"))
      ;; no following comma, don't report error but restore match-data
      (set-match-data match-data)
      "{")))

(defun sidebar-filter-find-comma-or-brace (filter index)
  "Return index of first top-level `,' or `}' after `{' in string FILTER at
index INDEX.  Skip nested `{...}' and ignore characters escaped by a preceding
`\\'.
Adapted from F.J.Wright's `ls-lisp-wildcard-to-regexp-find-,}'."
  (let ((index (string-match "[^\\][{,}]" filter index)))
    (while (and index (eq (aref filter (1+ index)) ?{))
      (setq index (sidebar-filter-skip-braces filter (1+ index)))
      (when index (setq index (string-match "[^\\][{,}]" filter index))))
    (when index (1+ index))))

(defun sidebar-filter-skip-braces (filter index)
  "Return index of `}' matching `{' in string FILTER at index INDEX.  Skip
nested `{...}' and ignore characters escaped by a preceding `\\'.
Adapted from F.J.Wright's `ls-lisp-wildcard-to-regexp-skip-{}'."
  (let ((index (string-match "[^\\][{}]" filter index)))
    (while (and index (eq (aref filter (1+ index)) ?{))
      (setq index (sidebar-filter-skip-braces filter (1+ index)))
      (when index (setq index (string-match "[^\\][{}]" filter index))))
    (when index (1+ index))))

(defvar sidebar-filter-system-extensions
  (sidebar-string-list-to-extension
   (cond ((memq system-type '(ms-dos windows-nt))
          '(".o" ".bin" ".bak" ".obj" ".map" ".a" ".ln"))
         ((eq system-type 'vax-vms)
          '(".obj" ".exe" ".bin" ".lbin" ".sbin"
            ".brn" ".rnt" ".mem" ".lni" ".lis"
            ".olb" ".tlb" ".mlb" ".hlb"))
         (t
          '(".o" ".bin" ".lbin" ".fasl" ".ufsl" ".a" ".ln")))
   t))

(defvar sidebar-filter-text-extensions
  (sidebar-string-list-to-extension
   '(".lof" ".glo" ".idx" ".lot" ".dvi" ".fmt" ".tfm"
     ".pdf" ".toc" ".log" ".aux" ".cp" ".fn" ".ky" ".pg"
     ".tp" ".vr" ".cps" ".fns" ".kys" ".pgs" ".tps" ".vrs"
     ".blg" ".bbl") t))

(defvar sidebar-filter-other-extensions
  (sidebar-string-list-to-extension
  '(".elc" ".class" ".fas" ".lib" ".x86f" ".sparcf" ".lo" ".la"
    ".rej" ".orig") t))

(defvar sidebar-filter-dot-dirs-wildcard
  (sidebar-string-list-to-wildcard '("." "..")))

(defvar sidebar-filter-backup-files-wildcard
  (sidebar-string-list-to-wildcard '("#*#" "*~")))

(defvar sidebar-filter-dot-dirs-regexp "^\\.\\.?$")

(defvar sidebar-filter-backup-files-regexp "^#.*#$\\|.*~$")

(defvar sidebar-filter-list-default
  (sort
   (list
    (list t 'dirs "dot-dirs-regexp" 'block nil 1
          (cons 'regexp sidebar-filter-dot-dirs-regexp))
    (list nil 'dirs "dot-dirs-wildcard" 'block nil 1
          (cons 'wildcard sidebar-filter-dot-dirs-wildcard))
    ;;; (list t 'files "backup-files-regexp" 'block nil 1
          ;;; (cons 'regexp sidebar-filter-backup-files-regexp))
    ;;; (list nil 'files "backup-files-wildcard" 'block nil 1
          ;;; (cons 'wildcard sidebar-filter-backup-files-wildcard))
    (list t 'files "system-extensions" 'block nil 1
          (cons 'extension sidebar-filter-system-extensions))
    (list t 'files "text-extensions" 'block nil 1
          (cons 'extension sidebar-filter-text-extensions))
    (list t 'files "other-extensions" 'block nil 1
          (cons 'extension sidebar-filter-other-extensions)))
   'sidebar-filter-lessp))

(defun sidebar-filter-list-lessp (filter1 filter2)
   (if (eq sidebar-filter-list-sort 'phase)
       (< (sidebar-filter-phase filter1)
          (sidebar-filter-phase filter2))
     (let* ((filter-name1 (sidebar-filter-name filter1))
            (filter-name2 (sidebar-filter-name filter2))
            (name1 (if (string-match "\\S " filter-name1)
                       filter-name1
                     (sidebar-filter-string filter1)))
            (name2 (if (string-match "\\S " filter-name2)
                       filter-name2
                     (sidebar-filter-string filter2))))
       (cond
        ((eq sidebar-filter-list-sort 'name)
         (string-lessp name1 name2))
        ((eq sidebar-filter-list-sort 'phase-name)
         (or (< (sidebar-filter-phase filter1)
                (sidebar-filter-phase filter2))
             (string-lessp name1 name2)))))))

(defun sidebar-filter-list-get (symbol)
  "Prepare `sidebar-filter-list' for customization."
  ;; fooling customize
  (if sidebar-filter-list-sort
      (let ((filter-list
             (sort (copy-sequence (if (boundp 'sidebar-filter-list)
                                      sidebar-filter-list
                                    sidebar-filter-list-default))
                   'sidebar-filter-list-lessp))
            (saved-list (get symbol 'saved-value))
            (customized-list (get symbol 'customized-value)))
        (when saved-list
          (put symbol 'saved-value
               (list
                (list 'quote
                      (sort (copy-sequence (eval (car saved-list)))
                            'sidebar-filter-list--lessp)))))
        (when customized-list
          (put symbol 'customized-value
               (list
                (list 'quote
                      (sort (copy-sequence (eval (car customized-list)))
                            'sidebar-filter-list-lessp)))))
        filter-list)
    (if (boundp 'sidebar-filter-list)
        sidebar-filter-list
      sidebar-filter-list-default)))

(defcustom sidebar-filter-list sidebar-filter-list-default
  "sidebar filter list.
The filters in this list apply to all operations on lists of files and
directories provided `sidebar-filter-apply' is not nil.  Insert or delete a
filter by respectively pushing `INS' or `DEL'.  You may customize the following
components of each filter:

 State: Push the `State' button to turn the filter `on' or `off'.

 Domain: Use the `Domain' menu to apply the filter to files only \(`files'),
       directories only \(`dirs'), or files and directories \(`any').

 Name: Each filter has a name which is used to identify it in the `Toggle
       Filters' menu.  If you don't assign a name here, a substring of the
       filter string (see below) is used.

 Type: A `pass' filter asserts that a filename matching the filter gets listed
       unless it has been blocked in an earlier phase.  A `block' filter asserts
       that a filename matching the filter is _not_ listed unless it has passed
       an earlier phase.  A `stop' filter behaves like a `pass' filter with the
       additional effect that a filename _not_ matching the filter is not listed
       unless it has passed an earlier phase.  For obvious reasons, a filtering
       process stops after applying a stop filter.

 Invert: Push the `Invert' button to invert the semantics of matching.  If, for
       example, `Invert' is turned `off', applying a pass filter will assert
       that a filename matching the filter gets listed as described above.  If
       `Invert' is turned `on', a filename gets listed iff it does _not_ match
       the filter.  An inverted `block' filter asserts that a filename is not
       listed iff it does _not_ match the filter.  An inverted `stop' filter,
       finally, behaves like a `block' filter with the additional effect that a
       filename not matching the filter gets listed.

 Phase: Filters are applied in numbered phases.  Filters with lower phase
       numbers are applied first.  If filters have the same phase number, pass
       filters are applied before block filters, and block filters before stop
       filters.

 String: May be `Regexp' - a regular expression, `Wildcard' - a shell wildcard
       expression including bash wildcard constructs enclosed in curly braces,
       or `Extension' - a string of file name extensions separated by spaces or
       newlines.

Regular expressions and wildcards must match complete filenames.  The
directory-part of a filename is _not_ considered by filters.  Extensions should
start with a dot to avoid that filenames not carrying an extension get
incidentally filtered too."
  :type
  '(repeat :entry-format "%i %d %v\n"
           (list :format "%v"
                 (boolean :tag "State" :value t :format "State: %[Toggle%] %v "
                          :on "on " :off "off")
                 (choice :tag "Domain" :value any :format "  Domain: %[Value Menu%] %v "
                         (const :format "any  " any)
                         (const :format "files" files)
                         (const :format "dirs " dirs))
                 (string :tag "Name" :value "" :format "  %t: %v\n" :size 10) 
                 (choice :tag "Type" :value block
                         :format "Type: %[Value Menu%] %v "
                         (const :format "block" block)
                         (const :format "pass " pass)
                         (const :format "stop " stop))
                 (boolean :tag "Invert" :value nil
                          :format "  Invert: %[Toggle%] %v "
                          :on "on " :off "off")
                 (number :tag "Phase" :value 1 :format "  %t: %v\n" :size 6)
                 (choice :tag "String" :value (regexp . "")
                         :format "String: %[Value Menu%] %v"
                         (cons :tag "Regexp" :format "%v"
                               (const :format "" regexp)
                               (regexp :tag "Regexp" :format "   %t: %v"))
                         (cons :tag "Wildcard" :format "%v"
                               (const :format "" wildcard)
                               (string :tag "Wildcard" :format " %t: %v"))
                         (cons :tag "Extension" :format "%v"
                               (const :format "" extension)
                               (string :tag "Extension" :format "%t: %v")))))
  :get 'sidebar-filter-list-get
  :set (lambda (symbol value)
         (set-default symbol value)
         (when (fboundp 'sidebar-filter-list-regexp)
           (sidebar-filter-list-regexp)))
  :group 'sidebar-filter)
(make-variable-buffer-local 'sidebar-filter-list)

(defvar sidebar-filter-file-list nil)
(make-variable-buffer-local 'sidebar-filter-file-list)

(defvar sidebar-filter-directory-list nil)
(make-variable-buffer-local 'sidebar-filter-directory-list)

(defun sidebar-filter-list-to-regexp-list (filter-list)
  (let (filter-regexp-list)
    (while filter-list
      (if (caar filter-list)
          ;; active filter
          (let* ((filter (car filter-list))
                 (type (sidebar-filter-type filter))
                 (invert (sidebar-filter-invert filter))
                 (phase (sidebar-filter-phase filter))
                 (interpret (sidebar-filter-interpret filter))
                 (regexp (if (eq interpret 'wildcard)
                             (sidebar-filter-wildcard-to-regexp
                              (sidebar-filter-string filter))
                           (sidebar-filter-string filter))))
            (setq filter-list (cdr filter-list))
            (while (and filter-list
                        (setq filter (car filter-list))
                        (sidebar-filter-state filter)
                        (= (sidebar-filter-phase filter) phase)
                        (eq (sidebar-filter-type filter) type)
                        (eq (sidebar-filter-invert filter) invert)
                        (eq (sidebar-filter-interpret filter) interpret))
              (setq regexp
                    (cond
                     ((eq interpret 'regexp)
                      (concat regexp "\\|" (sidebar-filter-string filter)))
                     ((eq interpret 'extension)
                      (concat regexp " " (sidebar-filter-string filter)))
                     ((eq interpret 'wildcard)
                      (concat regexp "\\|" (sidebar-filter-wildcard-to-regexp
                                            (sidebar-filter-string filter))))))
              (setq filter-list (cdr filter-list)))
            (setq filter-regexp-list
                  (cons (list type invert
                              (concat
                               "^\\(?:"
                               (if (memq interpret '(regexp wildcard))
                                   regexp  ; filter1\\|filter2\\|...\\|filtern
                                 (sidebar-filter-extension-to-regexp regexp))
                               "\\)$"))
                        filter-regexp-list)))
        ;; inactive filter
        (setq filter-list (cdr filter-list))))
    filter-regexp-list))

(defun sidebar-filter-list-regexp ()
  "Calculate current filter list regexps."
  (let ((filter-list
         (copy-sequence
          (if (boundp 'sidebar-filter-list)
              sidebar-filter-list
            sidebar-filter-list-default)))
        file-filter-list directory-filter-list
        file-regexp-list directory-regexp-list)
    ;; make separate lists for domains
    (dolist (filter filter-list)
      (let ((domain (sidebar-filter-domain filter)))
        (cond
         ((eq domain 'any)
          (setq file-filter-list
                (cons filter file-filter-list))
          (setq directory-filter-list
                (cons filter directory-filter-list)))
         ((eq domain 'files)
          (setq file-filter-list
                (cons filter file-filter-list)))
         ((eq domain 'dirs)
          (setq directory-filter-list
                (cons filter directory-filter-list))))))
    ;; sort filter lists
    (setq sidebar-filter-file-list
          (nreverse
           (sidebar-filter-list-to-regexp-list
            (sort file-filter-list 'sidebar-filter-lessp))))
    (setq sidebar-filter-directory-list
          (nreverse
           (sidebar-filter-list-to-regexp-list
            (sort directory-filter-list 'sidebar-filter-lessp))))))

(defsubst sidebar-filter-out (name filter-list)
  "Return non-nil if NAME shall be filtered out."
  (when sidebar-filter-apply
    (catch 'out
      ;; anything not caught by this dolist is not filtered
      (dolist (filter filter-list)
        (if (nth 1 filter)
            ;; invert filter
            (if (string-match (nth 2 filter) name)
                ;; match
                (when (eq (car filter) 'stop)
                  ;; matching inverted stop filter => reject
                  (throw 'out t))
              ;; mismatch
              (cond
               ((memq (car filter) '(pass stop))
                ;; mismatching inverted pass or stop filter => accept
                (throw 'out nil))
               ((eq (car filter) 'block)
                ;; mismatching inverted block filter => reject
                (throw 'out t))))
          ;; don't invert filter
          (if (string-match (nth 2 filter) name)
              ;; match
              (cond
               ((memq (car filter) '(pass stop))
                ;; matching pass or stop filter => accept
                (throw 'out nil))
               ((eq (car filter) 'block)
                ;; matching block filter => reject
                (throw 'out t)))
            (when (eq (car filter) 'stop)
              ;; mismatching stop filter => reject
              (throw 'out t))))))))

;; _____________________________________________________________________________
;;                                                                              
;;;			      faces and colors					
;; _____________________________________________________________________________
;;                                                                              
(defgroup sidebar-faces nil
  "Sidebar faces and colors."
  :version "22.1"
  :group 'sidebar)

(defface sidebar-modeline-prefix '((t (:foreground "Turquoise")))
  "Face for prefix in modeline.
The modeline prefix displays additional information about values like that of
`sidebar-files-sort' and `sidebar-reverse' in the modeline of
the sidebar."
  :group 'sidebar-faces)

(defface sidebar-hidden '((t (:inherit fringe)))
  "Face of empty lines at beginning and end of sidebars.
This face is used iff at least one of its attributes has been set.  By default,
this face equals that of the window fringes.  If no attribute has been set, the
face of the sidebar list is used.  You should specify the background attribute
only."
  :group 'sidebar-faces)

(defface sidebar-current-line-selected-window
  '((t (:inverse-video t)))
  "Face to highlight current line in a selected sidebar.
This face is mixed with that of the current sidebar list and the face of the
object on the current line. By default, `inverse-video' is used."
  :group 'sidebar-faces)

(defface sidebar-current-line-non-selected-window '((t (:box t)))
  "Face to highlight current line in non-selected sidebars.
This face is mixed with that of the current sidebar list and the face of the
object on the current line."
  :group 'sidebar-faces)

(defface sidebar-directory-name '((t (:foreground "Red3")))
  "Face for directory names in directory lists."
  :group 'sidebar-faces)

(defface sidebar-file-name '((t (:foreground "#000040")))
  "Face for file names in directory lists."
  :group 'sidebar-faces)

(defface sidebar-empty-directory '((t (:foreground "Red3")))
  "Face representing an empty directory in directory lists."
  :group 'sidebar-faces)

(defface sidebar-bookmark-directory-name
  '((t (:inherit sidebar-directory-name)))
  "Face for directory names in bookmark lists."
  :group 'sidebar-faces)

(defface sidebar-bookmark-file-name
  '((t (:inherit sidebar-file-name)))
  "Face for names of not expandable files in bookmarks lists."
  :group 'sidebar-faces)

(defface sidebar-bookmark-expandable-file-name
  '((t (:background "LightCyan3")))
  "Face for overlay of expandable files in bookmark lists."
  :group 'sidebar-faces)

(defface sidebar-bookmark-expanded-file-name
  '((t (:background "LightCyan3")))
  "Face for overlay of expanded files in bookmark lists."
  :group 'sidebar-faces)

(defface sidebar-bookmark-name
  '((t (:foreground "#000040" :background "LightCyan2")))
  "Face for bookmark names in bookmark lists."
  :group 'sidebar-faces)

(defface sidebar-bookmark-echo
  '((t (:foreground "LightCyan2" :background "#000040")))
  "Face for text corresponding to bookmark on current line in bookmark lists.
The default attributes equal that of the current sidebar line."
  :group 'sidebar-faces)

(defface sidebar-tag-name '((t (:foreground "#000040")))
  "Face for tag names in tag lists."
  :group 'sidebar-faces)

(defface sidebar-tag-header
  '((t (:inherit sidebar-tag-name :bold t)))
  "Face for tag headers in tag-lists.
Tag headers are group headers supplied by `imenu' for various languages
typically separating objects like variables, types, and functions."
  :group 'sidebar-faces)

(defface sidebar-tag-prefix
  '((t (:inherit sidebar-tag-name :bold t)))
  "Face for prefixes of tag groups in tag lists.
Tag group prefixes are common substrings of subsequent entries in tag lists."
  :group 'sidebar-faces)

(defface sidebar-tag-echo
  '((t (:foreground "#FFEFDF" :background "#000040")))
  "Face for text corresponding to tag on current line in tag-mode.
The default attributes equal that of the current sidebar line."
  :group 'sidebar-faces)

(defface sidebar-buffer-directory-name
  '((t (:inherit sidebar-directory-name)))
  "Face for directory names in buffer lists.
Related buffers are usually in `dired-mode'."
  :group 'sidebar-faces)

(defface sidebar-buffer-name '((t (:foreground "#000040")))
  "Face for names of buffers not representing a directory in buffer lists."
  :group 'sidebar-faces)

(defface sidebar-buffer-modified '((t (:background "Orange2")))
  "Face for names of modified buffers in buffer lists."
  :group 'sidebar-faces)

(defface sidebar-match '((t (:foreground "#000040")))
  "Face for line numbers in match lists."
  :group 'sidebar-faces)

(defface sidebar-match-echo '((t (:background "OliveDrab2")))
  "Face for matching text.
This face will overlay text in the buffer associated with a sidebar.  Hence,
you should only set the background of this face to avoid obscuring faces
assigned by `font-lock-mode' and related packages.  Usually text appears in this
face immediately after invoking `sidebar-match-query' and related commands."
  :group 'sidebar-faces)

(defface sidebar-replace-echo '((t (:background "Cyan1")))
  "Face for replaced text.
This face will overlay text in the sidebar _and_ the buffer associated with
the sidebar.  Hence, you should only set the background of this face to avoid
obscuring faces assigned by `font-lock-mode' and related packages.  Usually text
appears in this face immediately after invoking `sidebar-match-query' and
related commands."
  :group 'sidebar-faces)

(defface sidebar-mismatch-echo '((t (:background "Orange2")))
  "Face for mismatching text.
This face will overlay text in the sidebar _and_ the buffer associated
with the sidebar.  Hence, you should only set the background of this face
to avoid obscuring faces assigned by `font-lock-mode' and related
packages.  Usually text appears in this face after executing a
`sidebar-space' or related command or after modifying the associated
buffer while sidebar displays a match list."
  :group 'sidebar-faces)

(defface sidebar-current-match-echo
  '((t (:foreground "OliveDrab2" :background "#000040")))
  "Face for matching text corresponding to current line in match list.
This face will overlay text in the buffer associated with the corresponding line
of a sidebar match list.  Default foreground is that of
`sidebar-match-echo'.  Default background is that of the foreground of
`sidebar-match'."
  :group 'sidebar-faces)

(defface sidebar-current-replace-echo
  '((t (:foreground "Cyan1" :background "#000040")))
  "Face for replaced text corresponding to current line in match list.
This face will overlay text in the buffer associated with the corresponding line
of a sidebar match list.  Default foreground is that of
`sidebar-replace-echo'.  Default background is that of the foreground of
`sidebar-match'."
  :group 'sidebar-faces)

(defface sidebar-current-mismatch-echo
  '((t (:foreground "Orange2" :background "#000040")))
  "Face for mismatching text corresponding to current line in match list.
This face will overlay text in the buffer associated with the corresponding line
of a sidebar match list.  Default foreground is that of
`sidebar-mismatch-echo'.  Default background is that of the foreground of
`sidebar-match'."
  :group 'sidebar-faces)

(defface sidebar-context '((t (:background "Honeydew")))
  "Face for match context.
This face will overlay text in the buffer associated with the corresponding line
of a sidebar match list.  Hence you should specify its background attribute
only."
  :group 'sidebar-faces)

(defface sidebar-grep-name '((t (:foreground "#000040")))
  "Face of file names in grep lists."
  :group 'sidebar-faces)

(defface sidebar-rename
  '((t (:foreground "black" :background "grey92"
        :box (:line-width 2 :color "grey75" :style pressed-button))))
  "Face of rename overlay.
Foreground _should_ be specified here."
  :group 'sidebar-faces)

(defface sidebar-mark
  '((t (:background "MediumPurple1")))
  "Face for marked items.
This face will overlay text in the buffer associated with the
corresponding line of a sidebar match list.  Hence you should specify its
background attribute only."
  :group 'sidebar-faces)

(defcustom sidebar-directory-background "Ivory"
  "Background of directory lists."
  :type 'color
  :group 'sidebar-faces)

(defcustom sidebar-tag-background "Bisque"
  "Background of tag lists."
  :type 'color
  :group 'sidebar-faces)

(defcustom sidebar-bookmark-background "LightCyan1"
  "Background of bookmark lists."
  :type 'color
  :group 'sidebar-faces)

(defcustom sidebar-buffer-background "Khaki1"
  "Background of buffer lists."
  :type 'color
  :group 'sidebar-faces)

(defcustom sidebar-match-background "OliveDrab2"
  "Background of match lists."
  :type 'color
  :group 'sidebar-faces)

(defcustom sidebar-grep-background "GoldenRod"
  "Background of grep lists."
  :type 'color
  :group 'sidebar-faces)

;; _____________________________________________________________________________
;;                                                                              
;;;                                   macros                                    
;; _____________________________________________________________________________
;;                                                                              
(eval-when-compile
  (defmacro with-valid-sidebar (&rest body)
    "Execute BODY with valid and writable sidebar.
If sidebar is invalid run `sidebar-reconfigure'.  Save match data during
sidebar operations."
    `(let ((inhibit-read-only t))
       (save-match-data
	 (if (sidebar-valid-p)
	     (progn ,@body)
	   (sidebar-reconfigure)))))
  (def-edebug-spec with-valid-sidebar t))

(eval-when-compile
  (defmacro with-saved-window-and-buffer (&rest body)
    "Like `save-selected-window' but make the window's buffer current at restore."
    `(let ((save-window-and-buffer-window (selected-window)))
       (unwind-protect
           (progn ,@body)
         (when (window-live-p save-window-and-buffer-window)
           (select-window save-window-and-buffer-window)
           (set-buffer (window-buffer save-window-and-buffer-window))))))
  (def-edebug-spec with-saved-window-and-buffer t))

;; _____________________________________________________________________________
;;                                                                              
;;;                                requirements                                 
;; _____________________________________________________________________________
;;                                                                              
(require 'sort)
(require 'bookmark)

;; _____________________________________________________________________________
;;                                                                              
;;;                              support functions                              
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-message (format &rest args)
  "Display message, ding, sit-for a second, and return nil."
    (apply 'message format args)
    (ding)
    (sit-for 1)
    nil)

(defsubst sidebar-select-window (window &optional norecord)
  "Select window WINDOW and make it's buffer current."
  (select-window window norecord)
  (set-buffer (window-buffer window))
  window)

(defun sidebar-count-line (&optional at)
  "Return current line number.  With optional argument AT number of line
containing position AT.  Widens buffer before counting.  Doesn't consider
selective display !"
  (save-excursion
    (save-restriction
      (widen)
      (when at (goto-char at))
      (narrow-to-region (point-min) (line-beginning-position))
      (goto-char (point-min))
      (- (buffer-size) (forward-line (buffer-size))))))

;; *** TODO: Make this customizable and use the m&d stuff ***
(defun sidebar-word-symbol ()
  "Return cons cell containing bounds of word-symbol around point.
Return nil if no such word-symbol can be found."
  (let ((at (point))
        from to)
    (skip-syntax-backward "w_")
    (skip-syntax-forward "_")
    (setq from (point))
    (goto-char at)
    (skip-syntax-forward "w_")
    (skip-syntax-backward "_")
    (setq to (point))
    (goto-char at)
    (when (> to from) (cons from to))))

(defun sidebar-name-around-point ()
  "Return begin and end of name around point.
Return nil if there is no such name."
  (save-excursion
    (let (from to bound)
      (when (cond
             ;; search name at point
             ((and (save-excursion
                     (skip-syntax-backward "w_") (setq from (point)))
                   (save-excursion
                     (skip-syntax-forward "w_") (setq to (point)))
                   (> to from)))
             ;; search name between point and line-end-position
             ((and (setq bound (line-end-position))
                   (skip-syntax-forward "^w_" bound)
                   (< (setq from (point)) bound)
                   (skip-syntax-forward "w_")
                   (setq to (point))))
             ;; search name between line-beginning-position and point
             ((and (setq bound (line-beginning-position))
                   (skip-syntax-backward "^w_" bound)
                   (> (setq to (point)) bound)
                   (skip-syntax-backward "w_")
                   (setq from (point)))))
        (cons from to)))))

(defun sidebar-defun-forward (&optional arg)
  (interactive)
  (cond
   ((memq major-mode '(emacs-lisp-mode python-mode cperl-mode))
    ;; privileged modes where end-of-defun more or less works in the
    ;; intended way
    (end-of-defun (or arg 1)))
   ((eq major-mode 'c-mode)
    (re-search-forward "^}" nil t))
   ((and (eq major-mode 'change-log-mode) (not arg))
    (unless (eobp) (forward-char))
    (when (re-search-forward "^[^ \t\n]" nil 'eob)
      (goto-char (match-beginning 0)))
    (skip-syntax-backward " ")
    (forward-line))
   ((and (eq major-mode 'text-mode) (not arg)
	 (prog1
	     (save-excursion
	       (goto-char (point-max))
	       (skip-chars-backward " \t\n")
	       (looking-back "^-+"))
	   (prog1
	       (re-search-forward "^-+" nil t)
	     (goto-char (match-beginning 0))))))
   (t
    (forward-paragraph (or arg 1))
    (skip-syntax-backward " ")
    (forward-line))))

(defun sidebar-defun-backward (&optional arg)
  (interactive)
  (cond
   ((memq major-mode '(emacs-lisp-mode python-mode cperl-mode))
    ;; privileged modes where beginning-of-defun works more or less in
    ;; the indended way
    (beginning-of-defun (or arg 1)))
   ((eq major-mode 'c-mode)
    (re-search-backward "^{"))
   ((and (eq major-mode 'change-log-mode) (not arg))
    (re-search-backward "^[^ \t\n]" nil 'bob))
   ((and (eq major-mode 'text-mode) (not arg)
	 (prog1
	     (save-excursion
	       (goto-char (point-max))
	       (skip-chars-backward " \t\n")
	       (looking-back "^-+"))
	   (or (prog1
		   (re-search-backward "^-+" nil t)
		 (forward-line))
	       (goto-char (point-min))))))
   (t (forward-paragraph (if arg (- arg) -1))
      (skip-syntax-forward " "))))

;;; (global-set-key [(shift control f5)] 'sidebar-defun-forward)
;;; (global-set-key [(shift meta f5)] 'sidebar-defun-backward)

(defun sidebar-defun ()
  "Return cons cell of `beginning-of-defun' and `end-of-defun' of defun `point' is in."
  (save-excursion
    (save-restriction
      (widen)
      (condition-case nil
          (let* ((start 'within)
                 (at (cond
                      ((eobp) (point))
                      ((= (point)
                          (save-excursion
                            (forward-char) (sidebar-defun-backward) (point)))
                       ;; point = beginning of a defun
                       (setq start 'start)
                       (goto-char (1+ (point))))
                      (t (point))))
                 (from (save-excursion (sidebar-defun-backward) (point)))
                 (to (save-excursion (sidebar-defun-forward) (point))))
            ;; we're at TO now
            (condition-case nil
                (if (and (save-excursion
                           (goto-char to) (beginning-of-line)
                           (sidebar-defun-backward) (= (point) from))
                         (<= from at) (<= at to)
                         ;; rule out trivial case
                         (or (/= from (point-min)) (/= to (point-max))))
                    ;; within or at start of defun
                    (list from to start)
                  ;; not within defun
                  (condition-case nil
                      (when (and (save-excursion
                                   (sidebar-defun-backward -1)
                                   (<= at (setq to (point))))
                                 (save-excursion
                                   (sidebar-defun-forward -1)
                                   (<= (setq from (point)) at)))
                        ;; between defuns
                        (list from to 'between))
                    ;; preceding first defun
                    (error (list (point-min) to 'before))))
              ;; within unclosed defun
              (error (list from (point-max) 'after))))
        (error nil)))))

;; _____________________________________________________________________________
;; 										
;;;			 Window / buffer management				
;; _____________________________________________________________________________
;; 										
(defun sidebar-overlays ()
  "Install sidebar overlays."
  (when sidebar-mode
    (let* ((overlay-face
	    (cons
	     'background-color
	     (eval (intern (concat "sidebar-" (symbol-name sidebar-mode) "-background")))))
	   (before-after-face
	    (if (face-differs-from-default-p 'sidebar-hidden)
		'sidebar-hidden
	      overlay-face))
	   before-to after-from)
      (save-excursion
	;; `sidebar-before-overlay'.
	(goto-char (point-min))
	(skip-chars-forward "\n")
	(setq before-to (point))
	(overlay-put sidebar-before-overlay 'face before-after-face)
	(move-overlay sidebar-before-overlay (point-min) before-to)
	;; `sidebar-after-overlay'
	(goto-char (point-max))
	(beginning-of-line)
	(setq after-from (point))
	;; the following space is utterly needed
	(unless (looking-at " ") (insert " "))
	(overlay-put sidebar-after-overlay 'face before-after-face)
	(overlay-put
	 sidebar-after-overlay 'after-string
	 (propertize sidebar-after-string 'face before-after-face))
	(move-overlay sidebar-after-overlay after-from (point-max))
	;; `sidebar-overlay'
	(overlay-put sidebar-overlay 'face overlay-face)
	(move-overlay sidebar-overlay before-to after-from)))))

(defun sidebar-valid-p ()
  "Return t when current buffer is a valid sidebar buffer."
  (let ((buffer-window (get-buffer-window (current-buffer) t)))
    (and
     ;; (1) The current buffer is in `sidebar-mode'.
     (eq major-mode 'sidebar-mode)
     ;; (2) The current buffer is displayed in exactly one window.
     (window-live-p buffer-window)
     (= (length (get-buffer-window-list (current-buffer) nil t)) 1)
     ;; (3) `sidebar-client-window' is live, appears in the same frame,
     ;;     has the same top-line, and is adjacent to our window.
     (window-live-p sidebar-client-window)
     (eq (window-frame sidebar-client-window)
	 (window-frame buffer-window))
     (let ((edges (window-edges buffer-window))
	   (other-edges (window-edges sidebar-client-window)))
       (and (= (nth 1 edges) (nth 1 other-edges))
	    (= (nth 2 edges) (nth 0 other-edges))))
     ;; (4) `sidebar-client-window' does not display a sidebar.
     (with-current-buffer (window-buffer sidebar-client-window)
       (not (eq major-mode 'sidebar-mode)))
     ;; (5) In tag and match lists `sidebar-client-buffer' is live.
     (or (not (memq sidebar-mode '(tag match)))
	 (buffer-live-p sidebar-client-buffer)))))

(defun sidebar-delete-window (&optional window)
  "Delete window WINDOW where WINDOW defaults to `selected-window'."
  (setq window (or window (selected-window)))
  (when (window-live-p window)
    (set-window-dedicated-p window nil)
    ;; Added this on 2009-08-31 because for some reason it always jumped
    ;; to some other window instead.  Shouldn't be necessary with
    ;; present code.
    (let ((client-window sidebar-client-window))
      (condition-case nil
	  (delete-window window)
	(error nil))
      (when (window-live-p client-window)
	(select-window client-window)))))

(defun sidebar-kill (sidebar)
  "Delete sidebar SIDEBAR and all windows displaying it.
Also remove SIDEBAR from `sidebar-list'."
  (setq sidebar-list (delq sidebar sidebar-list))
  (when (buffer-live-p sidebar)
    (dolist (window (get-buffer-window-list sidebar))
      (when (window-live-p window)
	(set-window-dedicated-p window nil)
	(when (fboundp 'set-window-parameter)
	  (set-window-parameter window 'delete-window t))))
    (delete-windows-on sidebar)
    (kill-buffer sidebar)))

(defun sidebar-reconfigure ()
  "Reconfigure sidebars and associated windows.
Kill all invalid sidebars and match overlays in orphaned buffers.  If no
valid sidebars remain remove all sidebar hooks.  For any valid sidebar put
window-overlay."
  (let ((selected-window (selected-window))
	(sidebar-inhibit-reconfigure t)
	window-list)
    ;; Scan `sidebar-list'.
    (dolist (sidebar sidebar-list)
      (if (not (buffer-live-p sidebar))
	  ;; Dead sidebar, remove from `sidebar-list' 
	  (setq sidebar-list (delq sidebar sidebar-list))
	;; Live sidebar.
	(with-current-buffer sidebar
	  (if (not (sidebar-valid-p))
	      ;; Invalid sidebar, kill (should kill rename in progress too).
	      (sidebar-kill sidebar)
	    ;; Valid sidebar.
	    (when (eq sidebar-mode 'match)
	      (when (and (eq (get-buffer-window sidebar) selected-window)
			 (not (eq (window-buffer sidebar-client-window) sidebar-client-buffer)))
		;; Selected sidebar in match-mode requires `sidebar-client-window' to
		;; display `sidebar-client-buffer' (Note: `set-window-buffer' won't make
		;; `sidebar-client-buffer' current since `sidebar-client-window' is not selected).
		(set-window-and-current-buffer sidebar-client-window sidebar-client-buffer))
	      ;; Add `sidebar-client-buffer' to `window-list'.
	      (setq window-list (sidebar-addq sidebar-client-window window-list)))
	    ;; The following three lines added on 2009-12-22 to make
	    ;; sure sidebar is readjusted after `delete-other-windows'.
	    (when (memq this-command '(delete-other-windows split-or-delete-other-windows))
	      (with-selected-window (get-buffer-window sidebar)
		(sidebar-adjust-window)))))))
    ;; Scan `sidebar-buffer-window-list' for invalid entries.
    (dolist (pair sidebar-buffer-window-list)
      (cond
       ((not (buffer-live-p (car pair)))
	;; Dead buffer, remove window from `sidebar-buffer-window-list'.
	(setq sidebar-buffer-window-list
	      (delete pair sidebar-buffer-window-list)))
       ((not (memq (cdr pair) window-list))
	;; Window has no associated sidebar (no sidebar in match mode has added
	;; `window' to `window-list'): Remove window from
	;; `sidebar-buffer-window-list' and remove all match overlays with
	;; according window property.
	(sidebar-match-cleanup (car pair) (cdr pair))
	(setq sidebar-buffer-window-list (delete pair sidebar-buffer-window-list)))))
    ;; Difficult to say whether the following gets done:
    ;; Restore current buffer and selected window.
    (if (or sidebar-list sidebar-buffer-window-list)
	;; Remove ourselves.
	(remove-hook 'post-command-hook 'sidebar-reconfigure)
      ;; All sidebars got killed.
      (sidebar-remove-hooks)
      (when (timerp sidebar-window-scroll-timer)
        (cancel-timer sidebar-window-scroll-timer)))
    (when (window-live-p selected-window)
      ;; Actually select window.
      (select-window selected-window))))

(defun sidebar-split-window-atom (window size side)
  "Our `split-window' function."
  (let ((client-window
	 (with-current-buffer (window-buffer window)
	   sidebar-client-window)))
    (if (window-live-p client-window)
	(with-selected-window client-window
	  (split-window client-window size side))
    (split-window window size side))))

(defun sidebar-delete-window-special (window)
  "Our `delete-window' function."
  (with-selected-window window
    (when (window-live-p sidebar-client-window)
      (let ((client-window sidebar-client-window)
	    (ignore-window-parameters 'pre))
	(set-window-parameter client-window 'delete-other-windows nil)
	(set-window-parameter client-window 'delete-other-windows-function nil)
	(delete-window window)
	(select-window client-window)))))

(defun sidebar-delete-window-atom (window)
  "Our `delete-window' function."
  (if (window-parameter window 'window-atom)
      (with-selected-window window
	(when (window-live-p sidebar-client-window)
	  (let ((client-window sidebar-client-window)
		(ignore-window-parameters 'pre))
	    (set-window-parameter client-window 'delete-window-function nil)
	    (set-window-parameter client-window 'delete-window nil)
	    (set-window-parameter client-window 'delete-other-windows-function nil)
	    (set-window-parameter client-window 'delete-other-windows nil)
	    (delete-window window)
	    (select-window client-window))))
    (set-window-parameter window 'delete-window-function nil)
    (set-window-parameter window 'delete-window nil)
    (set-window-parameter window 'delete-other-windows-function nil)
    (set-window-parameter window 'delete-other-windows nil)
    (delete-window window)))

(defun sidebar-delete-other-windows-atom (window)
  "Our `delete-other-windows' function."
  (if (window-parameter window 'window-atom)
      (with-selected-window window
	(let ((root (window-atom-root window)))
	  (if (eq root (frame-root-window (window-frame window)))
	      (if sidebar-client-window
		  ;; When called in the sidebar, delete the sidebar.
		  (delete-window window)
		;; When called in another window delete the sidebar as well
		;; but don't delete any other subwindows.
		(let ((ignore-window-parameters 'pre))
		  (sidebar-delete-sidebar)
		  ;; (delete-other-windows window)
		  (set-window-parameter window 'delete-other-windows-function nil)))
	    ;; Do what `delete-other-windows' would do here.
	    (delete-other-windows root))))
    (set-window-parameter window 'delete-window-function nil)
    (set-window-parameter window 'delete-window nil)
    (set-window-parameter window 'delete-other-windows nil)
    (delete-other-windows window)))

(defun sidebar-make (&optional nocleanup)
  "Make or reuse existing sidebar.
If current-buffer is a sidebar, try to reuse it.  Otherwise, try to create
a new sidebar.  May remove (by side-effect) invalid sidebars.  Selects and
returns window of sidebar."
  (cond
   ((and (eq major-mode 'sidebar-mode)
	 (or (sidebar-valid-p)
	     ;; Invalid sidebar, reconfigure.
	     (and (sidebar-reconfigure) nil)))
    ;; The current buffer is a valid sidebar, use it.  Quit rename in progress.
    (when sidebar-rename (sidebar-rename-restore))
    ;; Save these values for `sidebar-kill-or-restore'.
    (setq sidebar-old (list sidebar-client-buffer sidebar-file default-directory))
    (select-window (get-buffer-window (current-buffer))))
   ((let ((buffer (current-buffer))
	  (buffer-of-sidebar
	   (catch 'found
	     (dolist (sidebar sidebar-list)
	       (with-current-buffer sidebar
		 (when (and (eq sidebar-client-window (selected-window))
			    (or (sidebar-valid-p)
				;; Invalid sidebar, reconfigure.
				(and (sidebar-reconfigure) nil)))
		   (throw 'found sidebar)))))))
      (when buffer-of-sidebar
	;; Current buffer has a valid sidebar, use it.
	(let* ((file (buffer-file-name (or (buffer-base-buffer buffer) buffer)))
	       (directory (if file (file-name-directory file) default-directory)))
	  (with-current-buffer buffer-of-sidebar
	    ;; Quit rename in progress.
	    (when sidebar-rename (sidebar-rename-restore))
	    (setq sidebar-old (list sidebar-client-buffer sidebar-file default-directory))
	    (setq sidebar-client-buffer buffer)
	    (setq sidebar-file file)
	    (setq default-directory directory)))
	(select-window (get-buffer-window buffer-of-sidebar)))))
   ((or (fboundp 'display-buffer-pop-up-window)
	(fboundp 'display-buffer-pop-up-window-special))
    (condition-case nil
	;; Current buffer is not a sidebar and has no associated
	;; sidebar.
	(let* ((buffer (current-buffer))
	       ;; Handle case where buffer is indirect.
	       (file (buffer-file-name (or (buffer-base-buffer buffer) buffer)))
	       (directory (if file (file-name-directory file) default-directory))
	       (buffer-of-sidebar (generate-new-buffer " *sidebar*"))
	       (window
		(if (fboundp 'display-buffer-pop-up-window-special)
		    (display-buffer-pop-up-window-special
		     buffer-of-sidebar '((selected . left)) '((atomic . nest)))
		  (display-buffer-pop-up-window
		   buffer-of-sidebar '((selected . left)) '((atomic . nest)))))
	       (old-window (selected-window)))
	  (select-window window)
	  (setq sidebar-old nil)
	  (set-window-and-current-buffer (selected-window) buffer-of-sidebar)
	  (sidebar-mode)		; Enter `sidebar-mode' here and only here.
	  (set-window-dedicated-p (selected-window) t)
	  ;; Split client window instead of ours.
	  (set-window-parameter
	   window 'split-window-function 'sidebar-split-window-atom)
	  (set-window-parameter
	   window 'split-window 'sidebar-split-window-atom)
	  ;; Delete window as usual.
	  (set-window-parameter
	   window 'delete-window-function 'sidebar-delete-window-atom)
	  (set-window-parameter
	   window 'delete-window 'sidebar-delete-window-atom)
	  ;; Handle `delete-other-windows'.
	  (set-window-parameter
	   window 'delete-other-windows-function
	   'sidebar-delete-other-windows-atom)
	  (set-window-parameter
	   window 'delete-other-windows 'sidebar-delete-other-windows-atom)
	  (set-window-parameter
	   old-window 'delete-other-windows-function
	   'sidebar-delete-other-windows-atom)
	  (set-window-parameter
	   old-window 'delete-other-windows 'sidebar-delete-other-windows-atom)
	  ;; Let `other-window' select it.
	  (set-window-parameter window 'maybe-other-window t)
	  (setq sidebar-client-buffer buffer)
	  (setq sidebar-file file)
	  (setq default-directory directory)
	  (setq sidebar-client-window old-window))
      (error nil)))
   ((boundp 'window-24)
    (condition-case nil
	;; Current buffer is not a sidebar and has no associated
	;; sidebar.
	(let* ((buffer (current-buffer))
	       ;; Handle case where buffer is indirect.
	       (file (buffer-file-name (or (buffer-base-buffer buffer) buffer)))
	       (directory (if file (file-name-directory file) default-directory))
	       (window
		(if (and (fboundp 'window-nest) (fboundp 'quit-restore-window))
		    (let ((window-nest t))
		      (set-window-parameter nil 'no-split nil)
		      (split-window nil nil 'left))
		  (let ((window-splits 'nest))
		    (set-window-parameter nil 'no-split nil)
		    (split-window nil nil 'left))))
	       (buffer-of-sidebar (generate-new-buffer " *sidebar*"))
	       (old-window (selected-window)))
	  (set-window-parameter old-window 'no-split 'parent)
	  (set-window-parameter old-window 'no-delete 'parent)
	  (set-window-parameter window 'no-split 'parent)
	  (set-window-parameter window 'no-delete 'this)
	  (select-window window)
	  (setq sidebar-old nil)
	  (set-window-and-current-buffer (selected-window) buffer-of-sidebar)
	  (sidebar-mode)		; Enter `sidebar-mode' here and only here.
	  (set-window-dedicated-p (selected-window) t)
	  (setq sidebar-client-buffer buffer)
	  (setq sidebar-file file)
	  (setq default-directory directory)
	  (setq sidebar-client-window old-window))
      (error nil)))
   ;; The following case will be removed.
   ((condition-case nil
	;; Current buffer is not a sidebar and has no associated sidebar, create
	;; new sidebar.
	(let* ((buffer (current-buffer))
	       ;; Handle case where buffer is indirect.
	       (file (buffer-file-name (or (buffer-base-buffer buffer) buffer)))
	       (directory (if file (file-name-directory file) default-directory))
	       (window (split-window nil nil t))
	       (buffer-of-sidebar (generate-new-buffer " *sidebar*")))
	  (setq sidebar-old nil)
	  (set-window-and-current-buffer (selected-window) buffer-of-sidebar)
	  (sidebar-mode)		; Enter `sidebar-mode' here and only here.
	  (set-window-dedicated-p (selected-window) t)
	  (setq sidebar-client-buffer buffer)
	  (setq sidebar-file file)
	  (setq default-directory directory)
	  (setq sidebar-client-window window))
      (error nil)))))

(defun sidebar-switch-mode (mode)
 "Switch to new mode MODE."
 (erase-buffer)
 (when (eq sidebar-mode 'match)
   (sidebar-match-cleanup sidebar-client-buffer sidebar-client-window)
   (setq sidebar-buffer-window-list
	 (delete (cons sidebar-client-buffer sidebar-client-window)
		 sidebar-buffer-window-list)))
 (setq sidebar-mode mode))

(defun sidebar-kill-or-restore ()
  "Kill sidebar or restore old configuration."
  (if (not sidebar-old)
      (sidebar-kill (current-buffer))
    ;; Restore "old" sidebar.
    (setq sidebar-client-buffer (nth 0 sidebar-old))
    (setq sidebar-file (nth 1 sidebar-old))
    (setq default-directory (nth 2 sidebar-old))))

(defsubst sidebar-window-height-min ()
  "Return minimum of heights of `window-of-sidebar' and `sidebar-client-window'."
  (min (window-height (get-buffer-window (current-buffer)))
       (window-height sidebar-client-window)))

(defun sidebar-shrink-window ()
  (let ((window-min-width 1)
        (sidebar-inhibit-reconfigure t))
    (setq window-size-fixed nil)
    (condition-case nil
        (shrink-window (max (- (window-width) sidebar-minimum-width) 0) t)
      (error nil))
    (if (eq sidebar-mode 'match)
	(setq window-size-fixed 'width)
      (setq window-size-fixed nil))))

(defun sidebar-adjust-window ()
  "Adjust width of sidebar."
  (unless (or (with-current-buffer (window-buffer (selected-window))
		;; Do the following for sidebars only.
		(not (eq major-mode 'sidebar-mode)))
	      (string-equal (symbol-name this-command) "mouse-drag-vertical-line")
	      ;; `mouse-drag-vertical-line' when run by down-mouse-1 is followed
	      ;; by `mouse-select-window' when the mouse button is released.
	      (and (string-equal (symbol-name last-command) "mouse-drag-vertical-line")
		   (string-equal (symbol-name this-command) "mouse-select-window")))
    (let ((excess-width (- (window-width) (max sidebar-width 2)))
	  (window-min-width 2)
	  (sidebar-inhibit-reconfigure t))
      (setq window-size-fixed nil)
      ;; the following condition-case is needed to handle other fixed-size windows
      (condition-case nil
	  (cond
	   ((< excess-width 0)
	    (let* ((enlarge-excess-width
		    (if (> (- (window-width) excess-width) sidebar-maximum-width)
			(- sidebar-maximum-width (window-width))
		      (- excess-width)))
		   (sidebar-window-width (window-width sidebar-client-window)))
	      (when (> (- sidebar-window-width enlarge-excess-width) 0)
		;; enlarge iff we don't kill sidebar-client-window
		(enlarge-window-horizontally enlarge-excess-width))))
	   ((or (> excess-width sidebar-width-tolerance)
		(eq this-command 'sidebar-decrease-by-one)
		;; in match-mode we may get pretty narrow
		(and (> excess-width 0) (eq sidebar-mode 'match)))
	    (shrink-window excess-width t)))
	(error nil))
      ;; calculated width becomes new standard
      (setq sidebar-width (window-width))
      (if (eq sidebar-mode 'match)
	  (setq window-size-fixed 'width)
	(setq window-size-fixed nil)))))

(defun sidebar-window-line (&optional point)
  "Return window line of POINT which defaults to point.
This works correctly iff the size of the selected window doesn't change until
next redisplay.  Otherwise `vertical-motion' may return an inaccurate value."
  (save-excursion
    (goto-char (or point (point)))
    (if (<= (point) (window-start))
	0
      (save-restriction
        (widen)
        (narrow-to-region
         (window-start)
         (if (or truncate-partial-width-windows truncate-lines)
             ;; when lines are truncated, narrow to line-beginning-position, otherwise
             ;; vertical-motion is periodically off by one
             (line-beginning-position)
           ;; when lines are not truncated narrow to point perforce
           (point)))
        (goto-char (point-min))
        (vertical-motion (buffer-size))))))

;; *** TODO: The following must be removed ***
(defsubst sidebar-line-number ()
  "Number of current line.  Point must be at bol."
  (count-lines (point-min) (point)))

(defsubst sidebar-line-face ()
  (get-text-property (line-beginning-position) 'face))

(defsubst sidebar-same-files (name-1 name-2)
  "Return t iff NAME-1 and NAME-2 yield same `file-truename'."
  (string-equal (file-truename name-1) (file-truename name-2)))

;;; (defsubst sidebar-same-files (name-1 name-2)
;;;   "Return t iff NAME-1 and NAME-2 yield same `file-truename'."
;;;   (if (memq system-type '(ms-dos windows-nt))
;;;       ;; unconditionally downcase on w32, `w32-long-file-name' might be
;;;       ;; better here, but a general solution for w32 is needed anyway
;;;       (string-equal (downcase (file-truename name-1))
;;;                     (downcase (file-truename name-2)))
;;;     (string-equal (file-truename name-1) (file-truename name-2))))

(defsubst sidebar-property ()
   (get-text-property (line-beginning-position) 'sidebar-property))

(defsubst sidebar-level ()
   (get-text-property (line-beginning-position) 'sidebar-level))

(defun sidebar-put-line (text face &optional property)
  "Insert text, update `sidebar-width', and insert newline."
  (setq sidebar-width (max sidebar-width (length text)))
  (insert (concat text "\n"))
  (put-text-property (line-beginning-position 0) (line-beginning-position) 'face face)
  (when property
    (put-text-property
     (line-beginning-position 0) (line-beginning-position) 'sidebar-property property))
  (sidebar-put-overlay))

(defun sidebar-put-dir-line (text face &optional property)
  "Insert text, update `sidebar-width', and insert newline."
  (setq sidebar-width (max sidebar-width (length text)))
  (insert (concat text "\n"))
  (put-text-property (line-beginning-position 0) (line-beginning-position) 'face face)
  (when property
    (put-text-property
     (line-beginning-position 0) (line-beginning-position) 'sidebar-property property))
  (sidebar-put-overlay t))

(defun sidebar-put-text (face &optional property)
  "Put text properties for text on _previous_ line."
  (put-text-property (line-beginning-position 0) (line-beginning-position) 'face face)
  (when property
    (put-text-property
     (line-beginning-position 0) (line-beginning-position) 'sidebar-property property)))

(defun sidebar-put-char (property value &optional line)
  "Put text PROPERTY for character at beginning of current line to VALUE."
  (put-text-property (line-beginning-position line) (1+ (line-beginning-position line)) property value))

(defun sidebar-attributes-help (window object position)
  (with-saved-window-and-buffer
    (save-excursion
      (sidebar-select-window window)
      (goto-char position)
      (beginning-of-line)
      (sidebar-attributes (window-buffer (selected-window))))))

(defun sidebar-put-overlay (&optional dir)
  "Puts overlay properties for text on _previous_ line."
  (let ((overlay (make-overlay (line-beginning-position 0) (line-end-position 0) nil t)))
    (overlay-put overlay 'evaporate t)
    (when (and dir sidebar-directory-brackets)
      ;; Here overlays don't inherit any more from the surrounding text
      ;; hence they get the default background :-(.
      (overlay-put overlay 'before-string "[")
      (overlay-put overlay 'after-string "]"))
    (overlay-put overlay 'mouse-face 'highlight)
    (overlay-put overlay 'help-echo 'sidebar-attributes-help)))

;; The following version puts a space in front of each line, breaks mouse-face
;; etc. but would make sidebar "nicer" 
;; (defun sidebar-put-overlay (&optional dir)
;;   "Puts overlay properties for text on _previous_ line."
;;   (let ((overlay (make-overlay (line-beginning-position 0) (line-end-position 0) nil t)))
;;     (overlay-put overlay 'evaporate t)
;;     (cond
;;      ((and dir sidebar-directory-brackets)
;;       (overlay-put overlay 'before-string " [")
;;       (overlay-put overlay 'after-string "]"))
;;      (t
;;       (overlay-put overlay 'before-string " ")))
;;     ;; the following is broken if we (i) use brackets and (ii) are not on the
;;     ;; first line of the directory (possibly in some other cases too)
;;     (overlay-put overlay 'mouse-face 'highlight)
;;     ;; the following appears broken too
;;     (overlay-put overlay 'help-echo 'sidebar-attributes-help)))

(defun sidebar-line-file ()
  "Return file-name corresponding to current line of sidebar.
If no such name can be construed return `sidebar-file'."
  (cond
   ((and (memq (sidebar-line-face)
               '(sidebar-directory-name sidebar-file-name))
         (not sidebar-unfolded))
    (expand-file-name
     (buffer-substring-no-properties (line-beginning-position) (line-end-position))))
   ((memq (sidebar-line-face)
          '(sidebar-bookmark-directory-name
            sidebar-bookmark-file-name sidebar-grep-name))
    (sidebar-property))
   ((and (eq (sidebar-line-face) 'sidebar-file-name)
         sidebar-unfolded)
    (if sidebar-file-list-ls
        ;; ls creates a list of full file names
        (elt sidebar-file-list (sidebar-line-number))
      ;; dfa creates a list of conses of file-name-sans and full file name with
      ;; attributes
      (cadr (elt sidebar-file-list (sidebar-line-number)))))
   ((memq (sidebar-line-face)
          '(sidebar-buffer-directory-name sidebar-buffer-name))
    (buffer-file-name (sidebar-property))) 
   (t sidebar-file)))

(defun sidebar-view ()
  "Orderly display `sidebar-client-window'."
  (let ((window-line (sidebar-window-line)))
    (cond
     ((memq (sidebar-line-face)
            '(sidebar-file-name sidebar-bookmark-file-name sidebar-grep-name))
      (with-saved-window-and-buffer
       (sidebar-find-file (sidebar-line-file))
       (recenter window-line)))
     ((eq (sidebar-line-face) 'sidebar-buffer-name)
      (set-window-and-current-buffer ; and select it too
       sidebar-client-window
       (buffer-substring-no-properties (line-beginning-position) (line-end-position))))
     ((eq (sidebar-line-face) 'sidebar-tag-name)
      (let* ((name (get-text-property (point) 'help-echo))
             (from (sidebar-property))
             (overlay sidebar-line-echo-overlay)
	     (window sidebar-client-window))
        (with-saved-window-and-buffer
         (sidebar-find-file)
         (goto-char from)
         (beginning-of-line)
         (when (string-match "^[ \t]*" name)
           ;; remove leading spaces from name, for cperl we might have
           ;; to strip the package prefix as well
           (setq name (substring name (match-end 0))))
         (if (re-search-forward (concat "\\(?:\\W\\|^\\)\\("
					(regexp-quote name)
					"\\)\\(?:\\W\\|$\\)")
				(line-end-position) t)
             ;; highlight what the tag says, this fails if name appears
             ;; before tag on same line, blame the user
             (progn
               (move-overlay
                overlay (match-beginning 1) (match-end 1) (current-buffer))
               (goto-char from))
           ;; highlight entire line if we can't find our string
           (move-overlay overlay (line-beginning-position) (line-end-position) (current-buffer)))
         (overlay-put overlay 'face 'sidebar-tag-echo)
         (overlay-put overlay 'window window)
         ;; the following does not work with Emacs 22.1 imenu's beg/start philosophy
;;;            (goto-char from)
;;;            (let ((to (save-excursion (forward-sexp) (point))))
;;;              (when to
;;;                ;; forward-sexp as fallback 
;;;                (overlay-put overlay 'face 'sidebar-tag-echo)
;;;                (overlay-put overlay 'window (get-buffer-window (current-buffer)))
;;;                (move-overlay overlay from to (current-buffer)))))
         (recenter window-line)
         (setq window-line (sidebar-window-line)))
        (recenter window-line)))
     ((eq (sidebar-line-face) 'sidebar-bookmark-name)
      (if (eq last-command 'sidebar-delete)
          (overlay-put sidebar-line-overlay
                       ;; limbo after deletion
                       'face 'sidebar-current-line-non-selected-window)
        (let* ((bookmark (sidebar-property))
               (bookmark-name (car bookmark))
               (bookmark-name-sans-suffix
                (if (string-match "<[0-9]+>$" bookmark-name)
                    (substring bookmark-name 0 (match-beginning 0))
                  bookmark-name))
               (overlay sidebar-line-echo-overlay)
               (window sidebar-client-window)
	       bookmark-line)
          (with-saved-window-and-buffer
           (let ((offset (sidebar-bookmark-sidebar bookmark)))
             (sidebar-find-file (sidebar-bookmark-file-name bookmark))
             (bookmark-jump bookmark-name)
             (when (numberp offset)
               (condition-case nil
                   (backward-char offset)
                 (error nil))))
           (setq bookmark-line (sidebar-count-line))
           (overlay-put overlay 'face 'sidebar-bookmark-echo)
           (overlay-put overlay 'window window)
           (move-overlay overlay (point)
                         (if (looking-at (regexp-quote bookmark-name-sans-suffix))
                             (match-end 0)
                           (line-end-position))
                         (current-buffer)))
          (sidebar-update-mode-line)
          (when (> window-line bookmark-line)
            ;; bookmark near beginning
            (recenter bookmark-line)
            (setq window-line bookmark-line))
          (with-saved-window-and-buffer
           (sidebar-select-window sidebar-client-window)
           (recenter window-line)))))
     ((eq (sidebar-line-face) 'sidebar-match)
      (let (overlay)
        ;; synchronizing lines is a bit tricky
        (unless (string-match
                 "sidebar-\\(down\\|up\\)-\\(drag\\|preserve\\)"
                 (symbol-name this-command))
          (when (> window-line (- (window-height) sidebar-match-lines-after 3))
            ;; at least `sidebar-match-after-lines' should remain visible after
            ;; each match; "3" is a magic value that might not always work
            (recenter (- (window-height) sidebar-match-lines-after 3))
            (setq window-line (sidebar-window-line)))
          (setq window-line
                (max window-line
                     ;; try to make start of defun visible
                     (+ sidebar-match-lines-before
                        (if sidebar-match-context 2 0)))))
        (with-saved-window-and-buffer
         (setq overlay (sidebar-property))
         (sidebar-select-window sidebar-client-window)
         (when overlay (goto-char (overlay-start overlay)))
         ;; for the use of "3" see above
         (recenter (min window-line (- (window-height) 3)))
         (setq window-line (sidebar-window-line)))
        (recenter window-line))))))

(defun sidebar-find-file (&optional file)
  "Display file FILE in `sidebar-client-window'.  FILE defaults to `sidebar-file'.
In addition set `sidebar-client-buffer', `sidebar-file', and `default-directory'.
Exit with `sidebar-client-window' selected."
  (let* ((file (or file sidebar-file))
         (buffer
	  (let (inhibit-read-only)
	    (find-file-noselect file))))
    (when buffer
      (setq buffer (or (buffer-base-buffer buffer) buffer))
      (set-window-and-current-buffer sidebar-client-window buffer)
      (setq sidebar-client-buffer buffer)
      (setq sidebar-file file)
      (setq default-directory (or default-directory (file-name-directory file)))
      (sidebar-select-window sidebar-client-window))))

;; _____________________________________________________________________________
;;                                                                              
;;;                            interactive functions                            
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-revert (&optional ignore-auto noconfirm)
  "Revert sidebar."
  (interactive)
  (with-valid-sidebar
   (cond
    ((eq sidebar-mode 'directory)
     (if sidebar-unfolded
         (sidebar-file-list)
       (let ((name (buffer-substring-no-properties (line-beginning-position) (line-end-position))))
         ;; stay with current object
         (sidebar-directory-list
          default-directory name
          (eq (sidebar-line-face) 'sidebar-directory-name)))))
    ((eq sidebar-mode 'tag)
     (sidebar-select-window sidebar-client-window)
     (sidebar-tag-list))
    ((eq sidebar-mode 'buffer)
     (sidebar-buffers (buffer-list)))
    ((eq sidebar-mode 'bookmark)
     (if sidebar-unfolded
         (sidebar-bookmark-unfold) (sidebar-bookmarks)))
    ((eq sidebar-mode 'match)
     ;; an approximation of the earlier match
     (sidebar-match-refresh)))))

(defun sidebar-down (&optional arg)
  "Move to next line of sidebar.  With ARG move by ARG lines.
If necessary scroll window to make target line visible.  When on last line of
sidebar and `sidebar-wrap' is non-nil move to first line of sidebar."
  (interactive "P")
  (setq mark-active nil)
  (sidebar-next-visible-line arg)
  (when (and sidebar-wrap (save-excursion (skip-syntax-forward " ") (eobp)))
    (goto-char (point-min)))
  (sidebar-reposition))
;;   (unless (pos-visible-in-window-p (point) (selected-window))
;;     (recenter (- (window-height) 2)))) ; what for did I use that ?

(defun sidebar-down-preserve (&optional arg)
  "Move to next line of sidebar preserving screen position.  With ARG move by ARG lines.
When on last line of sidebar and `sidebar-wrap' is non-nil move to first
line of sidebar."
  (interactive "P")
  (let ((window-line (sidebar-window-line)))
    (sidebar-down (or arg 1))
    (recenter window-line)))

(defun sidebar-down-drag (&optional arg)
  "Drag sidebar down by one line."
  (interactive)
  (setq mark-active nil)
  (let ((window-line (sidebar-window-line)))
    (if (or (> window-line (- (sidebar-window-height-min) 3))
            ;; for the use of "3" compare `sidebar-view'
            (save-excursion
              (goto-char (point-min))
              (skip-syntax-forward " ")
              (pos-visible-in-window-p (point))))
        (ding)
      (setq window-line (+ window-line (or arg 1)))
      (recenter window-line))
    ;; hopefully the following is not needed
;;       (when (or (memq sidebar-mode '(tag match))
;;                 (eq (sidebar-line-face) 'sidebar-bookmark-name))
;;         (unless (with-current-buffer sidebar-client-buffer
;;                   (or truncate-partial-width-windows truncate-lines))
;;           ;; continuation lines must show up
;;           (sidebar-adjust-window)))
    (when (memq (sidebar-line-face)
                '(sidebar-tag-name sidebar-bookmark-name))
      (sidebar-view))))

(defun sidebar-down-view (&optional arg)
  "Move to next line of sidebar and synchronize `sidebar-client-window'.
Move to beginning of sidebar if at end and `sidebar-wrap' is non-nil.  In
match-mode attempt to restore last grep and display next file containing
matches."
  (interactive "P")
  (if (and (eq sidebar-mode 'match)
           sidebar-grep-last)
      (sidebar-grep-restore 'down)
    (sidebar-down (or arg 1))
    (sidebar-view)))

(defun sidebar-up (&optional arg)
  "Move to previous line of sidebar.
Move to end of sidebar if at beginning and `sidebar-wrap' is non-nil."
  (interactive "P")
  (setq mark-active nil)
  (sidebar-prev-visible-line arg)
  (when (and (save-excursion (skip-syntax-backward " ") (bobp))
             sidebar-wrap)
    (goto-char (point-max)))
  (sidebar-reposition))
;;   (unless (pos-visible-in-window-p (point) (selected-window))
;;     ;; recenter on first line of window, what for ?
;;     (recenter 0)))

(defun sidebar-up-preserve (&optional arg)
  "Move to previous line of sidebar preserving screen position.
Move to end of sidebar if at beginning and `sidebar-wrap' is non-nil."
  (interactive "P")
  (let ((window-line (sidebar-window-line)))
    (sidebar-up (or arg 1))
    (recenter window-line)))

(defun sidebar-up-drag (&optional arg)
  "Move sidebar up by one line."
  (interactive)
  (setq mark-active nil)
  (let ((window-line (sidebar-window-line)))
    ;; see remark to sidebar-down-drag
    (if (= window-line 0)
        (ding)
      (setq window-line (- window-line (or arg 1)))
      (recenter window-line))
    ;; see above
;;       (when (or (memq sidebar-mode '(tag match))
;;                 (eq (sidebar-line-face) 'sidebar-bookmark-name))
;;         (unless (with-current-buffer sidebar-client-buffer
;;                   (or truncate-partial-width-windows truncate-lines))
;;           ;; continuation lines must show up
;;           (sidebar-adjust-window)))
    (when (memq (sidebar-line-face)
                '(sidebar-tag-name sidebar-bookmark-name))
      (sidebar-view))))

(defun sidebar-up-view (&optional arg)
  "Move to previous line of sidebar and synchronize `sidebar-client-window'.
Move to end of sidebar if at beginning and `sidebar-wrap' is non-nil.  In
match-mode attempt to restore last grep and display previous file containing
matches."
  (interactive "P")
  (if (and (eq sidebar-mode 'match)
           sidebar-grep-last)
      (sidebar-grep-restore 'up)
    (sidebar-up (or arg 1))
    (sidebar-view)))

(defun sidebar-scroll-down ()
  "Scroll sidebar down."
  (interactive)
  ;; 1 counts mode-line
  (sidebar-prev-visible-line (- (window-height) next-screen-context-lines 1)))

(defun sidebar-scroll-up ()
  "Scroll sidebar up."
  (interactive)
  ;; 1 counts  mode-line
  (sidebar-next-visible-line (- (window-height) next-screen-context-lines 1)))

(defun sidebar-home ()
  "Go to beginning of sidebar."
  (interactive)
  (goto-char (point-min)))

(defun sidebar-home-top ()
  "Move current sidebar line to top."
  (interactive)
  (recenter 0))

(defun sidebar-end-center ()
  "Move current sidebar line somewhere to center of screen."
  (interactive)
  (recenter (/ (window-height) 2)))

(defun sidebar-end ()
  "Go to end of sidebar."
  (interactive)
  (goto-char (point-max)))

(defun sidebar-down-jump (&optional arg)
  "Jump forward to line respecting current context.
In match-mode jump to line listing first occurrence in another defun.  In other
modes jump to line starting with next letter.  Move to beginning of sidebar if
at end and `sidebar-wrap' is non-nil.  ARG is ignored."
  (interactive)
  (sidebar-jump-down-or-up 'down))

(defun sidebar-up-jump (&optional arg)
  "Jump backward to line respecting current context.
In match-mode jump to line listing first occurrence in current defun or the line
listing last occurrence in preceding defun.  In other modes jump to line
starting with previous letter.  Move to end of sidebar if at beginning and
`sidebar-wrap' is non-nil.  May not work properly when `sidebar-indent'
equals zero.  ARG is ignored."
  (interactive)
  (sidebar-jump-down-or-up 'up))

(defun sidebar-jump-down-or-up (move)
  "Jump to appropriate line."
  (let ((window-line (sidebar-window-line))
        (permit-wrap t) point)
    (cond
     ((and (memq (sidebar-line-face)
                 '(sidebar-bookmark-directory-name
                   sidebar-bookmark-file-name))
           (save-excursion
             (forward-line (when (eq move 'up) -1))
             (eq (sidebar-line-face) 'sidebar-bookmark-name)))
      ;; skip bookmarks and go to next file/directory entry
      (while (and (zerop (forward-line (when (eq move 'up) -1)))
                  (eq (sidebar-line-face) 'sidebar-bookmark-name)))
      (beginning-of-line))
     ;; jump to context start/end
     ((and (eq (sidebar-line-face) 'sidebar-match)
           sidebar-context-overlay)
      (let* ((at (point))
             (to (if (eq move 'down)
                     (or (overlay-end sidebar-context-overlay)
                         (line-beginning-position 2))
                   (goto-char
                    (or (overlay-start sidebar-context-overlay) at))
                   (line-beginning-position))))
        (if (= to at)
            ;; move in any case
            (cond
             ((eq move 'down)
              (unless (save-excursion (skip-syntax-forward " ") (eobp))
                (forward-line)
                (setq permit-wrap nil)))
             ((eq move 'up)
              (unless (save-excursion (skip-syntax-backward " ") (bobp))
                (forward-line -1)
                (setq permit-wrap nil))))
          (goto-char to)
          (unless (looking-at " *$")
            (setq permit-wrap nil)))))
     (t
      ;; jump to line starting with different character
      (let ((bookmark-dir-or-file-p
             (memq (sidebar-line-face)
                   '(sidebar-bookmark-directory-name
                     sidebar-bookmark-file-name)))
            (regexp
             (save-excursion
	       (skip-syntax-forward " ")
	       (unless (eolp)
		 (concat "\\s-*"
			 (regexp-quote (buffer-substring-no-properties
					(point) (1+ (point))))))))
            (point (point)))
        ;; skip invisible lines or lines starting with same character sequence
        (when regexp
          (when (eq move 'down) (forward-line))
          (while (and (not (eobp))
		      (not (bobp))
		      ;; <---------- we need a special fun here to skip
		      ;; invisible text till min / max invisible overlay around
		      ;; point
		      (or (get-char-property (point) 'invisible)
			  (looking-at regexp)))
	    (forward-line (when (eq move 'up) -1))
            ;; forbid wrapping, <----- needed ?
            (setq permit-wrap nil)))
        (when (= point (point))
          ;; move one line in _any_ case, this won't work with invisible text
          (forward-line (when (eq move 'up) -1))
	  (when (get-char-property (point) 'invisible)
	    (goto-char point)))
        (when (and bookmark-dir-or-file-p
                   (eq (sidebar-line-face) 'sidebar-bookmark-name))
          ;; skip bookmarks and go to next file/directory entry (if we were on a
          ;; non-bookmark item stay on a non-bookmark item)
          (while (and (zerop (forward-line (when (eq move 'up) -1)))
                      (eq (sidebar-line-face) 'sidebar-bookmark-name)))
          (beginning-of-line)))))
    (cond
     ((and (eq move 'down)
           (save-excursion (skip-syntax-forward " ") (eobp))
           sidebar-wrap permit-wrap)
      (goto-char (point-min)))
     ((and (eq move 'up)
           (save-excursion (skip-syntax-backward " ") (bobp))
           sidebar-wrap permit-wrap)
      (goto-char (point-max))))
    (sidebar-reposition)))

(defun sidebar-grep-restore (&optional move)
  "Restore last grep when available."
  (interactive)
  (let ((inhibit-read-only t))
    (when sidebar-grep-last
      ;; restore last grep
      (sidebar-switch-mode 'grep)
      (insert sidebar-grep-last)
      (setq sidebar-width sidebar-grep-last-width)
      (goto-char sidebar-grep-last-point)
      ;; recenter, maybe we can do better some time
      (recenter (/ (window-height) 2))
      (when move
        (if (eq move 'down) (sidebar-down) (sidebar-up))
        (sidebar-open)))))

(defun sidebar-close ()
  "Close current object.
Act according to the type of the object on the current line:

- for a directory or file display parent directory

- for a tag header or prefix collapse associated tags

- for a tag go to line of associated header or prefix

- for a bookmarked file collapse associated bookmarks

- for a bookmark go to line of bookmarked file

- for a match go to last grep list"
  (interactive)
  (with-valid-sidebar
   (cond
    ((memq (sidebar-line-face)
           '(sidebar-file-name
             sidebar-directory-name sidebar-empty-directory))
     ;; directory or file: display parent directory
     (let ((directory
            (file-name-directory (directory-file-name default-directory)))
           (file
            (file-name-nondirectory (directory-file-name default-directory))))
       (when (and directory
                  (or (not (string-equal directory default-directory))
                      sidebar-unfolded))
         (setq default-directory directory)
         (sidebar-directory-list directory file t))))
    ((eq (sidebar-line-face) 'sidebar-bookmark-file-name)
     ;; bookmarked file: collapse associated bookmarks
     (sidebar-bookmark-collapse t))
    ((eq (sidebar-line-face) 'sidebar-bookmark-name)
     ;; bookmark: go to associated file
     (while (and (zerop (forward-line -1))
                 (eq (sidebar-line-face) 'sidebar-bookmark-name))))
    ((memq (sidebar-line-face)
           '(sidebar-tag-name sidebar-tag-header sidebar-tag-prefix))
     (let (overlay at)
       (cond
	((and (setq overlay (get-text-property (point) 'overlay))
	      (not (overlay-get overlay 'invisible)))
	 ;; tag header/prefix with visible overlay ~> make invisible
	 (overlay-put overlay 'invisible t))
	((let ((overlays (overlays-at (point)))
	       (last 0)
	       start)
	   (when overlays
	     ;; try to go to header/prefix with highest start position before
	     ;; point; this header/prefix must have a visible overlay that
	     ;; covers point, hence search for overlay with highest start and go
	     ;; to its header/prefix
	     (dolist (overlay overlays)
	       (when (overlay-get overlay 'sidebar-tag)
		 (setq start (overlay-start overlay))
		 (when (> start last)
		   (setq last start))))
	     (when (> last 0)
	       (goto-char last)
	       ;; header/prefix is on previous line
	       (sidebar-prev-visible-line)
	       t))))
	((and (setq at (previous-single-char-property-change (point) 'sidebar-tag))
	      (> at (point-min))
	      (or (goto-char at) t)
	      (setq at (previous-single-char-property-change at 'sidebar-tag))
;;; 	      (> at (point-min))
	      (let ((overlays (overlays-at at))
		    (last (point-max))
		    start)
		;; try to go to highest level / highest start position
		;; header/prefix preceding the current position; this is the
		;; highest position header/prefix preceding point whose overlay
		;; covers all overlays between the header/prefix and point
		(when overlays
		  (dolist (overlay overlays)
		    (when (overlay-get overlay 'sidebar-tag)
		      (setq start (overlay-start overlay))
		      (when (< start last)
			(setq last start))))
		  (when (< last (point-max))
		    (goto-char last)))
		t)
	      (sidebar-prev-visible-line)
	      t)))))
    ((eq (sidebar-line-face) 'sidebar-match)
     ;; match: restore last grep
     (sidebar-grep-restore)))))

(defun sidebar-open-tag ()
  "Display tags for file on current line of sidebar."
  (interactive)
  (sidebar-open 'tag))

(defun sidebar-open-bookmark ()
  "Display bookmarks for file on current line of sidebar."
  (interactive)
  (sidebar-open 'bookmark))

(defun sidebar-open-kill ()
  "Display object on current line and kill sidebar."
  (interactive)
  (sidebar-open 'kill))

(defun sidebar-open-view ()
  "Display object on current line in sidebar-client-window but don't switch to it."
  (interactive)
  (sidebar-open 'view))

(defun sidebar-file-symlink (file)
  "Return referenced file if FILE is a symlink and can be parsed and FILE otherwise."
  (file-truename
   (if (and file (file-symlink-p file))
       (if (memq system-type '(ms-dos windows-nt))
           (when (fboundp 'w32-symlinks-parse-shortcut)
             (or (w32-symlinks-parse-shortcut file) file))
         (file-symlink-p file))
     file)))

(defun sidebar-open (&optional option)
  "Display object on current line.  OPTION is ignored in interactive use.
Interactively display is controlled by the type of the object on the current line:

- a directory is expanded

- a file is displayed in sidebar-client-window

- for a bookmarked file either display its bookmarks or display the file in the
  sidebar-client-window

- for a grepped file display matches in sidebar-client-window

- tag-headers and tag-prefixes are expanded

- for a tag or bookmark jump to it in sidebar-client-window

OPTION may be tag, bookmark, view, or kill."
  (interactive)
  (with-valid-sidebar
   (let ((window-line (sidebar-window-line))
         (file (sidebar-line-file)))
     (when (eq (sidebar-line-face) 'sidebar-grep-name)
       ;; save grep list
       (setq sidebar-grep-last (buffer-substring (point-min) (point-max)))
       (setq sidebar-grep-last-point (point))
       (setq sidebar-grep-last-width sidebar-width))
     (cond
      ((eq (sidebar-line-face) 'sidebar-match)
       ;; match: go to corresponding line, ignore all options but kill
       (with-saved-window-and-buffer
	(let ((overlay (sidebar-property)))
	  (sidebar-select-window sidebar-client-window)
	  (goto-char (overlay-start overlay))
	  (recenter window-line)))
       (if (eq option 'kill)
           (sidebar-delete-window)
         ;; match lists don't shrink or expand
         (sidebar-select-window sidebar-client-window)))
      ((memq (sidebar-line-face)
             '(sidebar-buffer-name sidebar-buffer-directory-name))
       ;; buffer: display corresponding buffer, optionally tags or bookmarks
       (let ((buffer (sidebar-property)))
         (when (and buffer (and (buffer-live-p buffer)))
           (set-window-and-current-buffer sidebar-client-window buffer)
           (setq sidebar-client-buffer buffer)
           (setq sidebar-file (buffer-file-name buffer))
           (cond
            ((and (eq option 'tag) file (file-readable-p file))
             (sidebar-tag-list))
            ((and (eq option 'bookmark) file (file-readable-p file))
             (if (sidebar-bookmark-named-file file)
                 (progn                 ; display bookmarks
                   (sidebar-bookmark-list file)
                   (with-saved-window-and-buffer (sidebar-find-file file)))
               (sidebar-message "No bookmarks for %s" file)))
            ((eq option 'kill) (sidebar-delete-window))
            ((eq option 'view))
            (t (sidebar-shrink-window)
               (sidebar-select-window sidebar-client-window))))))
      ;; anything below requires an existing and readable file, for a symlink
      ;; try - by side-effect - to get referenced file first
      ((and (setq file (sidebar-file-symlink file)) (not file))
       (sidebar-message "No file on this line"))
      ((not (file-exists-p file))
       (sidebar-message "File %s does not exist" file))
      ((not (file-readable-p file))
       (sidebar-message "File %s is not readable" file))
      ((or (memq (sidebar-line-face)
                 '(sidebar-directory-name
                   sidebar-bookmark-directory-name))
           (and (eq (sidebar-line-face) 'sidebar-file-name)
                (file-directory-p file)))
       ;; directory: display in sidebar or dired
       (if option
           ;; dired file in sidebar-client-window
           (let ((window sidebar-client-window))
             (with-saved-window-and-buffer
	      (sidebar-select-window window)
	      (dired (file-name-as-directory file)))
             (cond
              ((memq option '(tag bookmark))
               (sidebar-shrink-window)
               (sidebar-select-window window))
              ((eq option 'kill) 
               (sidebar-delete-window)
               (sidebar-select-window window))))
         ;; sidebar directory
         (setq sidebar-mode 'directory)
         (setq default-directory (file-name-as-directory file))
         (sidebar-directory-list default-directory)))
      ((memq (sidebar-line-face)
             '(sidebar-file-name sidebar-bookmark-file-name
                                        sidebar-grep-name))
       ;; file: display file, optionally tags or bookmarks
       (cond
        ((eq option 'tag)
         (sidebar-tag-list))
        ((eq option 'bookmark)
         (if (sidebar-bookmark-named-file file)
             (progn                     ; display bookmarks
               (sidebar-bookmark-list file)
               (with-saved-window-and-buffer (sidebar-find-file file)))
           (sidebar-message "No bookmarks for %s" file)))
        ((eq (sidebar-line-face) 'sidebar-bookmark-file-name)
         ;; ignore all options but kill and view
         (when (or (let ((overlay-list (overlays-at (point))))
                     (catch 'found
                       (dolist (overlay overlay-list)
                         (when (eq (overlay-get overlay 'face)
                                   'sidebar-bookmark-expanded-file-name)
                           (throw 'found t)))))
                   (not (sidebar-bookmark-expand))) ; side-effect
           (with-saved-window-and-buffer (sidebar-find-file file))
           (let ((window sidebar-client-window))
             (cond ((eq option 'kill) (sidebar-delete-window))
                   ((not (eq option 'view)) (sidebar-shrink-window)))
             (sidebar-select-window window))))
        ((eq (sidebar-line-face) 'sidebar-grep-name)
         ;; grep: display file for view and kill and match list otherwise
         (with-saved-window-and-buffer (sidebar-find-file file))
         (let ((window sidebar-client-window))
           (cond ((eq option 'kill)
                  (sidebar-delete-window)
                  (sidebar-select-window window))
                 ((eq option 'view))
                 (t (sidebar-match-list
                     sidebar-grep-string sidebar-grep-type
                     (selected-window) nil nil)))))
        ((eq option 'kill)
         (let ((window sidebar-client-window))
           (with-saved-window-and-buffer (sidebar-find-file file))
           (sidebar-delete-window)
           (sidebar-select-window window)))
        ((eq option 'view)
         (with-saved-window-and-buffer (sidebar-find-file file)))
        (t
         (sidebar-shrink-window)
         (sidebar-find-file file))))
      ((memq (sidebar-line-face)
	     '(sidebar-tag-name sidebar-tag-header sidebar-tag-prefix))
       (let ((from (point))
	     to overlay)
	 (cond
	  ((setq to (get-text-property from 'sidebar-property))
	   (if (eq option 'view)
	       (sidebar-view)
	     (let ((window sidebar-client-window))
	       (sidebar-shrink-window)
	       (with-saved-window-and-buffer
		(sidebar-find-file)
		(goto-char to)
		(recenter window-line))
	       (when (eq option 'kill) (sidebar-delete-window))
	       (sidebar-select-window window))))
	  ((and (setq overlay (get-text-property (point) 'overlay))
		(overlay-get overlay 'invisible))
	   (overlay-put overlay 'invisible nil)
	   (goto-char (or (progn
			    (sidebar-next-visible-line)
			    (when (get-text-property (point) 'overlay)
			      (point)))
			  from)))
	  (overlay
	   (setq overlay nil)
	   (while (and (not overlay)
		       (setq to (next-single-property-change (point) 'overlay)))
	     (goto-char to)
	     (when (setq overlay (get-text-property to 'overlay))
	       ;; likely we need a better recenter here, otherwise not sufficient
	       ;; information is visible, it would be best to see the entire prefix
	       ;; headed group here (it's simple, we have to count the number of
	       ;; lines of overlay)
	       (when (overlay-get overlay 'invisible)
		 (overlay-put overlay 'invisible nil))
	       ;; go down iff the next line has an overlay
	       (setq to (or (progn
			      (sidebar-next-visible-line)
			      (when (get-text-property (point) 'overlay)
				(point)))
			    to))))
	   ;; goto to iff overlay was found else stay at from
	   (goto-char (if overlay to from))))))
      ((eq (sidebar-line-face) 'sidebar-bookmark-name)
       ;; bookmark: display bookmark applying view and kill
       (if (eq option 'view)
           (sidebar-view)
         (let ((window sidebar-client-window)
               (bookmark-name (car (sidebar-property))))
           (sidebar-shrink-window)
           (with-saved-window-and-buffer
	    (let ((offset (sidebar-bookmark-sidebar (sidebar-property))))
	      (sidebar-find-file)
	      (bookmark-jump bookmark-name)
	      (when (numberp offset)
		(condition-case nil
		    (backward-char offset)
		  (error nil))))
	    (recenter window-line))
           (when (eq option 'kill) (sidebar-delete-window))
           (sidebar-select-window window))))
      (t
       ;; this case should not occur
       (ding))))))

(defun sidebar-mouse-right (event)
  "Emulate sidebar-open."
  (interactive "@e")
  (setq mark-active nil)
  (goto-char (posn-point (event-start event)))
  (when (and (eq sidebar-mode 'match)
	     (not (eq (window-buffer sidebar-client-window) sidebar-client-buffer)))
    ;; Assert that `sidebar-client-window' displays `sidebar-client-buffer'.
    (set-window-and-current-buffer sidebar-client-window sidebar-client-buffer))
  (sidebar-reposition)
  (sidebar-open))

;; _____________________________________________________________________________
;; 										
;;;				 flash group					
;; _____________________________________________________________________________
;; 										
(defface sidebar-flash
  '((t (:height 1.2 :inverse-video t)))
  "Face for flashing line."
  :group 'sidebar-faces)

(defun sidebar-flash-line ()
  "Flash sidebar line."
  (let ((overlay (make-overlay
		  ;; Let overlay cover newline too, looks better.
		  (line-beginning-position) (line-beginning-position 2))))
    (overlay-put overlay 'face 'sidebar-flash)
    (overlay-put overlay 'priority 10000)
    (unwind-protect
	;; We could make this customizable. 
	(sit-for 0.5)
      (delete-overlay overlay))))

(defun sidebar-flash-path ()
  "Flash path of file on this line and add it to `kill-ring'." 
  (interactive)
  (with-valid-sidebar
   (let ((window-line (sidebar-window-line))
         (file (sidebar-line-file)))
     (when file
       (kill-new file (and kill-ring (string-equal file (car kill-ring))))
       (sidebar-flash-line)
       file))))

(defun sidebar-flash-standard-path ()
  "Flash standard path of file on this line and add it to `kill-ring'." 
  (interactive)
  (with-valid-sidebar
   (let ((window-line (sidebar-window-line))
         (file (sidebar-line-file)))
     (when file
       (setq file (convert-standard-filename file))
       (kill-new file (and kill-ring (string-equal file (car kill-ring))))
       (sidebar-flash-line)
       file))))

(defun sidebar-flash-name ()
  "Flash name of file on this line and add it to `kill-ring'." 
  (interactive)
  (with-valid-sidebar
   (let ((window-line (sidebar-window-line))
         (file (sidebar-line-file)))
     (when file
       (setq file (file-name-nondirectory file))
       (kill-new file (and kill-ring (string-equal file (car kill-ring))))
       (sidebar-flash-line)
       file))))

;; _____________________________________________________________________________
;;                                                                              
;;;                     vertical scroll-bar and mouse-wheel                     
;; _____________________________________________________________________________
;;                                                                              
(defcustom sidebar-scroll-up-command 'sidebar-up
  "Vertical scroll-bar up command."
  :type '(choice (const sidebar-up)
                 (const sidebar-up-jump)
                 (const sidebar-up-preserve)
                 (const sidebar-up-drag))
  :group 'sidebar)

(defcustom sidebar-scroll-down-command 'sidebar-down
  "Vertical scroll-bar down command."
  :type  '(choice (const sidebar-down)
                  (const sidebar-down-jump)
                  (const sidebar-down-preserve)
                  (const sidebar-down-drag))
  :group 'sidebar)

(defun sidebar-vertical-scroll-bar (event)
  "Vertical-scroll-bar scroll sidebar and associated window in match lists.
Return non-nil if scrolling can be applied."
  (let* ((position (event-start event))
         (window (nth 0 position))
         (bar-part (nth 4 position))
         (selected-window (selected-window)))
    (when (with-current-buffer (window-buffer window)
            ;; scroll iff sidebar is selected
            (and (eq major-mode 'sidebar-mode)
		 (eq sidebar-mode 'match)
		 (eq selected-window window)
                 (memq bar-part '(up down))))
      (with-saved-window-and-buffer
       (sidebar-select-window window)
       ;; discard pending input: needed for mice enhanced by special software
       (discard-input)
       (cond
	((eq bar-part 'up)
	 (let ((sidebar-wrap nil))
	   (funcall sidebar-scroll-up-command)
	   (setq this-command sidebar-scroll-up-command)))
	((eq bar-part 'down)
	 (let ((sidebar-wrap nil))
	   (funcall sidebar-scroll-down-command)
	   (setq this-command sidebar-scroll-down-command))))
       t))))

(defun sidebar-mouse-wheel-scroll-line (event)
  "Mouse-wheel-scroll sidebar and associated window in match lists.
Return non-nil if scrolling can be applied."
  (interactive "e")
  (let* ((position (event-start event))
         (window (nth 0 position))
         (selected-window (selected-window)))
    (when (with-current-buffer (window-buffer window)
            ;; scroll iff sidebar is selected
	    (and (eq major-mode 'sidebar-mode)
		 (eq sidebar-mode 'match)
		 (eq selected-window window)))
      (with-saved-window-and-buffer
        (sidebar-select-window window)
        ;; discard pending input: needed for mice enhanced by special software
        (discard-input)
        (if (< (car (cdr (cdr event))) 0)
            (progn
              (funcall sidebar-scroll-up-command)
              (setq this-command sidebar-scroll-up-command))
          (funcall sidebar-scroll-down-command)
          (setq this-command sidebar-scroll-down-command))
	t))))

;; _____________________________________________________________________________
;;                                                                              
;;;                               directory group                               
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-sort-file-list-function (&optional sans)
  "Return function to sort a list of (name attributes ...) or
\(name-sans name attributes ...) lists according to current setting of
`sidebar-files-sort'."
  (cond
   ((eq sidebar-files-sort 'name) sidebar-string-car-lessp-function)
   ((eq sidebar-files-sort 'extension)
    (lambda (file1 file2)
      (let ((extension1 (file-name-extension (car file1) t))
            (extension2 (file-name-extension (car file2) t)))
        (or (sidebar-string-lessp extension1 extension2)
            (and (sidebar-string-equal extension1 extension2)
                 (sidebar-string-car-lessp file1 file2))))))
   ((eq sidebar-files-sort 'time)
    ;; by modification time, more recent first
    (if sans
        (lambda (file1 file2)
          (let ((time1 (nth 7 file1))
                (time2 (nth 7 file2)))
            (or (> (car time1) (car time2))
                (and (= (car time1) (car time2))
                     (> (cadr time1) (cadr time2))))))
      (lambda (file1 file2)
        (let ((time1 (nth 6 file1))
              (time2 (nth 6 file2)))
          (or (> (car time1) (car time2))
              (and (= (car time1) (car time2))
                   (> (cadr time1) (cadr time2))))))))
   ((eq sidebar-files-sort 'size)
    ;; by size, larger first
    (if sans
        (lambda (file1 file2)
          (> (nth 9 file1) (nth 9 file2)))
      (lambda (file1 file2)
        (> (nth 8 file1) (nth 8 file2)))))))

(defun sidebar-directory ()
  "List current directory in sidebar."
  (interactive)
  (when (sidebar-make)
    (let ((inhibit-read-only t))
      (sidebar-switch-mode 'directory)
      (sidebar-directory-list default-directory))))

(defun sidebar-directory-list (directory &optional current-file current-file-is-dir)
  "List directory DIRECTORY in sidebar.
Optional parameter CURRENT-FILE denotes (relative) filename to highlight.
CURRENT-FILE-IS-DIR when non-nil means search CURRENT-FILE is a directory.
DIRECTORY is displayed with directories first followed by files.  Directories
and files are sorted respecting the current value of
`sidebar-files-sort'."
  (let ((directory-files-and-attributes
         (directory-files-and-attributes directory nil nil t))
	(inhibit-read-only t)
	dir-list file-list delimiter)
    ;; build dir-list and file-list
    (dolist (file directory-files-and-attributes)
      (if (eq (cadr file) t)
          (unless (and sidebar-filter-apply
                       (sidebar-filter-out
                        (car file) sidebar-filter-directory-list))
            (setq dir-list (cons file dir-list)))
        (unless (and sidebar-filter-apply
                     (sidebar-filter-out (car file) sidebar-filter-file-list))
          (setq file-list (cons file file-list)))))
    ;; sort and reverse
    (if (eq sidebar-files-sort 'unsorted)
        (unless sidebar-reverse
          (setq dir-list (nreverse dir-list))
          (setq file-list (nreverse file-list)))
      (setq dir-list (sort dir-list sidebar-string-car-lessp-function))
      (setq file-list
            (sort file-list (sidebar-sort-file-list-function)))
      (when sidebar-reverse
        (setq dir-list (nreverse dir-list))
        (setq file-list (nreverse file-list))))
    (erase-buffer)
    (setq sidebar-unfolded nil)
    (setq sidebar-width sidebar-minimum-width)
    ;; process dir-list
    (dolist (dir dir-list)
      (sidebar-put-dir-line (car dir) 'sidebar-directory-name))
    ;; recalculate sidebar-width
    (setq sidebar-width
          (max (+ sidebar-width
                  (if (and dir-list sidebar-directory-brackets) 2 0))
               ;; permit display of directory in modeline 
               (+ (length (file-name-nondirectory
                           (directory-file-name directory)))
                  (if sidebar-directory-brackets 1 0))))
    ;; set delimiter
    (setq delimiter (point))
    ;; process file-list
    (dolist (file file-list)
      (sidebar-put-line (car file) 'sidebar-file-name))
    ;; move to suitable line
    (cond
     ((= (point-min) (point-max))
      ;; empty directory
      (message "Empty directory")
      (insert-char ?\  sidebar-width)
      (insert "\n")
      (sidebar-put-text 'sidebar-empty-directory)
      (goto-char (point-min)))
     (current-file
      ;; search for file
      (if current-file-is-dir
          (progn
            (goto-char (point-min))
            (re-search-forward
             (concat "^" (regexp-quote current-file) "$") delimiter t))
        (goto-char (point-max))
        (re-search-backward
         (concat "^" (regexp-quote current-file) "$") delimiter 'start)))
     ((and sidebar-file
           ;; maybe the following is insufficient
           (not (string-match "sidebar-open" (symbol-name this-command)))
           (string-equal ; sidebar-string-equal ?
            (file-truename (file-name-directory sidebar-file))
            (file-truename directory)))
      ;; current-file undefined but sidebar-file is in directory
      (goto-char delimiter)
      (re-search-forward
       (concat "^" (regexp-quote (file-name-nondirectory sidebar-file)) "$")
       nil t))
     (t (goto-char (point-min))))))

(defvar sidebar-drive-history ()
  "Sidebar drive history.  Global.")

;; _____________________________________________________________________________
;;                                                                              
;;;                               file list group                               
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-file-list-query-non-recursive ()
  "List all files matching queried expression in current directory."
  (interactive)
  (sidebar-file-list-query t))

(defun sidebar-file-list-query (&optional non-recurse)
  "List all files matching queried expression.
With optional argument NON-RECURSE non-nil don't recurse into subdirectories."
  (interactive)
  (let* ((default (car sidebar-file-list-history))
         (string
          (read-from-minibuffer
           (if default
               (format "List files matching (default `%s'): " default)
             "List files matching: ")
           nil nil nil 'sidebar-file-list-history default t)))
    (setq string (if (string-equal string "") default string))
    (if string
        (sidebar-file-list
         non-recurse nil (sidebar-filter-wildcard-to-regexp string))
      (sidebar-message "Can't list"))))

(defun sidebar-file-list (&optional non-recurse directory match-regexp)
  (interactive)
  (if (and sidebar-files-use-ls
           (memq sidebar-files-sort '(name unsorted)))
      (sidebar-file-list-ls non-recurse directory match-regexp)
    (sidebar-file-list-dfa non-recurse directory match-regexp)))

(defun sidebar-file-list-dfa (&optional non-recurse directory regexp)
  (interactive)
  (let* ((directory (or directory default-directory))
         ;; can't use regexp in directory-files since it filters out dirs too
         (dir-list (directory-files-and-attributes directory t nil t))
	 (inhibit-read-only t)
	 file-list)
    (setq sidebar-width sidebar-minimum-width)
    (message "Scanning %s ..." directory)
    (while dir-list
      (let* ((file (car dir-list))
             (file-name (car file))
             (file-name-sans (file-name-nondirectory file-name)))
        (if (eq (cadr file) t)
            ;; directory
            (if (or non-recurse
                    (string-match "^\\.\\.?$" file-name-sans)
                    (sidebar-filter-out
                     ;; filters only relative names
                     file-name-sans sidebar-filter-directory-list))
                ;; ignore file
                (setq dir-list (cdr dir-list))
              ;; recurse
              (message "Scanning %s ..." file-name)
              (setq dir-list (nconc
                              ;; depth-first
                              (directory-files-and-attributes file-name t nil t)
                              (cdr dir-list))))
          ;; file or symlink
          (unless (or (and sidebar-filter-apply
                           (sidebar-filter-out
                            file-name-sans sidebar-filter-file-list))
                      (and regexp (string-match regexp file-name-sans)))
            (setq file-list (cons (cons file-name-sans file) file-list)))
          (setq sidebar-width (max sidebar-width (length file-name-sans)))
          (setq dir-list (cdr dir-list)))))
    (if (not file-list)
        (sidebar-message "No results available")
      (if (eq sidebar-files-sort 'unsorted)
          (setq sidebar-file-list
                (if sidebar-reverse (nreverse file-list) file-list))
        (message "Sorting %s ..." directory)
        (setq sidebar-file-list
              (sort file-list (sidebar-sort-file-list-function t)))
        (when sidebar-reverse
          (setq sidebar-file-list (nreverse sidebar-file-list))))
      (erase-buffer)
      (setq sidebar-unfolded t)
      (setq sidebar-file-list-ls nil)
      (message "Inserting %s ..." directory)
      (dolist (file sidebar-file-list)
        (insert (concat (car file) "\n"))
        (sidebar-put-text 'sidebar-file-name))
      (message "Listing %s complete" directory)
      (goto-char (point-min)))))

(defun sidebar-file-list-ls (&optional non-recurse base-directory regexp)
  (interactive)
  (let* ((buffer-of-sidebar (current-buffer))
         (command "ls")
         (options (if non-recurse "-U" "-RU"))
         (base-directory
          ;;; (string-make-unibyte          ; handle umlauts in file-names
           (directory-file-name (or base-directory default-directory)));)
         (width 0)
         (filter-apply sidebar-filter-apply)
         (filter-directory-list sidebar-filter-directory-list)
         (filter-file-list sidebar-filter-file-list)
         file-list directory base-base-directory)
    (when (memq system-type '(ms-dos windows-nt))
      ;; replace slashes by backslashes
      (while (string-match "/" base-directory)
        (setq base-directory (replace-match "\\" t t base-directory)))
      (when (string-match ":\\\\\\'" base-directory)
        ;; mswindows root-directory, we have to add a backslash here
        (setq base-base-directory (concat base-directory "\\"))))
    (message "Scanning %s ..." base-directory)
    (setq sidebar-ls-buffer (get-buffer-create "*sidebar-ls*"))
    (with-current-buffer sidebar-ls-buffer (erase-buffer))
    (call-process shell-file-name nil sidebar-ls-buffer nil
                  shell-command-switch
                  (concat command " " options " "
                          (shell-quote-argument
                           (or base-base-directory
                               base-directory))))
    (message "Filtering %s ..." base-directory)
    (with-current-buffer sidebar-ls-buffer
      (setq directory base-directory)
      (goto-char (point-min))
      (while (not (eobp))
        (cond
         ((looking-at "\\(.*\\)=:$")
          (setq directory (match-string 1))
          (when (memq system-type '(ms-dos windows-nt))
            ;; replace slashes by backslashes
            (while (string-match "/" directory)
              (setq directory (replace-match "\\" t t directory))))
          (forward-line)
          (when (and filter-apply
                     (sidebar-filter-out
                      (file-name-nondirectory directory) filter-directory-list))
            (while (and (not (eobp))
                        (not (looking-at "\\(.*\\)=:$")))
              (forward-line))))
         ((not (looking-at "[ \t]*$"))
          (let* ((file-name-sans
                  (buffer-substring (line-beginning-position) (line-end-position)))
                 (file (concat
                        directory
                        (if (memq system-type '(ms-dos windows-nt))
                            "\\" "/" )
                        file-name-sans)))
            (when (and (or (file-symlink-p file)
                           (not (file-directory-p file)))
                       (or (not filter-apply)
                           (not (sidebar-filter-out
                                 file-name-sans filter-file-list)))
                       (or (not regexp)
                           (string-match regexp file-name-sans)))
              (setq file-list (cons (cons file-name-sans file) file-list))
              (setq width (max width (length file-name-sans))))
            (forward-line)))
         (t (forward-line)))))
    (with-current-buffer buffer-of-sidebar
      (let ((inhibit-read-only t))
        (if (not file-list)
            (sidebar-message "No results available ...")
          (cond
           ((eq sidebar-files-sort 'name)
            (message "Sorting %s ..." base-directory)
            (setq file-list (sort file-list sidebar-string-car-lessp-function))
            (when sidebar-reverse
              (setq file-list (nreverse file-list))))
           ((and (eq sidebar-files-sort 'unsorted)
                 (not sidebar-reverse))
            (setq file-list (nreverse file-list))))
          (erase-buffer)
          (setq sidebar-unfolded t)
          (setq sidebar-file-list-ls t)
          (setq sidebar-width (max width sidebar-minimum-width))
          (setq sidebar-file-list nil)
          (message "Inserting %s ..." base-directory)
          (dolist (file file-list)
            (let ((name (car file)))
              (setq sidebar-file-list
                    (cons (cdr file) sidebar-file-list))
              (insert (concat name "\n"))))
          (put-text-property (point-min) (point-max) 'face 'sidebar-file-name)
          (setq sidebar-file-list (nreverse sidebar-file-list))
          (message "Listing %s complete" base-directory)
          (goto-char (point-min))
          (sidebar-post-command))))))

;; _____________________________________________________________________________
;;                                                                              
;;;                                  tag group                                  
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-tag-or-kill (&optional arg)
  "List tags in sidebar.  With prefix argument kill sidebar.
Compare `sidebar-tag-list' and `sidebar-delete-sidebar'."
  (interactive "P")
  (if arg (sidebar-delete-sidebar) (sidebar-tag-list)))

(defun sidebar-tag-list-get ()
  "Get tags for current buffer."
  (require 'local-tags "~/site-lisp/local-tags.el")
;;;   (require 'imenu)
  ;; the following `or' is utterly needed when switching from match-mode to
  ;; tag-mode since otherwise markers will point to a deleted buffer (we call
  ;; sidebar-tag-list-get _before_ reverting to the base buffer since there might
  ;; no be any tags and we want to preserve the previous state in that case)
  (condition-case nil
      (save-excursion
	(save-restriction
	  (widen)
	  (let ((local-tags-etags-charpos t))
	    (cdr (local-tags-list)))))
    (error nil)))

(defun sidebar-tag-pop (pdl)
  (let* ((first (car pdl))
	 (next (cadr pdl))
	 (rest (cddr pdl)))
    (cons
     (cons
      ;; cdr non-nil retain
      (car next)
      (if (>= (length (cdr first)) sidebar-tag-group-minimum-size)
	  ;; create separate subgroup here
	  (cons first (cdr next))
	;; merge into the next group
	(append (cdr first) (cdr next))))
     rest)))

(defun sidebar-tag-list-make (tags)
  (let ((tags (nreverse tags)) ;; eventually tags should arrive reversed
	(pdl (list (list "")))
	name name-length prefix prefix-length last index pdls)
    (dolist (tag tags)
      (cond
       ((number-or-marker-p (cdr tag))
	;; normal tag, process
	(setq name (car tag))
	(setq name-length (length name))
	(while (and (setq prefix (caar pdl))
		    (setq prefix-length (length prefix))
		    (or (>= prefix-length name-length)
			(not (string-equal
			      (substring name 0 prefix-length)
			      prefix))))
	  ;; prefix is not a prefix of name, pop prefix
	  (setq pdl (sidebar-tag-pop pdl)))
	(setq last nil)
	(setq index (1+ prefix-length))
	(while (and (< index name-length)
		    (string-match "\\(\\s_+\\)\\w+" name index))
	  (when last			; push prefix only
	    (setq pdl (cons (list last) pdl)))
	  (setq last (substring name 0 (match-end 1)))
	  (setq index (match-end 0)))
	(setq pdl
	      (if last
		  ;; push prefix plus tag
		  (cons (list last tag) pdl)
		;; merge tag into pdl
		(cons (cons (caar pdl) (cons tag (cdar pdl))) (cdr pdl)))))
       ((listp (cdr tag))
	;; group, recurse
	(setq pdls
	      (cons
	       (cons (car tag)
		     (list
		      (sidebar-tag-list-make
		       (sort (cdr tag) sidebar-string-car-lessp-function))))
	       pdls)))))
    ;; pop any prefixes
    (while (> (length pdl) 1)
      (setq pdl (sidebar-tag-pop pdl)))
    ;; combine groups and pdl
    (if pdls (append pdls pdl) (car pdl))))

(defun sidebar-tag-put (name length indent &optional properties)
  "Insert TEXT."
  (goto-char (point-max))
  (let ((from (point)))
    (insert (concat (make-string indent ?\ ) (substring name length) "\n"))
    ;; <-------------------------------------------- ????
    (setq sidebar-width (max sidebar-width (- (point) from 1)))
    ;; it would be nice to do this with propertize but apparently that's
    ;; impossible
    (when properties
      ;; add properties from beginning to end of line
      (add-text-properties
       from (1- (point))
       (append properties `(mouse-face highlight help-echo ,name))))))

(defun sidebar-tag-list ()
  "List tags in sidebar.
Display message if no tags are available."
  (interactive)
  (let ((selected-window (selected-window)))
    (when (sidebar-make t)
      (let ((truncate-partial-width-windows t)
	    (inhibit-read-only t)
	    tags)
        (cond
         ((and sidebar-client-buffer
	       (with-current-buffer sidebar-client-buffer
		 (setq tags (sidebar-tag-list-get))))
	  (setq sidebar-window-point (window-point sidebar-client-window))
	  (setq sidebar-tag-point 1)
	  (setq sidebar-tag-best-point 1)
	  ;; width non-nil means we were not necessarily in tag-mode before
	  (sidebar-switch-mode 'tag)
	  (setq sidebar-width 0)
	  (sidebar-tag-list-1
	   ;; make this depend on whether we should sort
	   (sidebar-tag-list-make
	    (sort (copy-sequence tags) sidebar-string-car-lessp-function))
	   0 0)
	  ;; the following was needed to make the last invisible overlay work,
	  ;; hopefully it isn't any more
	  (let ((overlays (overlays-at sidebar-tag-point)))
	    (dolist (overlay overlays)
	      (when (and (overlay-get overlay 'sidebar-tag)
			 (overlay-get overlay 'invisible))
		(overlay-put overlay 'invisible nil)))
	    (goto-char sidebar-tag-point)
	    (with-saved-window-and-buffer (sidebar-find-file))))
         (t
          (sidebar-message "No tags available")
          (sidebar-kill-or-restore)
	  (select-window selected-window)))))))

(defun sidebar-tag-list-1 (tags length indent)
  (overlay-recenter (point))
  (if (listp (car tags))
      (dolist (tag tags)
	(if (string-equal (car tag) "")
	    (dolist (tag (cdr tag))
	      (sidebar-tag-list-1 tag 0 indent))
	  (let ((from (point-max))
		to)
	    (sidebar-tag-put
	     (car tag) length indent '(face sidebar-tag-header))
	    (setq to (point-max))
	    (sidebar-tag-list-1 (cadr tag) 0 (1+ indent)) ; customize the 1+
	    (let ((overlay (make-overlay to (point-max))))
	      (overlay-put overlay 'invisible t)
	      (overlay-put overlay 'sidebar-tag t)
	      (put-text-property from to 'overlay overlay)))))
    (let ((name (car tags))
	  (prop (cdr tags)))
      (cond
       ((string-equal name "")
	(dolist (tag prop)
	  (sidebar-tag-list-1 tag 0 indent)))
       ((number-or-marker-p prop)
	(when (and (> prop sidebar-tag-best-point)
		   (<= prop sidebar-window-point))
	  ;; we have a new optimum
	  (setq sidebar-tag-best-point prop)
	  (setq sidebar-tag-point (point)))
	(sidebar-tag-put
	 name length indent `(face sidebar-tag-name sidebar-property ,prop)))
       ((listp prop)
	(let ((from (point-max))
	      to)
	  (sidebar-tag-put
	   name length indent '(face sidebar-tag-prefix))
	  (setq to (point-max))
	  (dolist (tag prop)
	    (sidebar-tag-list-1 tag (length name) (1+ indent))) ; customize 1+
	  (let ((overlay (make-overlay to (point-max))))
	    (overlay-put overlay 'invisible t)
	    (overlay-put overlay 'sidebar-tag t)
	    (put-text-property from to 'overlay overlay))))))))

(defun sidebar-re-search-forward (regexp &optional from to)
  "Like `re-search-forward' but skip invisible text.
Search starts at FROM, if present, or at `point' and terminates at TO. 
Don't report an error (just return nil if no occurrence is found) and do
not search for successive occurrences.  Does not handle empty matches."
  (let ((to (or to (point-max))))
    (when from (goto-char from))
    (catch 'end
      (while (< (point) to)
	(if (re-search-forward regexp to t)
	    (if (get-char-property (1- (point)) 'invisible)
		;; match in invisible text, skip invisible text and search again
		(goto-char (next-single-char-property-change
			    (1- (point)) 'invisible))
	      ;; match in visible text, throw
	      (throw 'end t))
	  ;; no match, throw
	  (throw 'end nil))))))

(defun sidebar-jump ()
  "Add character of `last-command-event' to or use it to replace
`sidebar-jump-string' and jump to corresponding line in sidebar \(if found).
Character is added iff `sidebar-jump-time' has not elapsed.  Searching starts at
current line if character was added and at next line if `sidebar-jump-string'
was replaced.  Original `sidebar-jump-string' is retained if no corresponding
line was found."
  (interactive)
  (let* ((current-point (point))
	 (current-time (current-time))
         (current-time-float (+ (float (nth 1 current-time))
				(/ (float (nth 2 current-time)) 1000000.0)))
	 (accumulate
	  (< (- current-time-float sidebar-jump-time) sidebar-jump-delay))
	 (string (concat (if accumulate sidebar-jump-string "")
			 (char-to-string last-command-event)))
	 (case-fold-search (string-equal (downcase string) string))
	 (regexp (concat "^\\s-*" (regexp-quote string)))
	 (start (line-beginning-position (unless accumulate 2))))
    ;; search for string preceded by whitespace at bol
    (if (or
	 ;; search from start until end of buffer
	 (sidebar-re-search-forward regexp start)
	 ;; search from beginning of buffer until start
	 ;; line containing start is investigated last
	 (sidebar-re-search-forward regexp (point-min) start)
	 ;; if we accumulated and string is made up of the same characters,
	 ;; search again for string starting with this character (this should
	 ;; handle the case where the user searches for a string starting with
	 ;; "c" and sidebar-jump-string was "cc")
	 (and accumulate
	      (string-equal string
			    (make-string
			     (length string)
			     (string-to-char (substring string 0 1))))
	      (let ((start (save-excursion
			     (goto-char start)
			     (line-beginning-position 2)))
		    (regexp
		     (concat "^\\s-*" (regexp-quote (substring string 0 1)))))
		(or (and (sidebar-re-search-forward regexp start)
			 ;; search from start until end of buffer
			 (setq string (substring string 0 1)))
		    (and (sidebar-re-search-forward regexp (point-min) start)
			  ;; search from beginning of buffer until seach-start
			 (setq string (substring string 0 1)))))))
        (progn
          (goto-char (match-beginning 0))
          (beginning-of-line)
          (unless (pos-visible-in-window-p) (recenter 3))
          (setq sidebar-jump-string string))
;;;           (cond
;;;            ((eq (sidebar-line-face) 'sidebar-tag-name)
;;;             (sidebar-view))))	  ; why is this needed (and not for bookmarks ?)
      (ding)
      (goto-char current-point))
    (setq sidebar-jump-time current-time-float)))

;; _____________________________________________________________________________
;;                                                                              
;;;                               bookmark group                                
;; _____________________________________________________________________________
;;                                                                              
(defsubst sidebar-bookmark-file-name (bookmark)
  "Return file of BOOKMARK, or nil if none."
  (file-truename (cdr (assq 'filename (cadr bookmark)))))

(defsubst sidebar-bookmark-position (bookmark)
  "Return position of BOOKMARK, or nil if none."
  (cdr (assq 'position (cadr bookmark))))

(defsubst sidebar-bookmark-sidebar (bookmark)
  "Return sidebar status of BOOKMARK, or nil if none."
  (cdr (assq 'sidebar (cadr bookmark))))

(defsubst sidebar-bookmark-file-p (bookmark)
  "Return t if sidebar status of BOOKMARK equals t."
  (eq (cdr (assq 'sidebar (cadr bookmark))) t))

(defsubst sidebar-bookmark-overlay (from to face)
  (let ((overlay (make-overlay from to)))
    (overlay-put overlay 'face face)
    (overlay-put overlay 'priority 700)
    (overlay-put overlay 'evaporate t)))

(defun sidebar-bookmark-load ()
  "Assert existence of bookmark components.
Require bookmark, load `bookmark-default-file', set
`sidebar-bookmark-default-file-modtime' and
`sidebar-bookmark-alist-modification-count'."
  (require 'bookmark)
  (bookmark-maybe-load-default-file)
  (when (and (not sidebar-bookmark-default-file-modtime)
             (file-exists-p bookmark-default-file))
    (setq sidebar-bookmark-default-file-modtime
          (nth 5 (file-attributes bookmark-default-file))))
  (unless sidebar-bookmark-alist-modification-count
    (setq sidebar-bookmark-alist-modification-count
          bookmark-alist-modification-count)))

(defun sidebar-bookmarks ()
  "List bookmarks in sidebar."
  (interactive)
  (sidebar-bookmark-load)
  (let ((selected-window (selected-window)))
    (when (sidebar-make t)
      (cond
       ((and sidebar-file
             (sidebar-bookmark-list (file-truename sidebar-file))))
       ((and (not sidebar-file) (sidebar-bookmark-list)))
       (t
        (sidebar-message "No bookmarks available")
        (sidebar-kill-or-restore)
	(select-window selected-window))))))

(defun sidebar-bookmark-index (item list)
  "Return position of first element of LIST whose cdr equals ITEM.
Return nil if no such element exists."
  (let ((index 0))
    (catch 'found
      (dolist (element list)
        (if (equal (cdr element) item)
            (throw 'found index)
          (setq index (1+ index)))))))

(defun sidebar-bookmark-list (&optional current-file)
  "List bookmarked directories, files, and bookmarks of current file in sidebar.
Directories are listed first.  Files that can be expanded because they
contain bookmarks are shown with `sidebar-bookmark-expandable-file-name'.
Expanded files are shown with `sidebar-bookmark-expanded-file-name'.  If
CURRENT-FILE is specified try to expand its bookmarks.  Otherwise, try
to expand bookmarks of `sidebar-file'."
  (sidebar-bookmark-load)
  (let* ((old-bookmark
          (when (or (eq (sidebar-line-face) 'sidebar-bookmark-name)
                    sidebar-unfolded)
            (sidebar-property)))
         (old-name (when old-bookmark (car old-bookmark)))
	 (inhibit-read-only t)
	 dir-list file-list delimiter index)
    (bookmark-maybe-load-default-file)
    ;; construct dir-list and file-list
    (dolist (bookmark bookmark-alist)
      (let* ((file (sidebar-bookmark-file-name bookmark))
             (name (file-name-nondirectory file)))
        (cond
         ((file-exists-p file)
          ;; file exists
          (if (file-directory-p file)
              (unless (rassoc file dir-list)
                (setq dir-list (cons (cons name file) dir-list)))
            (let ((element (rassoc file file-list)))
              (if element
                  ;; file already in file-list, update expandable status
                  (when (and (not (cdar element))
                             (not (sidebar-bookmark-file-p bookmark)))
                    (setcar element (cons name t)))
                ;; file not yet in file-list
                (setq file-list
                      (cons (cons (cons name
                                        (not (sidebar-bookmark-file-p
                                              bookmark)))
                                  file)
                            file-list))))))
          ;; file does not exist, remove bookmark iff it has sidebar status
         ((sidebar-bookmark-sidebar bookmark)
          (setq bookmark-alist (delq bookmark bookmark-alist))))))
    (when (or dir-list file-list)
      (let ((sort-function
             (cond
              ((eq sidebar-files-sort 'name)
               sidebar-string-caar-lessp-function)
              ((eq sidebar-files-sort 'extension)
               ;; should we care about anything else ?
               (lambda (file1 file2)
                 ;; use period to avoid value nil
                 (let ((extension1 (file-name-extension (cdr file1) t))
                       (extension2 (file-name-extension (cdr file2) t)))
                   (or (sidebar-string-lessp extension1 extension2)
                       (and (sidebar-string-equal extension1 extension2)
                            (sidebar-string-lessp (caar file1) (caar file2)))))))
              ((eq sidebar-files-sort 'time)
               ;; by time, more recent first
               (lambda (file1 file2)
                 (let ((time1 (nth 5 (file-attributes (cdr file1))))
                       (time2 (nth 5 (file-attributes (cdr file2)))))
                   (or (> (car time1) (car time2))
                       (and (= (car time1) (car time2))
                            (> (cadr time1) (cadr time2)))))))
              ((eq sidebar-files-sort 'size)
               ;; by size, larger first
               (lambda (file1 file2)
                 (> (nth 7 (file-attributes (cdr file1)))
                    (nth 7 (file-attributes (cdr file2)))))))))
        ;; sort and reverse
        (if (eq sidebar-files-sort 'unsorted)
            (unless sidebar-reverse
              (setq dir-list (nreverse dir-list))
              (setq file-list (nreverse file-list)))
          (setq dir-list (sort dir-list sidebar-string-car-lessp-function))
          (setq file-list (sort file-list sort-function))
          (when sidebar-reverse
            (setq dir-list (nreverse dir-list))
            (setq file-list (nreverse file-list)))))
      ;; prepare buffer
      (sidebar-switch-mode 'bookmark)
      (setq sidebar-width sidebar-minimum-width)
      ;; process dir-list
      (dolist (dir dir-list)
        (sidebar-put-dir-line
         (car dir) 'sidebar-bookmark-directory-name (cdr dir)))
      ;; recalculate sidebar-width
      (when (and dir-list sidebar-directory-brackets)
        (setq sidebar-width (+ sidebar-width 2)))
      ;; set delimiter
      (setq delimiter (point))
      ;; process file-list
      (dolist (file file-list)
        (sidebar-put-line
         (caar file) 'sidebar-bookmark-file-name (cdr file))
        (when (cdar file)
          (sidebar-bookmark-overlay
           (line-beginning-position 0) (line-beginning-position 2)
           'sidebar-bookmark-expandable-file-name)))
      (goto-char (point-min))
      ;; I hate true names
      (when current-file
        (setq current-file (file-truename current-file)))
      (cond
       ((and current-file
             (file-directory-p current-file)
             (setq index (sidebar-bookmark-index current-file dir-list)))
        (forward-line index))
       ((and current-file
             (setq index (sidebar-bookmark-index current-file file-list)))
        (goto-char delimiter)
        (forward-line index))
       ((and sidebar-file
             ;; not tested yet
             (not (string-match "sidebar-open" (symbol-name this-command)))
             (setq index (sidebar-bookmark-index sidebar-file file-list)))
        ;; current-file undefined but sidebar-file is in directory ???
        (goto-char delimiter)
        (forward-line index))
       (t (goto-char (point-min))))
      (beginning-of-line)
      (setq sidebar-unfolded nil)
      (unless (file-directory-p (sidebar-property))
        ;; try to list bookmarks
        (sidebar-bookmark-expand old-name old-bookmark))
      (when (file-exists-p bookmark-default-file)
        (setq sidebar-bookmark-default-file-modtime
              (nth 5 (file-attributes bookmark-default-file))))
      (setq sidebar-bookmark-alist-modification-count
            bookmark-alist-modification-count))))

(defun sidebar-bookmark-file-list (file)
  "Return list of bookmarks for file FILE."
  (sidebar-bookmark-load)
  (let (bookmark-file-list)
    (dolist (bookmark bookmark-alist bookmark-file-list)
      (when (and (sidebar-same-files
                  file (sidebar-bookmark-file-name bookmark))
                 (not (sidebar-bookmark-file-p bookmark)))
          (setq bookmark-file-list
                (cons (cons (car bookmark) (cdr bookmark)) bookmark-file-list))))))

(defun sidebar-bookmark-collapse (&optional check)
  (sidebar-bookmark-load)
  (when (eq (sidebar-line-face) 'sidebar-bookmark-file-name)
    (let ((file (sidebar-property))
          (at (point))
          (at-eol (line-end-position))
          (at-two (line-beginning-position 2))
          to)
      (while (and (zerop (forward-line))
                  (eq (sidebar-line-face) 'sidebar-bookmark-name)))
      (setq to (line-beginning-position))
      (delete-region at-two to)
      (when check
        (let ((overlay-list (overlays-at at)))
          (unless
              (catch 'found
                (dolist (overlay overlay-list)
                  (when (eq (overlay-get overlay 'face)
                            'sidebar-bookmark-expanded-file-name)
                    (if (sidebar-bookmark-file-list file)
                        (overlay-put
                         overlay 'face
                         'sidebar-bookmark-expandable-file-name)
                      (delete-overlay overlay))
                    (throw 'found t))))
            (when (sidebar-bookmark-file-list file)
              (sidebar-bookmark-overlay
               at at-two 'sidebar-bookmark-expandable-file-name)))))
      (goto-char at))))

(defun sidebar-bookmark-expand (&optional old-name old-bookmark)
  (sidebar-bookmark-load)
  (sidebar-bookmark-collapse)
  (let ((bookmark-file-list (sidebar-bookmark-file-list (sidebar-property)))
        (at (point))
        (at-eol (line-end-position))
        (at-two (line-beginning-position 2))
	(inhibit-read-only t)
	new-point)
    ;; sort bookmarks if requested
    (when bookmark-file-list
      (unless (eq sidebar-bookmarks-sort 'time)
        (setq bookmark-file-list
              (sort bookmark-file-list
                    (cond
                     ;; t is by creation time
                     ((eq sidebar-bookmarks-sort 'name)
                      'sidebar-string-car-lessp)
                     ((eq sidebar-bookmarks-sort 'position)
                      '(lambda (bookmark1 bookmark2)
                         (< (sidebar-bookmark-position bookmark1)
                            (sidebar-bookmark-position bookmark2))))))))
      ;; reverse bookmarks if requested
      (when sidebar-bookmarks-reverse
        (setq bookmark-file-list (nreverse bookmark-file-list)))
      ;; write out bookmarks
      (forward-line)
      (dolist (bookmark bookmark-file-list)
        (when (and (not new-point)
                   old-name (string-equal old-name (car bookmark))
                   (equal bookmark old-bookmark))
          (setq new-point (point)))
        (sidebar-put-line
         (concat (make-string sidebar-indent ?\ ) (car bookmark))
         'sidebar-bookmark-name bookmark))
       ;; override window-overlay
      (sidebar-bookmark-overlay at-two (point) 'sidebar-bookmark-name)
      ;; set this to avoid rescans in post-command-hook
      (setq sidebar-bookmark-alist-modification-count
            bookmark-alist-modification-count)
      ;; stay at file line
      (goto-char at)
      (let ((overlay-list (overlays-at at)))
        (unless (catch 'found
                  (dolist (overlay overlay-list)
                    (when (eq (overlay-get overlay 'face)
                              'sidebar-bookmark-expandable-file-name)
                      (overlay-put overlay 'face
                                   'sidebar-bookmark-expanded-file-name)
                      (throw 'found t))))
          (sidebar-bookmark-overlay
           at at-eol 'sidebar-bookmark-expanded-file-name)))
      (when new-point (goto-char new-point)))
    bookmark-file-list))

(defun sidebar-bookmark-named-file (file)
  "Non-nil when FILE has at least one named bookmark."
  (sidebar-bookmark-load)
  (setq file (file-truename file))
  (catch 'found
    (dolist (bookmark bookmark-alist)
      (when (and (string-equal
                  file (file-truename (sidebar-bookmark-file-name bookmark)))
                 (not (sidebar-bookmark-file-p bookmark)))
        (throw 'found t)))))

(defun sidebar-bookmark-file (file &optional suppress)
  "Bookmark file FILE."
  (sidebar-bookmark-load)
  (setq file (file-truename file))
  (if (catch 'found
        (dolist (bookmark bookmark-alist)
          (when (string-equal file (file-truename
                                    (sidebar-bookmark-file-name bookmark)))
            (throw 'found t))))
      (unless suppress
        (sidebar-message "File %s already bookmarked" file))
    (let* ((name (file-name-nondirectory file))
           (bookmark
            (list (concat name)
                  (list (cons 'filename file)
                        (cons 'front-context-string nil)
                        (cons 'rear-context-string nil)
                        (cons 'position 0)
                        (cons 'sidebar t)))))
      ;; append <2> ...
      (bookmark-maybe-rename bookmark (bookmark-all-names))
      (setq bookmark-alist (nconc (list bookmark) bookmark-alist))
      (setq bookmark-alist-modification-count
            (1+ bookmark-alist-modification-count))
      (when (bookmark-time-to-save-p) (bookmark-save)))))

(defun sidebar-bookmark-add ()
  "Add current sidebar entry to bookmarks."
  (interactive)
  (with-valid-sidebar
   (cond
    ((memq (sidebar-line-face)
           '(sidebar-directory-name sidebar-file-name sidebar-grep-name))
     (sidebar-bookmark-file (sidebar-line-file)))
    ((eq (sidebar-line-face) 'sidebar-tag-name)
     (let ((name (car (sidebar-property)))
           (from (cdr (sidebar-property))))
       (with-current-buffer sidebar-client-buffer
          (goto-char from)
          (beginning-of-line)
          (when (string-match "^[ \t]*" name)
            ;; remove leading spaces from name
            (setq name (substring name (match-end 0))))
          (if (re-search-forward (regexp-quote name) (line-end-position) t)
              ;; bookmark what the tag says
              (sidebar-bookmark-set
               name (match-beginning 0) (match-end 0))
            (sidebar-bookmark-set
             name from (save-excursion
                         (goto-char from) (forward-sexp) (point)))))))
    ((and (eq (sidebar-line-face) 'sidebar-match) sidebar-file)
     (let* ((from (overlay-start (sidebar-property)))
            (to (overlay-end (sidebar-property)))
            (name (with-current-buffer (buffer-base-buffer sidebar-client-buffer)
                    (buffer-substring-no-properties from to))))
       (with-current-buffer (buffer-base-buffer sidebar-client-buffer)
         (sidebar-bookmark-set name from to))))
    ((memq (sidebar-line-face)
           '(sidebar-buffer-directory-name sidebar-buffer-name))
     (let ((file (sidebar-line-file)))
       (if file
           (sidebar-bookmark-file file)
         (sidebar-message "Buffer %s not associated with file"
                            (sidebar-property)))))
    (t (sidebar-message "Nothing to add here")))))

(defun sidebar-bookmark-make-cell (annotation &optional info-node)
  "Return the record part of a new bookmark, given ANNOTATION.
Must be at the correct position in the buffer in which the bookmark is
being set.  This might change someday.
Optional second arg INFO-NODE means this bookmark is at info node
INFO-NODE, so record this fact in the bookmark's entry."
  (let ((the-record
         `((filename . ,(bookmark-buffer-file-name))
           (front-context-string
            . ,(if (>= (- (point-max) (point)) bookmark-search-size)
                   (buffer-substring-no-properties
                    (point)
                    (+ (point) bookmark-search-size))
                   nil))
           (rear-context-string
            . ,(if (>= (- (point) (point-min)) bookmark-search-size)
                   (buffer-substring-no-properties
                    (point)
                    (- (point) bookmark-search-size))
                   nil))
           (position . ,(point)))))

    ;; Now fill in the optional parts:

    ;; Take no chances with text properties
    (set-text-properties 0 (length annotation) nil annotation)
    (set-text-properties 0 (length info-node) nil info-node)

    (if annotation
        (nconc the-record (list (cons 'annotation annotation))))
    (if info-node
        (nconc the-record (list (cons 'info-node info-node))))

    ;; Finally, return the completed record.
    the-record))

(defun sidebar-bookmark-set (&optional name from to)
  "Set bookmark."
  (interactive)
  (sidebar-bookmark-load)
  (save-excursion
    (if (and (eq major-mode 'sidebar-mode)
	     (not (eq sidebar-mode 'bookmark)))
	(sidebar-bookmark-add)
      (when (eq major-mode 'sidebar-mode)
	;; This might be disconcerting, let's see.
	(set-buffer (window-buffer sidebar-client-window)))
      (cond
       ((and (not buffer-file-name) (not (buffer-base-buffer)))
	(sidebar-message "Buffer \"%s\" has no associated file" (buffer-name)))
       (name)					; name specified
       ((and mark-active
	     (setq from (save-excursion
			  (goto-char (min (point) (mark)))
			  (skip-syntax-forward " ")
			  (point)))
	     (setq to (min (save-excursion (goto-char from) (line-end-position))
			   (max (point) (mark))))
	     (< from to))
	(setq name (buffer-substring-no-properties from to)))
       ((let ((bounds (sidebar-name-around-point)))
	  (and bounds
	       (setq from (car bounds))
	       (setq to (cdr bounds))
	       (setq name (buffer-substring-no-properties from to)))))
       ;; *** TODO: Use something else here before giving up.  ***
       (t (sidebar-message "No suitable string around point")))
      (when name
	(when (buffer-base-buffer)
	  ;; for indirect buffer bookmark buffer-base-buffer instead, retain this,
	  ;; could be useful
	  (set-buffer (buffer-base-buffer)))
	(let* ((offset (/ (* 2 (- to from)) 3))
	       (bookmark
		(save-excursion
		  (goto-char (+ from offset))
		  (list name
			(nconc (sidebar-bookmark-make-cell nil)
			       (list (cons 'sidebar offset)))))))
	  (when (member bookmark bookmark-alist)
	    ;; avoid duplicates of same bookmark
	    (setq bookmark-alist (delete bookmark bookmark-alist)))
	  (bookmark-maybe-rename bookmark (bookmark-all-names))
	  (setq bookmark-alist (nconc (list bookmark) bookmark-alist))
	  (setq bookmark-alist-modification-count
		(1+ bookmark-alist-modification-count))
	  (when (bookmark-time-to-save-p) (bookmark-save))
	  (let ((overlay (make-overlay from to)))
	    (deactivate-mark)
	    (overlay-put overlay 'face 'sidebar-bookmark-name)
	    (sit-for 0.5)
	    (delete-overlay overlay)))))))

;; *** TODO: The following two should update and save bookmarks immediately ***
(defun sidebar-bookmark-find-file ()
  "Possibly add name of file found to bookmarks.
File is added iff `sidebar-bookmarks-update' equals `when-find'."
  (when (and buffer-file-name
             (eq sidebar-bookmarks-update 'when-find)) 
    (sidebar-bookmark-file buffer-file-name t))
  nil)

(defun sidebar-bookmark-write-file ()
  "Possibly add name of file written to bookmarks.
File is added iff `sidebar-bookmarks-update' equals `when-find' or
`when-write.'"
  ;; OBSERVE: THIS IS FLUSHED WHEN FILE_NAME CHANGES, RECONSIDER ????????????
  (when (and buffer-file-name
	     (not (member buffer-file-name sidebar-bookmarks-norecord))
	     (not (member (file-name-directory buffer-file-name)
			  sidebar-bookmarks-norecord))
	     (memq sidebar-bookmarks-update '(when-find when-write)))
    (sidebar-bookmark-file buffer-file-name t))
  nil)

(defun sidebar-bookmark-unfold ()
  (interactive)
  (let* ((old-bookmark
          (when (or (eq (sidebar-line-face) 'sidebar-bookmark-name)
                    sidebar-unfolded)
            (sidebar-property)))
         (old-name (when old-bookmark (car old-bookmark)))
	 (inhibit-read-only t)
	 bookmark-list new-point)
    (sidebar-bookmark-load)
    (dolist (bookmark bookmark-alist)
      (unless (sidebar-bookmark-file-p bookmark)
        (setq bookmark-list
              (cons (cons (car bookmark) (cdr bookmark)) bookmark-list))))
    (when bookmark-list
      (if (eq sidebar-bookmarks-sort 'name)
          (progn
            (setq bookmark-list
                  (sort bookmark-list 'sidebar-string-car-lessp))
            (when sidebar-bookmarks-reverse
              (setq bookmark-list (nreverse bookmark-list))))
        (unless sidebar-bookmarks-reverse
          (setq bookmark-list (nreverse bookmark-list))))
      ;; write out bookmarks
      (erase-buffer)
      (dolist (bookmark bookmark-list)
        (when (and (not new-point)
                   old-name (string-equal old-name (car bookmark))
                   (equal bookmark old-bookmark))
          (setq new-point (point)))
        (sidebar-put-line
         (car bookmark) 'sidebar-bookmark-name bookmark))
      ;; override window-overlay
      (sidebar-bookmark-overlay
       (point-min) (point-max) 'sidebar-bookmark-name)
      ;; set this to avoid rescans in post-command-hook
      (setq sidebar-bookmark-alist-modification-count
            bookmark-alist-modification-count)
      (if new-point (goto-char new-point) (goto-char (point-min)))
      (setq sidebar-unfolded t))))

;; _____________________________________________________________________________
;;                                                                              
;;;				Buffers group					
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-buffers (&optional buffer-list)
  "Display list of buffers in sidebar."
  (interactive)
  (when (or buffer-list
	    (and (sidebar-make) (setq buffer-list (buffer-list))))
    (let (name directories non-directories asterisks spaces others)
      (dolist (buffer buffer-list)
	(setq name (buffer-name buffer))
	(cond
	 ((string-equal (substring name 0 1) " ")
	  ;; Buffer name starts with space.
	  (unless sidebar-buffer-ignore-space
	    (setq spaces (cons (cons name buffer) spaces))))
	 ((string-equal (substring name 0 1) "*")
	  ;; Buffer name starts with asterisk.
	  (unless sidebar-buffer-ignore-asterisk
	    (setq asterisks (cons (cons name buffer) asterisks))))
	 ((with-current-buffer buffer (eq major-mode 'dired-mode))
	  ;; Directory.
	  (setq directories (cons (cons name buffer) directories)))
	 (t
	  ;; Non-directory.
	  (setq non-directories (cons (cons name buffer) non-directories)))))
      ;; *** TODO: We should use mapping functions here ***
      (if sidebar-buffers-sort
	  ;; Sort buffers.
	  (let ((fun sidebar-string-car-lessp-function))
	    (setq directories (sort directories fun))
	    (setq non-directories (sort non-directories fun))
	    (setq spaces (sort spaces fun))
	    (setq asterisks (sort asterisks fun)))
	;; Reverse buffers.
	(setq directories (nreverse directories))
	(setq non-directories (nreverse non-directories))
	(setq spaces (nreverse spaces))
	(setq asterisks (nreverse asterisks)))
      (setq others (nconc asterisks spaces))
      (if (or directories non-directories others)
	  (let ((inhibit-read-only t))
	    (sidebar-switch-mode 'buffer)
	    (setq sidebar-width sidebar-minimum-width)
	    (dolist (buffer directories)
	      (sidebar-put-dir-line
	       (car buffer) 'sidebar-buffer-directory-name (cdr buffer)))
	    (when (and directories sidebar-directory-brackets)
	      (setq sidebar-width (+ sidebar-width 2)))
	    (dolist (buffer non-directories)
	      (sidebar-put-line (car buffer) 'sidebar-buffer-name (cdr buffer))
	      (when (buffer-modified-p (cdr buffer))
		(let ((overlay
		       (make-overlay
			(line-beginning-position 0)
			;; *** TODO: Outrageous that min stuff !!! ***
			(min (line-end-position 0)
			     (+ (line-beginning-position 0)
				(length (car buffer))))
			nil t)))
		  ;; *** TODO: evaporate ?? ***
		  (overlay-put overlay 'evaporate t)
		  (overlay-put overlay 'face 'sidebar-buffer-modified))))
	    (dolist (buffer others)
	      ;; *** TODO: Modified likely distracting here ??? ***
	      (sidebar-put-line (car buffer) 'sidebar-buffer-name (cdr buffer)))
	    (goto-char (point-min))
	    (when (buffer-live-p sidebar-client-buffer)
	      ;; Occasionally sidebar-buffers get killed.
	      (re-search-forward
	       (concat "^" (regexp-quote (buffer-name sidebar-client-buffer))) nil t)))
	;; No buffers to list.  If the preceding sidebar contained a buffer-list
	;; we should probably delete it, but for the moment we won't care about
	;; this.
	(sidebar-message "No buffers to list")
	(sidebar-delete-window)))))

;; _____________________________________________________________________________
;;                                                                              
;;;                                 match group                                 
;; _____________________________________________________________________________
;;                                                                              
(defsubst sidebar-match-at-link ()
  "Return `sidebar-link' of first `sidebar-match' overlay found at `point'."
  (let ((overlays (overlays-at (point)))
	(window (selected-window)))
    (catch 'found
      (dolist (overlay overlays)
        (when (and (overlay-get overlay 'sidebar-match)
		   (eq (overlay-get overlay 'window) window))
          (throw 'found (overlay-get overlay 'sidebar-link)))))))

(defun sidebar-match-first-overlay (from)
  "Return `sidebar-link' of first `sidebar-match' overlay following FROM.
If no appropriate overlay can be found return nil."
  (let (link)
    (save-excursion
      (goto-char (next-overlay-change from))
      (or (sidebar-match-at-link)
          (progn
            (while (not (or (setq link (sidebar-match-at-link)) (eobp)))
              (goto-char (next-overlay-change (point))))
            link)))))

(defun sidebar-match-last-overlay (from)
  "Return `sidebar-link' of last `sidebar-match' overlay preceding FROM.
If no appropriate overlay can be found return nil."
  (let (link)
    (save-excursion
      (goto-char (previous-overlay-change from))
      (or (sidebar-match-at-link)
          (progn
            (while (not (or (setq link (sidebar-match-at-link)) (bobp)))
              (goto-char (previous-overlay-change (point))))
            link)))))

(defun sidebar-match-adjust (&optional before after)
  "Set `sidebar-match-lines-before' and `sidebar-match-lines-after' simultaneously.
Used by sidebar menu."
  (setq sidebar-match-lines-before
        (or before sidebar-match-lines-before))
  (setq sidebar-match-lines-after
        (or after sidebar-match-lines-after))
  (sidebar-match-refresh))

(defun sidebar-match-highlight ()
  "Highlight matches."
  (when (and (buffer-live-p (current-buffer))
             (eq major-mode 'sidebar-mode)
             (eq sidebar-mode 'match))
    ;; The following is certainly incorrect:
;;;              (eq (get-buffer-window sidebar-client-buffer) sidebar-client-window))
    (let ((buffer (current-buffer))
          (match-context sidebar-match-context)
          (match-lines-before sidebar-match-lines-before)
          (match-lines-after sidebar-match-lines-after)
          (replace-string sidebar-replace-string)
          (context-echo-overlay sidebar-context-echo-overlay)
          (context-overlay sidebar-context-overlay)
          (before-length sidebar-before-length)
          overlay-string overlay-match-string overlay-replace-string)
      (when (> (buffer-chars-modified-tick sidebar-client-buffer)
               sidebar-client-buffer-chars-modified-tick)
        ;; sidebar-client-buffer was modified, we may have to update mismatch overlays
        (setq sidebar-client-buffer-chars-modified-tick
              (buffer-chars-modified-tick sidebar-client-buffer))
	(overlay-recenter (point-max))
        (dolist (overlay (overlays-in (point-min) (point-max)))
          (when (overlay-get overlay 'match) (delete-overlay overlay)))
        (save-excursion
          (goto-char (point-min))
          (skip-syntax-forward " ")
          (while (eq (sidebar-line-face) 'sidebar-match) ; no eobp
            (let ((overlay (sidebar-property)))
              (setq overlay-match-string (overlay-get overlay 'sidebar-match))
              (setq overlay-string
                    (with-current-buffer sidebar-client-buffer
                      (buffer-substring-no-properties
                       (overlay-start overlay) (overlay-end overlay))))
              (unless (string-equal overlay-match-string overlay-string)
                (let ((bar-overlay
                       (make-overlay (line-beginning-position) (line-end-position))))
                  (overlay-put bar-overlay 'match t)
                  ;; priority must override that of window overlay
                  (overlay-put bar-overlay 'priority 1000)
                  ;; don't consider them in match-cleanup
                  (overlay-put bar-overlay 'evaporate t)
                  (overlay-put
                   bar-overlay 'face
                   (if (and (setq overlay-replace-string
                                  (overlay-get overlay 'sidebar-replace))
                            (string-equal overlay-string overlay-replace-string))
                       'sidebar-replace-echo
                     'sidebar-mismatch-echo))))
              (forward-line)))))
      (when context-echo-overlay (delete-overlay context-echo-overlay))
      (when context-overlay (delete-overlay context-overlay))
      (with-current-buffer sidebar-client-buffer
        (sidebar-add-to-invisibility-spec '(sidebar-invisible . t)))
      (sidebar-view)
      (with-saved-window-and-buffer
        (sidebar-select-window sidebar-client-window)
        (save-excursion
          (let* ((from (point))
		 (window (selected-window))
		 (overlays-from (window-start))
		 (overlays-to (window-end window t)))
            ;; put overlays for matches and mismatches in current window
            (dolist (overlay (overlays-in overlays-from overlays-to))
              ;; overlays-in delivers overlays UNSORTED
              (when (and (eq (overlay-get overlay 'window) window)
			 (setq overlay-match-string
                          (overlay-get overlay 'sidebar-match)))
                (setq overlay-string
                      (buffer-substring-no-properties
                       (overlay-start overlay) (overlay-end overlay)))
                (goto-char (overlay-start overlay))
                ;; an idiosyncratic approach to matches and mismatches
                (if (= (point) from)
                    (let ((context (when match-context (sidebar-defun))))
                      (overlay-put
                       overlay 'face
                       (cond
                        ((string-equal overlay-match-string overlay-string)
                         'sidebar-current-match-echo)
                        ((and (setq overlay-replace-string
                                    (overlay-get overlay 'sidebar-replace))
                              (string-equal overlay-string overlay-replace-string))
                         'sidebar-current-replace-echo)
                        (t 'sidebar-current-mismatch-echo)))
		      (overlay-put overlay 'window window)
                      (when context
                        ;; put context overlay for current defun
                        (let* ((context-2 (nth 2 context))
                               (context-within
                                (and (<= (car context) (overlay-start overlay))
                                     (<= (overlay-end overlay) (cadr context))))
                               (context-echo-start
                                (if context-within
                                    (if (memq context-2 '(before between start))
                                        (max (car context)
                                             (line-beginning-position
                                              (- (1- match-lines-before))))
                                      (car context))
                                  (line-beginning-position (- (1- match-lines-before)))))
                               (context-echo-end
                                (if context-within
                                    (if (memq context-2 '(after between start))
                                        (min (cadr context)
                                             (save-excursion
                                               (goto-char (overlay-end overlay))
                                               (line-beginning-position
                                                (+ match-lines-after 2))))
                                      (cadr context))
                                  (save-excursion
                                    (goto-char (overlay-end overlay))
                                    (line-beginning-position (+ match-lines-after 2)))))
                               context-start context-end)
                          (move-overlay
                           context-echo-overlay context-echo-start
                           context-echo-end (current-buffer))
			  (overlay-put context-echo-overlay 'window window)
                          (if (memq context-2 '(before between start))
                              (let ((point (overlay-get overlay 'sidebar-link)))
                                (with-current-buffer buffer
                                  (save-excursion
                                    (goto-char (+ point before-length))
                                    (move-overlay
                                     context-overlay (line-beginning-position)
                                     (line-beginning-position 2) buffer))))
                            (setq context-start
                                  (sidebar-match-first-overlay context-echo-start))
                            (setq context-end
                                  (sidebar-match-last-overlay context-echo-end))
                            (when (and context-start context-end)
                              (with-current-buffer buffer
                                (save-excursion
                                  (move-overlay
                                   context-overlay (+ context-start before-length)
                                   (progn
                                     (goto-char (+ context-end before-length))
                                     (line-beginning-position 2))
                                   buffer))))
                            (let ((message-log-max))
                              (message "%s" (save-excursion
                                              (goto-char context-echo-start)
                                              (buffer-substring
                                               ;; NEVER DISPLAY NEWLINES HERE,
                                               ;; THEY MAKE YOUR SCREEN FLASH
                                               (line-beginning-position) (line-end-position)))))))))
                  (overlay-put
                   overlay 'face
                   (cond
                    ((string-equal overlay-match-string overlay-string)
                     'sidebar-match-echo)
                    ((and (setq overlay-replace-string
                                (overlay-get overlay 'sidebar-replace))
                          (string-equal overlay-string overlay-replace-string))
                     'sidebar-replace-echo)
                    (t 'sidebar-mismatch-echo)))
		  (overlay-put overlay 'window window))))))))))

(defun sidebar-match-regexp-exit-minibuffer ()
  (interactive)
  (setq sidebar-match-exit 'regexp)
  (exit-minibuffer))

(defun sidebar-match-symbol-exit-minibuffer ()
  (interactive)
  (setq sidebar-match-exit 'symbol)
  (exit-minibuffer))

(defun sidebar-match-word-exit-minibuffer ()
  (interactive)
  (setq sidebar-match-exit 'word)
  (exit-minibuffer))

(defun sidebar-match-yank-next-character ()
  "Yank character following `sidebar-match-yank-at' and update latter."
  (interactive)
  (when (and (buffer-live-p sidebar-match-yank-buffer) sidebar-match-yank-at)
    (let ((char
           (with-current-buffer sidebar-match-yank-buffer
             (buffer-substring-no-properties
              sidebar-match-yank-at
              (setq sidebar-match-yank-at
                    (min (1+ sidebar-match-yank-at) (point-max)))))))
      (insert char))))

(defun sidebar-match-yank-word ()
  "Yank word following `sidebar-match-yank-at' and update latter."
  (interactive)
  (when (and (buffer-live-p sidebar-match-yank-buffer) sidebar-match-yank-at)
    (let ((word
           (with-current-buffer sidebar-match-yank-buffer
             (buffer-substring-no-properties
              sidebar-match-yank-at
              (save-excursion
                (goto-char sidebar-match-yank-at)
                (forward-word 1)
                (setq sidebar-match-yank-at (point)))))))
      (insert word))))

(defun sidebar-match-query (&optional from-to)
  "Find all occurrences of queried expression in `sidebar-client-buffer'.
If mark is active search occurrences in active region only."
  (interactive)
  ;; global settings
  (let* ((truncate-partial-width-windows t)
         (window (selected-window))
         string fun from to)
    (setq sidebar-match-exit nil)
    (cond
     ((eq major-mode 'sidebar-mode)
      (with-current-buffer sidebar-client-buffer
        (setq sidebar-match-yank-buffer (current-buffer))
        (setq sidebar-match-yank-at (point))))
     ((and (string-match "isearch" (symbol-name last-command))
           ;; last command was an isearch command, search for `isearch-string'
           (not (string-equal isearch-string "")))
      (setq string isearch-string)
      (setq sidebar-match-exit
            (cond (isearch-word 'word) (isearch-regexp 'regexp)))
      (setq from (point-min))
      (setq to (point-max)))
     (t
      (setq sidebar-match-yank-buffer (current-buffer))
      (setq sidebar-match-yank-at (point))))
    (unless from
      (setq from (save-excursion
                   (when (eq major-mode 'sidebar-mode)
                     (set-buffer sidebar-client-buffer))
                   (cond
                    (mark-active (min (point) (mark)))
                    ((and from-to (setq fun (sidebar-defun))) (car fun))
                    (t (point-min))))))
    (unless to
      (setq to (save-excursion
                 (when (eq major-mode 'sidebar-mode)
                   (set-buffer sidebar-client-buffer))
                 (cond
                  (mark-active (max (point) (mark)))
                  (fun (cadr fun))
                  (t (point-max))))))
    (unless string
      (let ((default (car sidebar-match-history)))
	(setq string
	      (read-from-minibuffer
	       (if default
		   (format "Match (default `%s'): " default)
		 "Match: ")
	       nil sidebar-match-minibuffer-local-map nil
	       'sidebar-match-history default t))
	(when (string-equal string "") (setq string default))))
    ;; deactivate mark in sidebar-client-buffer
    (setq mark-active nil)
    (when (and string (sidebar-make t))
      (sidebar-match-list string sidebar-match-exit window from to))))

(defun sidebar-match-query-from-to ()
  "List matches for queried expression.  If mark is active list matches in
active region only.  If current defun can be defined list matches in defun
only."
  (interactive)
  (sidebar-match-query t))

(defun sidebar-match-object (&optional from to)
  "List all matches for object around `point'.
If mark is active the object is the region between `point' and `mark'.
Otherwise the object is the `sidebar-word-symbol' at `point'."
  (interactive)
  (let* ((at (if (eq major-mode 'sidebar-mode)
                 (window-point sidebar-client-window)
               (point)))
         (selected-window (selected-window))
         string type)
    (when (and (not (eq major-mode 'sidebar-mode)) mark-active)
      (setq string (buffer-substring-no-properties
                    (min (mark) (point)) (max (mark) (point)))))
    (when (sidebar-make t)
      ;; needed for tags or bookmarks
      (unless string
        (with-current-buffer sidebar-client-buffer
          (let ((word-symbol (sidebar-word-symbol)))
            (when word-symbol
              (setq string (buffer-substring-no-properties
                            (car word-symbol) (cdr word-symbol)))
              (setq type 'symbol)))))
      (if string
          (progn
            (setq sidebar-match-history (cons string sidebar-match-history))
            (sidebar-match-list string type selected-window from to))
        (sidebar-kill-or-restore)
	(select-window selected-window)
        (sidebar-message "No suitable object around point")))))

(defun sidebar-match-object-from-to ()
  "List all matches within current defun for object around `point'.
If mark is active, object is the region between `point' and `mark'.  Otherwise
object is the symbol around `point', compare `sidebar-match-object'. If object
is not contained in a defun list all matches for object."
  (interactive)
  (let* ((fun
          (save-excursion
            (when (eq major-mode 'sidebar-mode)
              (set-buffer sidebar-client-buffer))
            (sidebar-defun)))
         (from (if fun (car fun) (point-min)))
         (to (if fun (cadr fun) (point-max))))
    (sidebar-match-object from to)))

(defun sidebar-match-list (string type selected-window &optional from to)
  "List matches for STRING.
Matching is case-sensitive if STRING contains upper case characters (excluding
those preceded by `\\').  TYPE indicates how to interpret STRING and may be nil,
`regexp', `symbol', or `word'.  If no matches are found, select window
SELECTED-WINDOW.  Optional arguments FROM and TO bound the search.
`buffer-of-sidebar' must be current when this function is called.  Matching is
aborted when an empty match is encountered."
  (condition-case nil
      (let* ((buffer-of-sidebar (current-buffer))
             (buffer sidebar-client-buffer)
	     (window sidebar-client-window)
             (match-count 0)
             (line-number 1)
             (previous-match (point-min))
             (match-context sidebar-match-context)
             (match-lines-before sidebar-match-lines-before)
             (this-before (1+ (- sidebar-match-lines-before)))
             (this-after (+ sidebar-match-lines-after 2))
             (buffer-case-fold-search
              (with-current-buffer buffer
                (and case-fold-search
                     (isearch-no-upper-case-p string (eq type 'regexp)))))
             (regexp
              (cond
               ((eq type 'regexp) (concat "\\(" string "\\)"))
               ((eq type 'symbol)
                (concat
                 "\\(?:\\`\\|\\S_\\|\\(?:\\`\\|\\S_\\)\\B\\s_+\\)\\<\\("
                 (regexp-quote string)
                 "\\)\\>\\(?:\\'\\|\\S_\\|\\s_+\\B\\(?:\\'\\|\\S_\\)\\)"))
               ((eq type 'word)
                (concat "\\(\\b" (regexp-quote string) "\\b\\)"))
               (t (concat "\\(" (regexp-quote string) "\\)"))))
             (inhibit-read-only t)
	     first-match point-old point-new line-number-format
             match-list match-from match-to this-from this-to last-to
             defun defun-from defun-from-two defun-to)
	;; *** TODO: Most stuff here is plain idiotism ***
        (set-buffer buffer)
        (setq from (or from (point-min)))
        (setq to (or to (point-max)))
        ;; narrow buffer for one search only
        (save-restriction
          (narrow-to-region from to)
          (let ((case-fold-search buffer-case-fold-search))
            (setq point-old (point))
            (goto-char (or from (point-min)))
            (if (and (re-search-forward regexp to t)
                     (or (> (match-end 1) (match-beginning 1))
                         (setq match-count nil)))
		(progn
		  (setq first-match (match-beginning 0))
		  (set-buffer buffer)
		  (widen))
	      ;; no matches
	      (with-current-buffer buffer-of-sidebar
		(sidebar-kill-or-restore))
	      ;; The following is a mess for grep's sake.
	      (if (window-live-p selected-window)
		  (sidebar-select-window selected-window)
		(sidebar-select-window window)))))
        (when first-match
          ;; at least one match exists
          (set-buffer buffer-of-sidebar)
	  (sidebar-switch-mode 'match)
          (setq sidebar-match-string string)
	  (setq sidebar-replace-string nil)
          (setq sidebar-match-type type)
          (unless (and (buffer-base-buffer sidebar-client-buffer)
                       (eq (with-current-buffer sidebar-client-buffer
                             ;; note: sidebar-client-buffer of sidebar-client-buffer
                             sidebar-client-buffer)
                           buffer-of-sidebar))
            (let ((flm font-lock-mode))
              (setq sidebar-client-buffer buffer)
              (setq sidebar-buffer-window-list
		    (sidebar-add
		     (cons sidebar-client-buffer sidebar-client-window) sidebar-buffer-window-list))
              (when flm (with-current-buffer buffer
                          (font-lock-fontify-buffer)))
              (set-window-and-current-buffer sidebar-client-window buffer)))
          (set-buffer sidebar-client-buffer)
          ;; narrow from _point-min_ to handle first invisibility overlay
          (narrow-to-region (point-min) to)
          (setq last-to (point-min))
          (goto-char first-match)
          ;; note: sidebar-client-buffer of sidebar-client-buffer !
          (setq sidebar-client-buffer buffer-of-sidebar)
          ;; invisibility
          (sidebar-add-to-invisibility-spec '(sidebar-invisible . t))
          (setq mark-active nil)
          (let ((case-fold-search buffer-case-fold-search))
            ;; search
            (while (and (re-search-forward regexp to t)
                        (setq match-from (match-beginning 1))
                        (setq match-to (match-end 1))
                        (or (> match-to match-from)
                            (sidebar-message "Encountered empty match, aborted")))
              (goto-char match-from)
              (beginning-of-line)
              (setq this-from (line-beginning-position this-before))
              (setq line-number
                    (+ line-number (count-lines previous-match (point))))
              (setq previous-match (point))
              (setq match-count (1+ match-count))
              (let (message-log-max)
                (message "%s matches for %s" match-count string))
              (let ((overlay (make-overlay match-from match-to nil t nil)))
                ;; a front-advance, rear-stay-behind overlay
                (setq match-list (cons (cons line-number overlay) match-list))
                (overlay-put
                 overlay 'sidebar-match
                 (buffer-substring-no-properties match-from match-to))
		(overlay-put overlay 'window window))
              (cond
               ;;       1. match-context
               ((and match-context
                     ;; 2. fun can be defined
                     (setq defun (sidebar-defun))
                     (setq defun-from (car defun))
                     (setq defun-from-two
                           (save-excursion
                             (goto-char defun-from) (line-beginning-position 2)))
                     (setq defun-to (cadr defun))
                     ;; 3. entire match contained in defun
                     (<= defun-from match-from)
                     (<= match-to defun-to))
                (if (memq (nth 2 defun) '(before between after))
                    (let ((from (max this-from defun-from)))
                      (when (> from last-to)
                        (let ((overlay (make-overlay last-to (1- from))))
                          (overlay-put overlay 'invisible 'sidebar-invisible)
			  (overlay-put overlay 'window window)))
                      (goto-char match-to)
                      (setq last-to
                            (max (point)
                                 (min (line-beginning-position this-after) defun-to))))
                  (when (> defun-from last-to)
                    ;; make invisibility preceding defun
                    (let ((overlay (make-overlay last-to (1- defun-from))))
                      (overlay-put overlay 'invisible 'sidebar-invisible)
		      (overlay-put overlay 'window window)))
                  (let ((overlay-from (max defun-from-two last-to)))
                    ;; make invisibility preceding match within defun
                    (when (> this-from overlay-from)
                      (let ((overlay (make-overlay overlay-from (1- this-from))))
                        (overlay-put overlay 'invisible 'sidebar-invisible)
			(overlay-put overlay 'window window))))
                  (goto-char match-to)
                  (setq last-to (max (point)
                                     (min (line-beginning-position this-after) defun-to)))))
               (t                ; no match-context or match not contained in defun
                (when (> this-from last-to)
                  ;; make invisibility preceding match
                  (let ((overlay (make-overlay last-to (1- this-from))))
                    (overlay-put overlay 'invisible 'sidebar-invisible)
		    (overlay-put overlay 'window window)))
                (goto-char match-to)
                (setq last-to (line-beginning-position this-after)))))
            ;; if we don't widen here the last invisibility overlay is not made
            (widen)
            ;; invisibility for text following last match
            (when (> (point-max) last-to)
              ;; for some strange reason we can't make an overlay that extends to
              ;; point-max, otherwise we may kill the fringe in sidebar-client-buffer
              ;; AND buffer-of-sidebar provided the match causes a hscroll,
              ;; hence we let the overlay terminate ONE CHARACTER BEFORE
              ;; point-max (until this gets us problems too)
              (let ((overlay (make-overlay last-to (1- (point-max)))))
                (overlay-put overlay 'invisible 'sidebar-invisible)
		(overlay-put overlay 'window window))))
          ;; back to buffer-of-sidebar
          (set-buffer buffer-of-sidebar)
          (setq match-list (reverse match-list))
          (let ((line-number-length (length (number-to-string line-number))))
            (setq line-number-format (format "%%%dd" line-number-length))
            (setq sidebar-width line-number-length)
            (sidebar-adjust-window))
          ;; insert match lines
          (dolist (match match-list)
            (let ((line (car match))
                  (overlay (cdr match)))
              ;; install reverse link (line-beginning-position of matching line)
              (overlay-put overlay 'sidebar-link (point))
              (unless (or point-new
                          (> point-old (overlay-end overlay)))
                ;; Note: point-old is in sidebar-client-buffer, point-new in
                ;; buffer-of-sidebar
                (setq point-new (line-beginning-position)))
              (insert (format line-number-format line) "\n")
              (sidebar-put-text 'sidebar-match overlay)
              (sidebar-put-overlay)))
          (setq sidebar-client-buffer-chars-modified-tick
                (buffer-chars-modified-tick sidebar-client-buffer))
          ;; insert (why?) 4 + sidebar-match-lines-before at head of sidebar
          (goto-char (point-min))
          (insert (make-string (+ sidebar-match-lines-before 4) ?\n))
          (setq sidebar-before-length (1- (point)))
          ;; recenter, maybe we can stay here
          (recenter (+ sidebar-match-lines-before (if match-context 2 0)))
          ;; go some place
          (when point-new
            (goto-char (+ point-new sidebar-before-length)))
          (unless (pos-visible-in-window-p)
            (recenter (+ sidebar-match-lines-before (if match-context 2 0)))))
        (let ((type-string
               (if (memq type '(regexp symbol word))
                   (concat "`" (symbol-name type) "' ")
                 ""))
              message-log-max)
          (cond
           ((not match-count) (sidebar-message "Empty match, aborted"))
           ((= match-count 0)
            (sidebar-message "No %smatches for \"%s\"" type-string string))
           ((= match-count 1)
            (message "One %smatch for \"%s\"" type-string string)
            (setq sidebar-match-count match-count))
           (t
            (message "%d %smatches for \"%s\"" match-count type-string string)
            (setq sidebar-match-count match-count)))))
    (quit (sidebar-message "Matching aborted")
          ;; we can't restore original sidebar here
          (sidebar-delete-window (selected-window)))))

(defun sidebar-match-refresh ()
  "Refresh match list."
  (with-valid-sidebar
   (let* ((match-context sidebar-match-context)
          (match-lines-before sidebar-match-lines-before)
          (this-before (1+ (- sidebar-match-lines-before)))
          (this-after (+ sidebar-match-lines-after 2))
          (first-match t)
          (inhibit-read-only t)
	  (window sidebar-client-window)
	  overlay-list before-first
	  match-from match-to this-from this-to last-to
          defun defun-from defun-from-two defun-to)
     (save-excursion
       (goto-char (point-min))
       (skip-syntax-forward " ")
       (when (eq (sidebar-line-face) 'sidebar-match)
         ;; the following is needed since match-list doesn't exist anymore
         (let ((from (overlay-start (sidebar-property))))
           (with-current-buffer sidebar-client-buffer
             (setq before-first
                   (count-lines
                    (point-min)
                    (save-excursion (goto-char from) (line-beginning-position)))))))
       (while (eq (sidebar-line-face) 'sidebar-match) ; no eobp
         (setq overlay-list (cons (sidebar-property) overlay-list))
         (forward-line))
       (setq overlay-list (nreverse overlay-list)))
     (with-current-buffer sidebar-client-buffer
       ;; remove invisibility overlays
       (save-restriction
         (widen)
         (remove-from-invisibility-spec '(sidebar-invisible . t))
	 (overlay-recenter (point-max))
         (dolist (overlay (overlays-in (point-min) (point-max)))
           (when (and (eq (overlay-get overlay 'invisible) 'sidebar-invisible)
		      (eq (overlay-get overlay 'window) window))
             (delete-overlay overlay)))
         (setq last-to (point-min))
         (dolist (overlay overlay-list)
           (when (and (overlay-get overlay 'sidebar-match)
		      (eq (overlay-get overlay 'window) window))
             (setq match-from (overlay-start overlay))
             (setq match-to (overlay-end overlay))
             (goto-char match-from)
             (beginning-of-line)
             (setq this-from (line-beginning-position this-before))
             (cond
              ;;       1. match-context
              ((and match-context
                    ;; 2. fun can be defined
                    (setq defun (sidebar-defun))
                    (setq defun-from (car defun))
                    (setq defun-from-two
                          (save-excursion
                            (goto-char defun-from) (line-beginning-position 2)))
                    (setq defun-to (cadr defun))
                    ;; 3. entire match contained in defun
                    (<= defun-from match-from)
                    (<= match-to defun-to))
               (if (memq (nth 2 defun) '(before between after))
                   (let ((from (max this-from defun-from)))
                     (when (> from last-to)
                       (let ((overlay (make-overlay last-to (1- from))))
                         (overlay-put overlay 'invisible 'sidebar-invisible)
			 (overlay-put overlay 'window window)))
                     (goto-char match-to)
                     (setq last-to (max (point) (min (line-beginning-position this-after)
                                                     defun-to))))
                 ;; make invisibility preceding defun, `1- defun-from' means
                 ;; that we don't always see the dots preceding the first match
                 (when (> defun-from last-to)
                   (let ((overlay (make-overlay last-to (1- defun-from))))
                     (overlay-put overlay 'invisible 'sidebar-invisible)
		     (overlay-put overlay 'window window)))
                 (let ((overlay-from (max defun-from-two last-to)))
                   (when (> this-from overlay-from)
                     ;; make invisibility preceding match within defun
                     (let ((overlay (make-overlay overlay-from (1- this-from))))
                       (overlay-put overlay 'invisible 'sidebar-invisible)
		       (overlay-put overlay 'window window))))
                 ;; next invisibility may already start at end of defun 
                 (goto-char match-to)
                 (setq last-to (max (point)
                                    (min (line-beginning-position this-after) defun-to)))))
              (t
               (when (> this-from last-to)
                 ;; make invisibility preceding match
                 (let ((overlay (make-overlay last-to (1- this-from))))
                   (overlay-put overlay 'invisible 'sidebar-invisible)
		   (overlay-put overlay 'window window)))
               (goto-char match-to)
               (setq last-to (line-beginning-position this-after))))
             (setq first-match nil)))
         ;; invisibility for text following last match
         (when (> (point-max) last-to)
           ;; see sidebar-match-list for comment on this
           (let ((overlay (make-overlay last-to (1- (point-max)))))
             (overlay-put overlay 'invisible 'sidebar-invisible)
	     (overlay-put overlay 'window window))))
       (sidebar-add-to-invisibility-spec '(sidebar-invisible . t)))
     ;; refresh hidden lines
     (save-excursion
       (delete-region (point-min) (1+ sidebar-before-length))
       (goto-char (point-min))
       ;; insert 4 + sidebar-match-lines-before at head of sidebar
       (insert (make-string (+ 4 sidebar-match-lines-before) ?\n))
       (recenter (+ sidebar-match-lines-before (if match-context 2 0)))
       (setq sidebar-before-length (1- (point)))))))

(defun sidebar-match-cleanup (buffer window)
  "Cleanup match stuff in buffer BUFFER displayed in window WINDOW."
  (with-current-buffer buffer
    (remove-from-invisibility-spec '(sidebar-invisible . t))
    (save-restriction
      (widen)
      (overlay-recenter (point-max))
      (dolist (overlay (overlays-in (point-min) (point-max)))
	(when (and (eq (overlay-get overlay 'window) window)
		   (or (eq (overlay-get overlay 'invisible) 'sidebar-invisible)
		       (overlay-get overlay 'sidebar-match)))
	  (delete-overlay overlay))))))

;; _____________________________________________________________________________
;;                                                                              
;;                                   replace                                    
;; _____________________________________________________________________________
;;                                                                              

;; *** TODO: The following two must go elsewhere, eventually we need different
;; *** keymaps ***
(defun sidebar-space ()
  (interactive)
  (cond
   ((eq sidebar-mode 'directory)
    (sidebar-mark-toggle)
    (sidebar-down))
   ((eq sidebar-mode 'match)
    (sidebar-replace))))

(defun sidebar-backspace ()
  (interactive)
  (cond
   ((eq sidebar-mode 'directory)
    (sidebar-up)
    (sidebar-mark-toggle))
   ((eq sidebar-mode 'match)
    (sidebar-replace t))))

(defun sidebar-replace-all ()
  (interactive)
  (sidebar-replace-where 'all))

(defun sidebar-replace-after ()
  (interactive)
  (sidebar-replace-where 'after))

(defun sidebar-replace-before ()
  (interactive)
  (sidebar-replace-where 'before))

(defun sidebar-replace-reset-forward ()
  (interactive)
  (sidebar-replace-reset))

(defun sidebar-replace-reset-backward ()
  (interactive)
  (sidebar-replace-reset t))

(defun sidebar-replace-reset-all ()
  (interactive)
  (sidebar-replace-reset-where 'all))

(defun sidebar-replace-reset-after ()
  (interactive)
  (sidebar-replace-reset-where 'after))

(defun sidebar-replace-reset-before ()
  (interactive)
  (sidebar-replace-reset-where 'before))

(defun sidebar-replace-edit-forward ()
  (interactive)
  (sidebar-replace-edit))

(defun sidebar-replace-edit-backward ()
  (interactive)
  (sidebar-replace-edit t))

(defun sidebar-replace-edit-all ()
  (interactive)
  (sidebar-replace-edit 'all))

(defun sidebar-replace-edit-after ()
  (interactive)
  (sidebar-replace-edit 'after))

(defun sidebar-replace-edit-before ()
  (interactive)
  (sidebar-replace-edit 'before))

(defun sidebar-replace-edit (&optional where)
  "Edit `sidebar-replace-string' and proceed as with standard replace commands.
If WHERE is a string call `sidebar-replace-where' else call `sidebar-replace'."
  (when (eq sidebar-mode 'match)
    ;; to edit from scratch provide empty string
    (setq sidebar-replace-string (or sidebar-replace-string ""))
    (setq sidebar-replace-string
          (read-from-minibuffer
           (format "Replace by `%s': " sidebar-replace-string)
           nil nil nil 'sidebar-match-history sidebar-replace-string t))
    (if (memq where '(all after before))
        (sidebar-replace-where where)
      (sidebar-replace where))))

(defun sidebar-replace-string ()
  "Read expression for `sidebar-replace'."
  (let* ((default (car sidebar-match-history))
         (input
          (read-from-minibuffer
           (format "Replace by: ") nil nil nil
           'sidebar-match-history default t)))
    input))

(defun sidebar-replace (&optional backward)
  "Replace match."
  (with-valid-sidebar
   (when (eq (sidebar-line-face) 'sidebar-match)
     (unless sidebar-replace-string
       (setq sidebar-replace-string (sidebar-replace-string)))
     (let* ((overlay (sidebar-property))
            (overlay-from (overlay-start overlay))
            (overlay-to (overlay-end overlay))
            (overlay-string
             ;; don't transfer properties to replace-string
             (with-current-buffer sidebar-client-buffer
               (buffer-substring-no-properties overlay-from overlay-to)))
            ;; nocasify lifted from replace.el
            (nocasify (not (and case-fold-search case-replace
                                (string-equal overlay-string
                                              (downcase overlay-string)))))
            (overlay-match-string (overlay-get overlay 'sidebar-match))
            (overlay-link (overlay-get overlay 'sidebar-link))
            (replace-string sidebar-replace-string))
       (with-current-buffer sidebar-client-buffer
         (delete-overlay overlay)
         (goto-char overlay-from)
         ;; looking-at must set match-data here, _needs_ a regexp
         (looking-at (regexp-quote overlay-string))
         (replace-match replace-string)
         (setq overlay
               (make-overlay
                overlay-from (+ overlay-from (length replace-string)) nil t nil))
         (overlay-put overlay 'sidebar-match overlay-match-string)
         (overlay-put overlay 'sidebar-replace
                      (buffer-substring-no-properties
                       (overlay-start overlay) (overlay-end overlay)))
         (overlay-put overlay 'sidebar-link overlay-link))
       (put-text-property
        (line-beginning-position) (line-beginning-position 2) 'sidebar-property overlay)
       (if backward (sidebar-up) (sidebar-down))
       (sidebar-view)))))

(defun sidebar-replace-reset (&optional backward)
  "Reset match."
  (with-valid-sidebar
   (when (eq (sidebar-line-face) 'sidebar-match)
     (let* ((overlay (sidebar-property))
            (overlay-from (overlay-start overlay))
            (overlay-to (overlay-end overlay))
            (overlay-match-string (overlay-get overlay 'sidebar-match))
            (overlay-point (overlay-get overlay 'sidebar-link)))
       (with-current-buffer sidebar-client-buffer
         (delete-overlay overlay)
         (delete-region overlay-from overlay-to)
         (goto-char overlay-from)
         (insert overlay-match-string)
         (setq overlay
               (make-overlay
                overlay-from (+ overlay-from (length overlay-match-string)) nil t nil))
         (overlay-put overlay 'sidebar-match overlay-match-string)
         (overlay-put overlay 'sidebar-replace nil)
         (overlay-put overlay 'sidebar-link overlay-point))
       (put-text-property
        (line-beginning-position) (line-beginning-position 2) 'sidebar-property overlay)
       (if backward (sidebar-up) (sidebar-down))
       (sidebar-view)))))

(defun sidebar-replace-where (&optional where)
  (with-valid-sidebar
   (when (eq (sidebar-line-face) 'sidebar-match)
     (unless sidebar-replace-string
       (setq sidebar-replace-string (sidebar-replace-string)))
     (let ((replace-string sidebar-replace-string)
           stop)
       (save-excursion
         (cond
          ((eq where 'all)
           (goto-char (point-min))
           (skip-syntax-forward " "))
          ((eq where 'before)
           (setq stop (/= (forward-line -1) 0))))
         (while (and (eq (sidebar-line-face) 'sidebar-match)
                     (not stop))
           (let* ((overlay (sidebar-property))
                  (overlay-from (overlay-start overlay))
                  (overlay-to (overlay-end overlay))
                  (overlay-string
                   (with-current-buffer sidebar-client-buffer
                     (buffer-substring-no-properties overlay-from overlay-to)))
                  (overlay-match-string (overlay-get overlay 'sidebar-match))
                  (overlay-point (overlay-get overlay 'sidebar-link)))
             (with-current-buffer sidebar-client-buffer
               (delete-overlay overlay)
               (delete-region overlay-from overlay-to)
               (goto-char overlay-from)
               (insert replace-string)
               (setq overlay
                     (make-overlay
                      overlay-from (+ overlay-from (length replace-string))
                      nil t nil))
               (overlay-put overlay 'sidebar-match overlay-match-string)
               (overlay-put overlay 'sidebar-replace replace-string)
               (overlay-put overlay 'sidebar-link overlay-point))
             (put-text-property
              (line-beginning-position) (line-beginning-position 2) 'sidebar-property overlay))
           (if (eq where 'before)
               (setq stop (/= (forward-line -1) 0))
             (forward-line 1)
             (setq stop (eobp)))))))))

(defun sidebar-replace-reset-where (&optional where)
  (with-valid-sidebar
   (when (eq (sidebar-line-face) 'sidebar-match)
     (let (stop)
       (save-excursion
         (cond
          ((eq where 'all)
           (goto-char (point-min))
           (skip-syntax-forward " "))
          ((eq where 'before)
           (setq stop (/= (forward-line -1) 0))))
         (while (and (eq (sidebar-line-face) 'sidebar-match)
                     (not stop))
           (let* ((overlay (sidebar-property))
                  (overlay-from (overlay-start overlay))
                  (overlay-to (overlay-end overlay))
                  (overlay-string
                   (with-current-buffer sidebar-client-buffer
                     (buffer-substring-no-properties overlay-from overlay-to)))
                  (overlay-match-string (overlay-get overlay 'sidebar-match))
                  (overlay-point (overlay-get overlay 'sidebar-link)))
             (with-current-buffer sidebar-client-buffer
               (delete-overlay overlay)
               (delete-region overlay-from overlay-to)
               (goto-char overlay-from)
               (insert overlay-match-string)
               (setq overlay
                     (make-overlay
                      overlay-from (+ overlay-from (length overlay-match-string))
                      nil t nil))
               (overlay-put overlay 'sidebar-match overlay-match-string)
               (overlay-put overlay 'sidebar-replace nil)
               (overlay-put overlay 'sidebar-link overlay-point))
             (put-text-property
              (line-beginning-position) (line-beginning-position 2) 'sidebar-property overlay))
           (if (eq where 'before)
               (setq stop (/= (forward-line -1) 0))
             (forward-line 1)
             (setq stop (eobp)))))))))

;; _____________________________________________________________________________
;;                                                                              
;;;                                 grep group                                  
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-symbol-string ()
  "Return string of all characters < 256 with symbol-syntax."
  (let ((index (1+ ?\ ))
        (string "")
        minus)
    (while (< index 256)
      (when (eq (char-syntax index) ?_)
        (cond
         ((eq index ?\-) (setq minus t))
         ;; eventually append `-'
         ((eq index ?\])
          ;; prepend `]'
          (setq string (concat "]" string)))
         ((eq index ?\%)
          ;; append `%%'
          (setq string (concat "%%" string)))
         (t (setq string (concat string (string index))))))
      (setq index (1+ index)))
    (when minus (setq string (concat string "-")))))

(defun sidebar-grep-file-list (&optional base-directory)
  "Run recursive ls on BASE-DIRECTORY (defaults to default-directory) and filter
results to a list of files (no directories or symlinks)."
  (interactive)
  (let* ((buffer-of-sidebar (current-buffer))
         (filter-apply sidebar-filter-apply)
         (filter-file-list sidebar-filter-file-list)
         file-list directory base-base-directory)
    (setq sidebar-ls-buffer (get-buffer-create "*sidebar-ls*"))
    (with-current-buffer sidebar-ls-buffer (erase-buffer))
    (call-process "ls" nil sidebar-ls-buffer nil "-RU"
		  (or base-directory default-directory))
    (with-current-buffer sidebar-ls-buffer
      (setq directory base-directory)
      (goto-char (point-min))
      (while (not (eobp))
        (cond
         ((looking-at "\\(.*\\):$")
          (setq directory (match-string 1)))
         ((not (looking-at " *$"))
          (let* ((file-name-sans
                  (buffer-substring (line-beginning-position) (line-end-position)))
                 (file (concat directory file-name-sans)))
            (when (and (or (file-symlink-p file)
                           (not (file-directory-p file)))
		       (or (not filter-apply)
                           (not (sidebar-filter-out
                                 file-name-sans filter-file-list))))
              (setq file-list (cons file file-list))))))
        (forward-line))
      file-list)))

(defun sidebar-grep-query ()
  "Note: Symbol means syntax of sidebar-client-buffer not of the files investigated!"
  (interactive)
  (setq sidebar-match-exit nil)
  (with-current-buffer sidebar-client-buffer
    (setq sidebar-match-yank-buffer (current-buffer))
    (setq sidebar-match-yank-at (point)))
  (let* ((default (car sidebar-match-history))
         (string
          (read-from-minibuffer
           (if default
               (format "Match (default `%s'): " default)
             "Match: ")
           nil sidebar-match-minibuffer-local-map nil
           'sidebar-match-history default t)))
    (when (string-equal string "") (setq string default))
    (if string
        (sidebar-grep-list string sidebar-match-exit)
      (sidebar-message "Empty string"))))

(defun sidebar-grep-list (string type)
  "Run synchronous grep query through all files in this directory.
Can't handle TYPE symbol as matching does word boundary presently."
  (let* ((options (concat "-Rls" ; -l "file names only", -s "no messages"
                          (when (and case-fold-search
                                     (isearch-no-upper-case-p
                                      string (eq type 'regexp)))
                            "i") ; -i "ignore case"
                          ;; the following must be passed to
                          ;; sidebar-grep-file-list !!!!!
                          (when (eq type 'word) "w")
                          "e")) ; -e "next argument is searchstring"
         (regexp
          (cond
           ((eq type 'regexp) string)
           ((eq type 'symbol)
            ;; unlikely that this coincides always with Emacs 22.1 \\_<
            ;; and \\_> but give it a try nevertheless
            (with-current-buffer sidebar-client-buffer
              (let ((symbol-string (sidebar-symbol-string)))
                (concat
                 "\\(?:\\`\\|[^" symbol-string "]\\|\\(?:\\`\\|[^" symbol-string
                 "]\\)\\B[" symbol-string "]+\\)\\<\\(" (regexp-quote string)
                 "\\)\\>\\(?:\\'\\|[^" symbol-string "]\\|[" symbol-string
                 "]+\\B\\(?:\\'\\|[^" symbol-string "]\\)\\)"))))
           (t (regexp-quote string))))
         (arguments (concat " " options " " (shell-quote-argument regexp)
                            " " (shell-quote-argument default-directory)))
         (sort-fold-case (and case-fold-search
                              (isearch-no-upper-case-p
                               string (eq type 'regexp))))
         (filter-apply sidebar-filter-apply)
         (filter-file-list sidebar-filter-file-list))
    ;; create grep buffer if needed
    (unless (buffer-live-p sidebar-grep-buffer)
      (setq sidebar-grep-buffer (get-buffer-create "*sidebar-grep*")))
    (with-current-buffer sidebar-grep-buffer (erase-buffer))
    (setq sidebar-grep-buffer-of-sidebar (current-buffer))
    (message "Grep searching for `%s\' ..." string)
    (call-process
     shell-file-name nil sidebar-grep-buffer nil
     shell-command-switch (concat "grep " arguments))
    (with-current-buffer sidebar-grep-buffer
      (if (= (point-min) (point-max))
          (sidebar-message "No results for `%s'" string)
        (goto-char (point-min))
        (with-current-buffer sidebar-grep-buffer-of-sidebar
          (setq buffer-read-only nil)
          (sidebar-switch-mode 'grep)
          (setq sidebar-grep-string string)
          (setq sidebar-grep-type type)
          (setq sidebar-width sidebar-minimum-width))
        (message "Inserting ...")
        (while (not (eobp))
          (let ((file (buffer-substring (line-beginning-position) (line-end-position))))
            ;; handle grep's ":" unless part of a ms-dos/windows drive name
            (unless (or (and filter-apply
                             (sidebar-filter-out file filter-file-list))
                        (string-match ": " file))
              (with-current-buffer sidebar-grep-buffer-of-sidebar
                (when (memq system-type '(ms-dos windows-nt))
                  (while (string-match "/" file)
                    (setq file (replace-match "\\" t t file))))
                (sidebar-put-line
                 (file-name-nondirectory file) 'sidebar-grep-name file)))
            (forward-line)))
        (with-current-buffer sidebar-grep-buffer-of-sidebar
          (goto-char (point-min))
          (message "Sorting ...")
          (sort-lines nil (point-min) (point-max))
          (message "Grep complete")
          (sidebar-post-command))))))

;; _____________________________________________________________________________
;;                                                                              
;;;                           customization and menu                            
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-menu-filter-toggle (index)
  "Toggle state of INDEXth filter in `sidebar-filter-list' and revert
sidebar if needed."
  (let (list)
    (setq list (sort (copy-sequence sidebar-filter-list)
                     'sidebar-filter-lessp))
    (setcar (nth index list) (not (car (nth index list))))
    (setq sidebar-filter-list (sort list 'sidebar-filter-lessp))
    (sidebar-filter-list-regexp)
    (when (and (eq major-mode 'sidebar-mode)
               (eq sidebar-mode 'directory))
      ;; we revert directories only
      (sidebar-revert))))

(defun sidebar-menu-filter (menu)
  "Build menu to toggle filters."
  (mapcar
   (lambda (filter)
     (vector (nth 1 filter)
             (list 'sidebar-menu-filter-toggle (nth 2 filter))
             :style 'toggle
             :active sidebar-filter-apply
             :selected (car filter)))
   (let ((count 0)
         ;; count is the argument used in callback
         (filter-list
            (if sidebar-filter-list-sort
                ;; operate on copy of sidebar-filter-list
                (sort (copy-sequence sidebar-filter-list)
                      'sidebar-filter-list-lessp)
              (copy-sequence sidebar-filter-list)))
         filters)
     (dolist (filter filter-list)
       (let* ((name (sidebar-filter-name filter))
              (menu-name
               (if (string-match "\\S " name)
                   ;; filter name contains a word character use name of filter
                   name
                 ;; construct name from filter string
                 (let* ((string (sidebar-filter-string filter))
                        (length (length string)))
                   (concat ;; maximum length of filter string is 48
                    "<"
                    (if (<= length 48)
                        string
                      (concat (substring string 0 44) "..."))
                    ">")))))
       (setq filters
               (cons
                (list
                 (sidebar-filter-state filter)
                 (if (string-match "^-+$" menu-name) ; separator line
                     (concat " " menu-name)
                   menu-name)
                 count)
                filters))
         (setq count (1+ count))))
     (nreverse filters))))

(defun sidebar-set-variable (variable value)
  "Set VARIABLE to VALUE and revert sidebar."
  (set variable value)
  (when (eq major-mode 'sidebar-mode) (sidebar-revert)))

(defun sidebar-negate-variable (variable)
  "Negate value of VARIABLE and revert sidebar."
  (set variable (not (symbol-value variable)))
  (when (eq major-mode 'sidebar-mode) (sidebar-revert)))

;; _____________________________________________________________________________
;;                                                                              
;;;                          rename, move, copy group                           
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-rename-visiting-buffer (file)
  "Return indirect or modified buffer visiting file FILE.  If there is no such
buffer return nil."
  (save-excursion
    (catch 'buffer
      (dolist (buffer (buffer-list)) ; awful
        (when (and (buffer-live-p buffer)
                   (or (buffer-modified-p buffer) (buffer-base-buffer buffer))
                   (buffer-file-name buffer)
                   (sidebar-same-files (buffer-file-name buffer) file))
          (throw 'buffer buffer))))))

(defun sidebar-rename-directory-visiting-buffer (directory)
  "Return indirect or modified buffer provided DIRECTORY is in its path.
If there is no such buffer return nil."
  (let ((regexp (concat "^" (regexp-quote (file-truename directory)))))
    (save-excursion
      (catch 'buffer
        (dolist (buffer (buffer-list))  ; awful
          (when (and (buffer-live-p buffer)
                     (or (buffer-modified-p buffer)
                         (buffer-base-buffer buffer))
                     (buffer-file-name buffer)
                     (string-match
                      regexp (file-truename (buffer-file-name buffer))))
            (throw 'buffer buffer)))))))

(defun sidebar-rename-directory (old new)
  "Rename directory OLD to NEW."
  (let ((regexp (concat "^" (regexp-quote (file-truename old)))))
    (let (modified)
      (bookmark-maybe-load-default-file)
      (dolist (bookmark bookmark-alist)
        (let ((name (car bookmark))
              (file (sidebar-bookmark-file-name bookmark)))
          (when (and file (string-match regexp file))
            ;; replaces new in name a weird thing to do
            (bookmark-set-filename
             name (concat new (substring file (match-end 0))))
            (setq modified t))))
      (when modified
        (setq bookmark-alist-modification-count
              (1+ bookmark-alist-modification-count)))
      (when (bookmark-time-to-save-p) (bookmark-save)))
    ;; search for all buffers with old in path
    (save-excursion
      (dolist (buffer (buffer-list))
        (when (and (buffer-live-p buffer)
                   (buffer-file-name buffer)
                   (not (buffer-base-buffer buffer))
                   (string-match
                    regexp (file-truename (buffer-file-name buffer))))
          ;; old is in path
          (with-current-buffer buffer
            (let ((buffer-modified-p (buffer-modified-p buffer)))
              ;; make buffer visit new
              (set-visited-file-name
               (concat new (substring buffer-file-name (match-end 0))) nil t)
              (clear-visited-file-modtime)
              (set-buffer-modified-p buffer-modified-p)))))
      (dolist (buffer (buffer-list))
        (when (buffer-live-p buffer)
          (with-current-buffer buffer
            (when (eq major-mode 'sidebar-mode)
              ;; should be able to use sidebar-file here
              (let ((file (file-truename default-directory))
                    modified)
                (when (string-match regexp file)
                  ;; name of default directory changed
                  (setq default-directory
                        (concat new (substring
                                     default-directory (match-end 0))))
                  (setq modified t))
                (when (and (stringp sidebar-file)
                           (string-match regexp sidebar-file))
                  (setq sidebar-file
                        (concat new (substring
                                     sidebar-file (match-end 0)))))
                (when modified (sidebar-update-mode-line)))))))
      (sidebar-reconfigure))
    ;; revert sidebar
    (cond
     ((eq sidebar-mode 'directory)
      (sidebar-directory-list
       default-directory (file-name-nondirectory new) t))
     ((eq sidebar-mode 'bookmark)
      (sidebar-bookmark-list (file-name-nondirectory new))))))

(defun sidebar-rename-non-directory (old new)
  ;; rename bookmarks
  (let (modified)
    (bookmark-maybe-load-default-file)
    (dolist (bookmark bookmark-alist)
      (let ((name (car bookmark))
            (file (sidebar-bookmark-file-name bookmark)))
        (when (sidebar-same-files file old)
          (bookmark-set-filename name new)
          (setq modified t))))
    (when modified
      (setq bookmark-alist-modification-count
            (1+ bookmark-alist-modification-count)))
    (when (bookmark-time-to-save-p) (bookmark-save)))
  ;; search for all buffers with old in path
  (save-excursion
    (dolist (buffer (buffer-list))
      (when (and (buffer-live-p buffer)
                 (buffer-file-name buffer)
                 (not (buffer-base-buffer buffer))
                 (sidebar-same-files (buffer-file-name buffer) old))
        ;; live buffer visiting old
        (with-current-buffer buffer
          (let ((buffer-modified-p (buffer-modified-p)))
            ;; make buffer visit new
            (set-visited-file-name new nil t)
            (clear-visited-file-modtime)
            (set-buffer-modified-p buffer-modified-p)))))
    (dolist (buffer (buffer-list))
      (when (buffer-live-p buffer)
        (with-current-buffer buffer
          (cond
           ((and (eq major-mode 'sidebar-mode)
                 ;; should be able to use sidebar-file here
                 sidebar-file
                 (sidebar-same-files sidebar-file old))
            (setq sidebar-file new)
            (sidebar-update-mode-line))
           ((and (buffer-base-buffer) sidebar-client-buffer)
            ;; this case should have been avoided by the user
            (rename-buffer
             (generate-new-buffer-name
              (buffer-name (buffer-base-buffer)))))))))
    (sidebar-reconfigure))
  ;; revert sidebar
  (cond
   ((and (eq sidebar-mode 'directory)
         (sidebar-same-files (file-name-directory new) default-directory))
    ;; new is in sidebar
    (sidebar-directory-list default-directory (file-name-nondirectory new)))
   ((and (eq sidebar-mode 'directory)
         (sidebar-same-files (file-name-directory old) default-directory))
    ;; new was in sidebar
    (sidebar-directory-list default-directory))
   ((eq sidebar-mode 'bookmark)
    (sidebar-bookmark-list (file-name-nondirectory new)))))

(defun sidebar-move ()
  "Move object on current line of sidebar."
  (interactive)
  (cond
   ((memq (sidebar-line-face)
          '(sidebar-directory-name sidebar-file-name
                                          sidebar-bookmark-directory-name
                                          sidebar-bookmark-file-name))
    (let ((old (sidebar-line-file))
          dir new)
      (when (and (or (file-exists-p old) ; old should exist
                     (sidebar-message "File %s does not exist" old))
                 (not (looking-at sidebar-filter-dot-dirs-regexp))
                 (or (not (file-directory-p old)) (setq dir t))
                 (or dir
                     (not (sidebar-rename-visiting-buffer old))
                     (y-or-n-p
                      (message
                       "Modified or base buffer of indirect buffer visiting `%s'; proceed? "
                       (file-name-nondirectory old))))
                 (or (not dir)
                     (not (sidebar-rename-directory-visiting-buffer old))
                     (y-or-n-p
                      (message
                       "`%s' in path of modified or base buffer of indirect buffer; proceed? "
                       (file-name-nondirectory old))))
                 (setq new
                       (read-file-name (format "Move \"%s\" to: "
                                               (file-name-nondirectory old))
                                       (file-truename (file-name-directory old))
                                       (file-truename old) nil
                                       (file-name-nondirectory old)))
                 (or (not (sidebar-rename-visiting-buffer new))
                     (sidebar-message
                      "Modified or base buffer of indirect buffer visiting `%s'; moving aborted"
                      (file-name-nondirectory new)))
                 (or (not (sidebar-rename-directory-visiting-buffer new))
                     (sidebar-message
                      "`%s' in path of modified or base buffer of indirect buffer; moving aborted"
                      (file-name-nondirectory new)))
                 (or (not (file-exists-p new))
		     ;; *** TODO: Windows specific only ***
		     (string-equal (downcase old) (downcase new))
                     (and (file-directory-p new)
                          (y-or-n-p (format "Move %s to %s ? "
                                            (file-name-nondirectory old) new))
                          (setq new (concat
                                     new "/" (file-name-nondirectory old))))
                     (not dir)
                     ;; avoid that rename overwrites non-directory with directory
                     (y-or-n-p (format "A file %s already exists, overwrite? "
                                       new))))
        ;; rename
        (rename-file old new t)
        ;; handle directories and non-directories separately
        (if dir
            (sidebar-rename-directory old new)
          (sidebar-rename-non-directory old new)))))
   ((and (eq (sidebar-line-face) 'sidebar-bookmark-name)
         ;; rename bookmark iff sidebar-client-buffer displays file whose bookmarks
         ;; are listed in sidebar
         (sidebar-same-files
          sidebar-file (sidebar-bookmark-file-name (sidebar-property))))
    (let* ((bookmark (sidebar-property))
           (old-name (car bookmark))
           (bookmark-current-buffer sidebar-client-buffer)
           (bookmark-yank-point
            (with-current-buffer bookmark-current-buffer (point)))
           new-name)
      (sidebar-bookmark-load)
      (bookmark-maybe-historicize-string old-name)
      (setq new-name (read-from-minibuffer
                      "New name: "
                      nil
                      (let ((now-map (copy-keymap minibuffer-local-map)))
                        (define-key now-map  "\C-w" 'bookmark-yank-word)
                        now-map)
                      nil
                      'bookmark-history))
      (bookmark-set-name old-name new-name)
      (setq bookmark-current-bookmark new-name)
      (setq bookmark-alist-modification-count
            (1+ bookmark-alist-modification-count))
      (when (bookmark-time-to-save-p) (bookmark-save))
      (sidebar-bookmarks)
      (re-search-forward (regexp-quote new-name) (point-max) t)))))

(defun sidebar-copy ()
  "Copy file on current line of sidebar."
  (interactive)
  (when (memq (sidebar-line-face)
              '(sidebar-directory-name sidebar-file-name))
    (let ((old (sidebar-line-file))
          dir new)
      (when (and (or (file-exists-p old) ; old should exist
                     (sidebar-message "File %s does not exist" old))
                 (not (looking-at sidebar-filter-dot-dirs-regexp))
                 (or (not (file-directory-p old)) (setq dir t))
                 (setq new
                       (read-file-name (format "Copy \"%s\" to: "
                                               (file-name-nondirectory old))
                                       (file-truename (file-name-directory old))
                                       (file-truename old) nil
                                       (file-name-nondirectory old)))
                 (or (not (file-directory-p new))
                     ;; user entered new as directory, concat old's file-name
                     (setq new (concat (file-name-as-directory new)
                                       (file-name-nondirectory old))))
;;;                  (y-or-n-p (format "Copy \"%s\" to %s ? "
;;;                                    (file-name-nondirectory old) new))
                 (or (not (file-exists-p new)) (not dir)
                     ;; avoid that rename overwrites non-directory with directory
                     (y-or-n-p (format "A file %s already exists, overwrite? "
                                       new))))
        ;; copy, keep-time
        (copy-file old new 0 t))
      (sidebar-directory-list
       default-directory (file-name-nondirectory old) dir))))

;; _____________________________________________________________________________
;;                                                                              
;;;                               rename at point                               
;; _____________________________________________________________________________
;;                                                                              

;; Invariant
;;   /\ `sidebar-rename-from' <= `sidebar-rename-mask-from'
;;   /\ `sidebar-rename-mask-from' <= `sidebar-rename-mask-to'
;;   /\ `sidebar-rename-mask-from' <= `sidebar-rename-to'
(defun sidebar-rename-reposition ()
  (let ((inhibit-read-only t)
        (mark-active-is mark-active)
        (point-is (point))
        (mark-is (mark)))
    (goto-char (max (point) sidebar-rename-from))
    (goto-char (min (point) sidebar-rename-to))
    (save-excursion
      ;; can't use newlines and form feeds in buffer-names
      (goto-char sidebar-rename-from)
      (while (re-search-forward "[\n\f]" sidebar-rename-to t)
        (replace-match " ")))
    (move-overlay sidebar-line-overlay 1 1)
    (sidebar-overlays)
    (sidebar-adjust-window)
    (cond
     ((< (point) sidebar-rename-mask-from)
      ;; point moved before mask
      (setq sidebar-rename-mask-from (point))
      (setq sidebar-rename-mask-to
            (min (+ sidebar-rename-mask-from sidebar-width)
                 sidebar-rename-to)))
     ((> (point) sidebar-rename-mask-to)
      (setq sidebar-rename-mask-to (point))
      (setq sidebar-rename-mask-from
            (max (- sidebar-rename-mask-to sidebar-width)
                 sidebar-rename-mask-from)))
     (t (setq sidebar-rename-mask-to
              (min (+ sidebar-rename-mask-from sidebar-width)
                   sidebar-rename-to))))
    (move-overlay sidebar-rename-overlay
                  sidebar-rename-from
                  (save-excursion (goto-char sidebar-rename-to) (line-beginning-position 2)))
    (move-overlay sidebar-rename-before-overlay
                  sidebar-rename-from sidebar-rename-mask-from)
    (move-overlay sidebar-rename-after-overlay
                  sidebar-rename-mask-to sidebar-rename-to)
    (put-text-property (point-min) sidebar-rename-from 'rear-nonsticky t)
    (put-text-property
     sidebar-rename-from (1+ sidebar-rename-to) 'local-map sidebar-rename-map)
    (put-text-property (point-min) (point-max) 'read-only nil)
    (put-text-property (point-min) sidebar-rename-from 'read-only t)
    (put-text-property sidebar-rename-to (point-max) 'read-only t)
    (put-text-property (point-min) (point-max) 'rear-non-sticky 'read-only)
    (when (and mark-active-is (= point-is (point)))
      ;; restore mark and mark-active
      (set-mark mark-is)
      (setq mark-active t))
    (if (and mark-active
             (<= sidebar-rename-from (point)) (<= sidebar-rename-from (mark))
             (<= (point) sidebar-rename-to) (<= (mark) sidebar-rename-to))
        (move-overlay sidebar-rename-highlight-overlay
                      (min (point) (mark)) (max (point) (mark)))
      (delete-overlay sidebar-rename-highlight-overlay))))

(defun sidebar-rename ()
  (interactive)
  (let ((inhibit-read-only `(t))
        (old (sidebar-line-file))
        dir)
    (cond
     ((and sidebar-rename-buffer
           (buffer-live-p sidebar-rename-buffer))
      (with-current-buffer sidebar-rename-buffer
        (sidebar-rename-cleanup t)))
     ((or (and (memq (sidebar-line-face)
                     ;; include grepped files ?
                     '(sidebar-directory-name
                       sidebar-file-name
                       sidebar-bookmark-directory-name
                       sidebar-bookmark-file-name))
               (not (looking-at sidebar-filter-dot-dirs-regexp))
               (or (file-exists-p old)  ; old should exist
                   (sidebar-message "File %s does not exist" old))
               (or (not (file-directory-p old)) (setq dir t))
               (setq sidebar-rename-string
                     (buffer-substring (line-beginning-position) (line-end-position)))
               (setq sidebar-rename-string-no-properties
                     (buffer-substring-no-properties (line-beginning-position) (line-end-position)))
               (set-marker sidebar-rename-from (line-beginning-position))
               (setq sidebar-rename-type (if dir 'dir 'file)))
          (and (eq (sidebar-line-face) 'sidebar-bookmark-name)
               ;; rename bookmark iff sidebar-client-buffer displays file whose
               ;; bookmarks are listed in sidebar
               (sidebar-same-files
                sidebar-file
                (sidebar-bookmark-file-name (sidebar-property)))
               (set-marker
                sidebar-rename-from
                (save-excursion
                  ;; looses when someone has changed `sidebar-indent' for this
                  ;; sidebar and not refreshed it at the same time
                  ;; (but do we need leading spaces for bookmarks or
                  ;; could we do skip-syntax here?)
                  (beginning-of-line)
                  (unless sidebar-unfolded
                    ;; unfolded bookmarks must start in first column
                    (forward-char sidebar-indent))
                  (point)))
               (setq sidebar-rename-string
                     (buffer-substring sidebar-rename-from (line-end-position)))
               (setq sidebar-rename-string-no-properties
                     (buffer-substring sidebar-rename-from (line-end-position)))
               (setq sidebar-rename-type 'bookmark-name)))
      (setq buffer-read-only nil)
      (setq sidebar-rename t)
      (setq sidebar-rename-buffer (current-buffer))
      ;; disable transient mark mode, the following comment out on 2011.03.22.
      ;; (set (make-local-variable 'pc-selection-mode) t)
      (set (make-local-variable 'transient-mark-mode) nil)
      (when (and dir sidebar-directory-brackets)
        (dolist (overlay (overlays-at (line-beginning-position)))
          (when (overlay-get overlay 'before-string)
            (delete-overlay overlay))))
      (set-marker sidebar-rename-to (line-end-position))
      (set-marker-insertion-type sidebar-rename-to t) ; advance-marker
      (setq sidebar-rename-before-overlay (make-overlay 1 1))
      (overlay-put sidebar-rename-before-overlay 'invisible 'sidebar-rename)
      (setq sidebar-rename-after-overlay (make-overlay 1 1))
      (overlay-put sidebar-rename-after-overlay 'invisible 'sidebar-rename)
      (sidebar-add-to-invisibility-spec 'sidebar-rename) ; make overlays invisible
      (setq sidebar-rename-overlay
            ;; this should extend to the end of this line
            (make-overlay sidebar-rename-from (line-beginning-position 2) nil nil t))
      (overlay-put sidebar-rename-overlay 'face 'sidebar-rename)
      (setq sidebar-rename-mask-from
            (marker-position sidebar-rename-from))
      (setq sidebar-rename-mask-to
            (min (+ sidebar-rename-mask-from (1- sidebar-width))
                 sidebar-rename-to))
      (setq cursor-type '(bar . 1)) ; make cursor visible
      ;; activate mark
      (set-mark sidebar-rename-from)
      (goto-char sidebar-rename-to)
      (setq mark-active t)
      (setq sidebar-rename-highlight-overlay
            (make-overlay sidebar-rename-from sidebar-rename-to))
      (overlay-put sidebar-rename-highlight-overlay 'face 'region)
      (overlay-put sidebar-rename-highlight-overlay 'priority 1000)))))

(defun sidebar-rename-return ()
  (interactive)
  (let ((dir (eq sidebar-rename-type 'dir))
        (old-string sidebar-rename-string-no-properties)
        (new-string (buffer-substring-no-properties
                     sidebar-rename-from sidebar-rename-to))
        old new)
    (cond
     ((string-equal old-string new-string)
      ;; cleanup if strings are equal
      (sidebar-rename-cleanup t))
     ((and (memq sidebar-rename-type '(dir file))
           (setq old (expand-file-name old-string))
           (setq new (expand-file-name new-string))
           (or (not (file-exists-p new))
	       ;; *** TODO: Windows specific only ***
	       (string-equal (downcase old) (downcase new))
               (and (file-directory-p new)
                    (y-or-n-p (format "Move %s to %s ? "
                                      (file-name-nondirectory old) new))
                    (setq new (concat
                               new "/" (file-name-nondirectory old))))
               (not dir)
               ;; avoid that rename overwrites non-directory with directory
               (y-or-n-p (format "A file %s already exists, overwrite? "
                                 new))))
      (condition-case nil
          (let ((rename-type sidebar-rename-type))
            (rename-file old new 7)
            (sidebar-rename-cleanup)
            ;; handle directories and non-directories separately
            (if (eq rename-type 'dir)
                (sidebar-rename-directory old new)
              (sidebar-rename-non-directory old new)))
        (error (sidebar-rename-cleanup t))))
     ((eq sidebar-rename-type 'bookmark-name)
      (if (and (assoc new-string bookmark-alist)
               (not (y-or-n-p
                     (format
                      "A bookmark with name %s already exists; rename nevertheless?"
                      new-string))))
          (sidebar-rename-cleanup t)
        (bookmark-maybe-historicize-string old-string)
        (bookmark-set-name old-string new-string)
        (setq bookmark-current-bookmark new-string)
        (setq bookmark-alist-modification-count
              (1+ bookmark-alist-modification-count))
        (when (bookmark-time-to-save-p) (bookmark-save))
        (sidebar-rename-cleanup)
        (sidebar-bookmarks)
        (re-search-forward (regexp-quote new-string) (point-max) t))))))

(defun sidebar-rename-restore ()
  (interactive)
  (sidebar-rename-cleanup t))

(defun sidebar-rename-cleanup (&optional restore)
  "Cleanup after rename operation.  With RESTORE non-nil restore original string."
  (let ((inhibit-read-only t))
    (when restore
      (delete-region sidebar-rename-from sidebar-rename-to)
      (insert sidebar-rename-string)
      (when (and (eq sidebar-rename-type 'dir)
                 sidebar-directory-brackets)
        (save-excursion
          (forward-line) (sidebar-put-overlay t))))
    (remove-text-properties
     (point-min) (point-max) '(read-only nil local-map nil))
    (setq cursor-type nil)
    (remove-from-invisibility-spec 'sidebar-rename)
    (delete-overlay sidebar-rename-before-overlay)
    (delete-overlay sidebar-rename-after-overlay)
    (delete-overlay sidebar-rename-overlay)
    (set-marker sidebar-rename-from nil)
    (set-marker sidebar-rename-to nil)
    (setq buffer-read-only t)
    (setq sidebar-rename-buffer nil)
    (setq sidebar-rename nil)))

;; _____________________________________________________________________________
;;                                                                              
;;;                             uncategorized group                             
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-visit-file ()
  "Create new file in current directory."
  (interactive)
  (with-valid-sidebar
   (if (not (eq sidebar-mode 'directory))
       (sidebar-message "Switch to directory first")
     (let ((file (read-file-name
                  (format "Find file: ") default-directory))
           (window (selected-window)))
       (sidebar-select-window sidebar-client-window)
       (find-file file)))))

(defun sidebar-make-directory ()
  "Make new directory in current directory."
  (interactive)
  (with-valid-sidebar
   (if (not (eq sidebar-mode 'directory))
       (sidebar-message "Switch to directory first")
     (let ((directory
            (read-file-name
             (format "New directory: ") default-directory "")))
       (make-directory directory t)
       (message "directory is: %s     dir-part: %s     file-part: %s"
                directory (file-name-directory directory) (file-name-nondirectory directory))
       (sit-for 2)
       (when (file-exists-p directory)
         ;; directory has been made, switch to its directory
         (setq default-directory (file-name-directory directory))
         (sidebar-directory-list
          (file-name-directory directory)
          (file-name-nondirectory directory) t))))))

(defun sidebar-delete ()
  "Remove current entry."
  (interactive)
  (with-valid-sidebar
   (let ((lines (sidebar-line-number)))
     (cond
      ((memq (sidebar-line-face)
             '(sidebar-bookmark-name
               sidebar-bookmark-directory-name
               sidebar-bookmark-file-name))
       (bookmark-maybe-load-default-file)
       (cond
        ((and (eq (sidebar-line-face) 'sidebar-bookmark-file-name)
              ;; don't delete entry that contains bookmarks
              (sidebar-bookmark-file-list (sidebar-property)))
         (sidebar-message "Delete bookmarks first"))
        ((memq (sidebar-line-face)
               '(sidebar-bookmark-directory-name
                 sidebar-bookmark-file-name))
         ;; directory or file
         (let ((file (file-truename (sidebar-property))))
           (catch 'found
             (dolist (bookmark bookmark-alist)
               (when (and (string-equal file (file-truename (cdaadr bookmark)))
                          (sidebar-bookmark-file-p bookmark))
                   (throw 'found
                          (setq bookmark-alist
                                (delq bookmark bookmark-alist)))))))
         (delete-region (line-beginning-position) (line-beginning-position 2)))
        ((eq (sidebar-line-face) 'sidebar-bookmark-name)
         ;; bookmark
         (let ((name (car (sidebar-property)))
               (file (file-truename (cdaadr (sidebar-property)))))
           (catch 'found
             (dolist (bookmark bookmark-alist)
               (when (and (string-equal name (car bookmark))
                          (string-equal file (file-truename (cdaadr bookmark))))
                 (throw 'found
                        (setq bookmark-alist (delq bookmark bookmark-alist)))))))
         (delete-region (line-beginning-position) (line-beginning-position 2))
         (when (or (save-excursion
                     (skip-syntax-forward " ") (eobp))
                   (not (eq (sidebar-line-face) 'sidebar-bookmark-name)))
           ;; deleted last bookmark line for this file
           (forward-line -1)
           (when (eq (sidebar-line-face) 'sidebar-bookmark-file-name)
             ;; deleted last bookmark for this file, search for
             ;; expanded-file-name overlay
             (let ((overlay-list (overlays-at (line-beginning-position))))
               (dolist (overlay overlay-list)
                 (when (eq (overlay-get overlay 'face)
                           'sidebar-bookmark-expanded-file-name)
                   (delete-overlay overlay))))))))
       (setq bookmark-alist-modification-count
             (1+ bookmark-alist-modification-count))
       (when (bookmark-time-to-save-p) (bookmark-save))
       (unless bookmark-alist
         ;; all bookmarks deleted, kill sidebar
         (sidebar-message "No bookmarks")
         (sidebar-delete-window)))
      ((memq (sidebar-line-face)
             '(sidebar-buffer-directory-name sidebar-buffer-name))
       ;; should we allow killing scratch ?
       (kill-buffer (sidebar-property))
       (sidebar-revert)
       (goto-char (point-min))
       (forward-line lines))
      (t
       (sidebar-message "Nothing to delete here"))))))

(defun sidebar-increase-by-one ()
  "Increase match context for match lists and sidebar-width for any other list."
  (interactive)
  (cond
   ((eq sidebar-mode 'match)
    (let ((window-line (sidebar-window-line)))
      (setq sidebar-match-lines-before (1+ sidebar-match-lines-before))
      (setq sidebar-match-lines-after (1+ sidebar-match-lines-after))
      (sidebar-match-refresh)
      (recenter window-line)))
   ((> (- (window-width sidebar-client-window)
          (- sidebar-width sidebar-minimum-width))
       ;; this might be a bit sloppy
       window-min-width)
    (setq sidebar-width (1+ sidebar-width))
    (setq sidebar-maximum-width (1+ sidebar-maximum-width)))))

(defun sidebar-decrease-by-one ()
  "Decrease match context for match lists and sidebar-width for any other list."
  (interactive)
  (cond
   ((eq sidebar-mode 'match)
    (let ((window-line (sidebar-window-line)))
      (unless (zerop sidebar-match-lines-before)
        (setq sidebar-match-lines-before (1- sidebar-match-lines-before))
        (unless (zerop sidebar-match-lines-after)
          (setq sidebar-match-lines-after (1- sidebar-match-lines-after)))
        (sidebar-match-refresh))
      (recenter window-line)))
   ((> sidebar-maximum-width sidebar-minimum-width)
    (setq sidebar-width (1- sidebar-width))
    (setq sidebar-maximum-width (1- sidebar-maximum-width)))))

(defun sidebar-unfolded ()
  (interactive)
  (cond
   ((eq sidebar-mode 'directory) (sidebar-file-list))
   ((eq sidebar-mode 'bookmark) (sidebar-bookmark-unfold))))

(defun sidebar-save (&optional args)
  "Save `sidebar-client-buffer' in its file."
  (interactive "p")
  (if (eq (window-buffer sidebar-client-window) sidebar-client-buffer)
      (with-current-buffer sidebar-client-buffer
        (save-buffer args)
        ;; let standard message show up
        (sit-for 1))
    (sidebar-message
     "%s does not display %s" sidebar-client-window sidebar-client-buffer)))

;; _____________________________________________________________________________
;;                                                                              
;;;                           keymap and syntax-table                           
;; _____________________________________________________________________________
;;                                                                              
(defvar sidebar-mode-map nil
  "Local keymap for sidebar-mode buffers.")

(defvar sidebar-change-drive-map nil)

(defun sidebar-change-drive ()
  (interactive)
  (let ((char last-input-event))
    (if (numberp char)
	(let* ((string (char-to-string char))
	       (drive (concat string ":")))
	  (cond
	   ((not (file-exists-p drive))
	    (sidebar-message "Drive %s does not exist" string))
	   (t
	    (unless (eq sidebar-mode 'directory)
	      (when (sidebar-make)
		(let ((inhibit-read-only t))
		  (sidebar-switch-mode 'directory))))
	    (setq default-directory drive)
	    (sidebar-directory-list drive))))
      (sidebar-message "Invalid command"))))

(defun sidebar-negate-filter-apply ()
  "Toggle filter application."
  (interactive)
  (sidebar-negate-variable 'sidebar-filter-apply))

(unless sidebar-mode-map
  (setq sidebar-mode-map (make-sparse-keymap))
  ;; Change drive map.
  (define-prefix-command 'sidebar-change-drive-map)
  ;; entering an ordinary printing character triggers sidebar-jump
  (let ((index (1+ ?\ )))
    (while (< index 256)
      (define-key sidebar-mode-map (vector index) 'sidebar-jump)
      (define-key sidebar-change-drive-map (vector index) 'sidebar-change-drive)
      (setq index (1+ index))))
  ;; commands
  ;; ..... CURSOR
  (define-key sidebar-mode-map [down] 'sidebar-down)
  (define-key sidebar-mode-map [(control down)] 'sidebar-down-jump)
  (define-key sidebar-mode-map [(meta down)] 'sidebar-down-preserve)
  (define-key sidebar-mode-map [(control meta down)] 'sidebar-down-drag)
  (define-key sidebar-mode-map [(shift control meta down)] 'sidebar-down-view)
  (define-key sidebar-mode-map [up] 'sidebar-up)
  (define-key sidebar-mode-map [(control up)] 'sidebar-up-jump)
  (define-key sidebar-mode-map [(meta up)] 'sidebar-up-preserve)
  (define-key sidebar-mode-map [(control meta up)] 'sidebar-up-drag)
  (define-key sidebar-mode-map [(shift control meta up)] 'sidebar-up-view)
  (define-key sidebar-mode-map [return] 'sidebar-open-kill)
  (define-key sidebar-mode-map [(control return)] 'sidebar-open-kill)
  (define-key sidebar-mode-map [left] 'sidebar-close)
  (define-key sidebar-mode-map [(shift left)] 'sidebar-grep-restore)
  (define-key sidebar-mode-map [right] 'sidebar-open)
  (define-key sidebar-mode-map [(control right)] 'sidebar-open-tag)
  (define-key sidebar-mode-map [(meta right)] 'sidebar-open-bookmark)
  (define-key sidebar-mode-map [(control meta right)] 'sidebar-open-kill)
  (define-key sidebar-mode-map [(shift control meta right)] 'sidebar-open-view)
  ;; ..... HOME END
  (define-key sidebar-mode-map [home] 'sidebar-home)
  (define-key sidebar-mode-map [(control meta home)] 'sidebar-home-top)
  (define-key sidebar-mode-map [(control meta end)] 'sidebar-end-center)
  (define-key sidebar-mode-map [end] 'sidebar-end)
  (define-key sidebar-mode-map [prior] 'sidebar-scroll-down)
  (define-key sidebar-mode-map [next] 'sidebar-scroll-up)
  ;; ..... INSERT DELETE
  (define-key sidebar-mode-map [(meta insert)] 'sidebar-bookmark-add)
  (define-key sidebar-mode-map [delete] 'sidebar-delete)
  ;; ..... MOUSE  (ignore all down events)
  (define-key sidebar-mode-map [down-mouse-1] 'ignore)
  (define-key sidebar-mode-map [down-mouse-2] 'ignore)
  (define-key sidebar-mode-map [down-mouse-3] 'ignore)
  (define-key sidebar-mode-map [mouse-1] 'sidebar-mouse-right)
  (define-key sidebar-mode-map [mouse-2] 'sidebar-mouse-right)
;;   (define-key sidebar-mode-map [mouse-3] 'sidebar-close)
  (define-key sidebar-mode-map [(meta ?\+)] 'sidebar-increase-by-one)
  (define-key sidebar-mode-map [(meta ?\-)] 'sidebar-decrease-by-one)
  (define-key sidebar-mode-map [?\*] 'sidebar-unfolded)
  (define-key sidebar-mode-map [(control ?\*)] 'sidebar-file-list-query)
  (define-key sidebar-mode-map [(meta ?\*)] 'sidebar-file-list-query-non-recursive)
  (define-key sidebar-mode-map [?\?] 'sidebar-grep-query)
  ;; The following might be contrived:
  (define-key sidebar-mode-map [(shift apps)] 'sidebar-grep-query)
  ;; MATCH REPLACE
  ;; *** TODO: We need a replace map here ***
  (define-key sidebar-mode-map [?\ ] 'sidebar-space)
  (define-key sidebar-mode-map [backspace] 'sidebar-backspace)
  (define-key sidebar-mode-map [(control ?\ )] 'sidebar-replace-reset-forward)
  (define-key sidebar-mode-map [(control backspace)] 'sidebar-replace-reset-backward)
  (define-key sidebar-mode-map [(meta ?\ )] 'sidebar-replace-edit-forward)
  (define-key sidebar-mode-map [(meta backspace)] 'sidebar-replace-edit-backward)
  (define-key sidebar-mode-map [?\!] 'sidebar-replace-all)
  (define-key sidebar-mode-map [(control ?\!)] 'sidebar-replace-reset-all)
  (define-key sidebar-mode-map [(meta ?\!)] 'sidebar-replace-edit-all)
  (define-key sidebar-mode-map [?\>] 'sidebar-replace-after)
  (define-key sidebar-mode-map [(control ?\>)] 'sidebar-replace-reset-after)
  (define-key sidebar-mode-map [(meta ?\>)] 'sidebar-replace-edit-after)
  (define-key sidebar-mode-map [?\<] 'sidebar-replace-before)
  (define-key sidebar-mode-map [(control ?\<)] 'sidebar-replace-reset-before)
  (define-key sidebar-mode-map [(meta ?\<)] 'sidebar-replace-edit-before)
  ;; ..... FUNCTION KEYS
  (define-key sidebar-mode-map [f2] 'sidebar-rename)
  (define-key sidebar-mode-map [f4] 'sidebar-visit-file)
  (define-key sidebar-mode-map [f5] 'sidebar-copy)
  (define-key sidebar-mode-map [(meta f5)] 'sidebar-backup-with-date)
  (define-key sidebar-mode-map [(shift meta f5)] 'sidebar-rename-with-date)
  (define-key sidebar-mode-map [f6] 'sidebar-move)
  (define-key sidebar-mode-map [f7] 'sidebar-make-directory)
  (define-key sidebar-mode-map [(shift control f10)] 'sidebar-ediff-base)
  (define-key sidebar-mode-map [(shift meta f10)] 'sidebar-ediff-trunk)
  ;; ..... SPECIAL
  (define-key sidebar-mode-map [(control +)] 'sidebar-mark-all)
  (define-key sidebar-mode-map [(control kp-add)] 'sidebar-mark-all)
  (define-key sidebar-mode-map [(control -)] 'sidebar-unmark-all)
  (define-key sidebar-mode-map [(control kp-subtract)] 'sidebar-unmark-all)
  (define-key sidebar-mode-map [(control kp-multiply)] 'sidebar-mark-list)
  (define-key sidebar-mode-map [(meta s)] 'sidebar-save)
  (define-key sidebar-mode-map [(control d)] 'sidebar-change-drive-map)
  (define-key sidebar-mode-map [(control t)] 'sidebar-negate-filter-apply)
  (define-key sidebar-mode-map [(control n)] 'sidebar-flash-name)
  (define-key sidebar-mode-map [(control p)] 'sidebar-flash-path)
  (define-key sidebar-mode-map [(control P)] 'sidebar-flash-standard-path))

(defvar sidebar-mode-syntax-table nil
  "Syntax table in use in sidebar-mode buffers.")

(unless sidebar-mode-syntax-table
  (setq sidebar-mode-syntax-table (make-syntax-table))
  ;; avoid paren highlighting
  (modify-syntax-entry ?( "w" sidebar-mode-syntax-table)
  (modify-syntax-entry ?) "w" sidebar-mode-syntax-table)
  (modify-syntax-entry ?{ "w" sidebar-mode-syntax-table)
  (modify-syntax-entry ?} "w" sidebar-mode-syntax-table)
  (modify-syntax-entry ?[ "w" sidebar-mode-syntax-table)
  (modify-syntax-entry ?] "w" sidebar-mode-syntax-table))

(easy-menu-define sidebar-menu sidebar-mode-map
  "Sidebar menu."
  (list
   "Sidebar "
   ["Balance Heights"
    (progn (setq sidebar-even-heights (not sidebar-even-heights))
           (sidebar-reconfigure))     ; -post does not work
    :style toggle :selected sidebar-even-heights]
   ["Wrapped Movement"
    (setq sidebar-wrap (not sidebar-wrap))
    :style toggle :selected sidebar-wrap]
   ["Directory Brackets"
    (sidebar-negate-variable 'sidebar-directory-brackets)
    :visible (not (memq sidebar-mode '(match grep)))
    :style toggle :selected sidebar-directory-brackets]
   (list "Minimum Width"
         :visible '(not (memq sidebar-mode '(match grep)))
         ["4" (sidebar-set-variable 'sidebar-minimum-width 4)
          :style radio :selected (= sidebar-minimum-width 4)]
         ["8" (sidebar-set-variable 'sidebar-minimum-width 8)
          :style radio :selected (= sidebar-minimum-width 8)]
         ["12" (sidebar-set-variable 'sidebar-minimum-width 12)
          :style radio :selected (= sidebar-minimum-width 12)]
         ["16" (sidebar-set-variable 'sidebar-minimum-width 16)
          :style radio :selected (= sidebar-minimum-width 16)])
   (list "Maximum Width"
         :visible '(not (memq sidebar-mode '(match grep)))
         ["12" (sidebar-set-variable 'sidebar-maximum-width 12)
          :style radio :selected (= sidebar-maximum-width 12)]
         ["24" (sidebar-set-variable 'sidebar-maximum-width 24)
          :style radio :selected (= sidebar-maximum-width 24)]
         ["36" (sidebar-set-variable 'sidebar-maximum-width 36)
          :style radio :selected (= sidebar-maximum-width 36)]
         ["48" (sidebar-set-variable 'sidebar-maximum-width 48)
          :style radio :selected (= sidebar-maximum-width 48)])
   (list "Width Tolerance"
         :visible '(not (memq sidebar-mode '(match grep)))
         ["2" (setq sidebar-width-tolerance 2)
          :style radio :selected (= sidebar-width-tolerance 2)]
         ["4" (setq sidebar-width-tolerance 4)
          :style radio :selected (= sidebar-width-tolerance 4)]
         ["6" (setq sidebar-width-tolerance 6)
          :style radio :selected (= sidebar-width-tolerance 6)]
         ["8" (setq sidebar-width-tolerance 8)
          :style radio :selected (= sidebar-width-tolerance 8)])
   (list "Indentation"
         :visible '(memq sidebar-mode '(tag bookmark))
         ["0" (sidebar-set-variable 'sidebar-indent 0)
          :style radio :selected (= sidebar-indent 0)]
         ["1" (sidebar-set-variable 'sidebar-indent 1)
          :style radio :selected (= sidebar-indent 1)]
         ["2" (sidebar-set-variable 'sidebar-indent 2)
          :style radio :selected (= sidebar-indent 2)]
         ["4" (sidebar-set-variable 'sidebar-indent 4)
          :style radio :selected (= sidebar-indent 4)])
   ["----" nil :visible (memq sidebar-mode '(directory grep))] 
   ;; filter items
   ["Apply Filters" (sidebar-negate-variable 'sidebar-filter-apply)
    :visible (memq sidebar-mode '(directory grep))
    :style toggle :selected sidebar-filter-apply]
   (list "Filter List"
         :visible '(memq sidebar-mode '(directory match grep))
         :filter 'sidebar-menu-filter)
   ["----" nil :visible (memq sidebar-mode '(directory bookmark))]
   ;; how files are sorted - affects files and bookmarks
   ["Sort By Name" (sidebar-set-variable 'sidebar-files-sort 'name)
    :visible (memq sidebar-mode '(directory bookmark))
    :style toggle :selected (eq sidebar-files-sort 'name)]
   ["Sort By Extension" (sidebar-set-variable 'sidebar-files-sort 'extension)
    :visible (memq sidebar-mode '(directory bookmark))
    :style toggle :selected (eq sidebar-files-sort 'extension)]
   ["Sort By Time" (sidebar-set-variable 'sidebar-files-sort 'time)
    :visible (memq sidebar-mode '(directory bookmark))
    :style toggle :selected (eq sidebar-files-sort 'time)]
   ["Sort By Size" (sidebar-set-variable 'sidebar-files-sort 'size)
    :visible (memq sidebar-mode '(directory bookmark))
    :style toggle :selected (eq sidebar-files-sort 'size)]
   ["Unsorted" (sidebar-set-variable 'sidebar-files-sort 'unsorted)
    :visible (memq sidebar-mode '(directory bookmark))
    :style toggle :selected (eq sidebar-files-sort 'unsorted)]
   ["Reverse Order"
    (sidebar-negate-variable 'sidebar-reverse)
    :visible (memq sidebar-mode '(directory bookmark))
    :style toggle :selected sidebar-reverse]
   ["File List Use \"ls\""
    (setq sidebar-files-use-ls (not sidebar-files-use-ls))
    :visible (memq sidebar-mode '(directory bookmark))
    :style toggle :selected sidebar-files-use-ls]
   ;; tag items (tag lists only)
   ["----" nil :visible (eq sidebar-mode 'tag)]
   (list "Sort Tags"
         :visible '(eq sidebar-mode 'tag)
         ["always"
          (sidebar-set-variable 'sidebar-tags-sort 'always)
          :style radio :selected (eq sidebar-tags-sort 'always)]
         ["unless-nosort"
          (sidebar-set-variable 'sidebar-tags-sort 'unless-nosort)
          :style radio :selected (eq sidebar-tags-sort 'unless-nosort)]
         ["when-sort"
          (sidebar-set-variable 'sidebar-tags-sort 'when-sort)
          :style radio :selected (eq sidebar-tags-sort 'when-sort)]
         ["never"
          (sidebar-set-variable 'sidebar-tags-sort 'never)
          :style radio :selected (eq sidebar-tags-sort 'never)])
   (list "Group Tags"
         :visible '(eq sidebar-mode 'tag)
         ["always"
          (sidebar-set-variable 'sidebar-tag-group 'always)
          :style radio :selected (eq sidebar-tag-group 'always)]
         ["when-sort"
          (sidebar-set-variable 'sidebar-tag-group 'when-sort)
          :style radio :selected (eq sidebar-tag-group 'when-sort)]
         ["never"
          (sidebar-set-variable 'sidebar-tag-group 'never)
          :style radio :selected (eq sidebar-tag-group 'never)])
   ;; bookmark items (bookmark lists only)
   ["----" nil :visible (eq sidebar-mode 'bookmark)]
   ["Bookmarks By Name " (sidebar-set-variable 'sidebar-bookmarks-sort 'name)
    :visible (eq sidebar-mode 'bookmark)
    :style toggle :selected (eq sidebar-bookmarks-sort 'name)]
   ["Bookmarks By Position " (sidebar-set-variable 'sidebar-bookmarks-sort 'position)
    :visible (eq sidebar-mode 'bookmark)
    :style toggle :selected (eq sidebar-bookmarks-sort 'position)]
   ["Bookmarks By Time " (sidebar-set-variable 'sidebar-bookmarks-sort 'time)
    :visible (eq sidebar-mode 'bookmark)
    :style toggle :selected (eq sidebar-bookmarks-sort 'time)]
   ["Reverse Bookmarks"
    (sidebar-negate-variable 'sidebar-bookmarks-reverse)
    :visible (eq sidebar-mode 'bookmark)
    :style toggle :selected sidebar-bookmarks-reverse]
   ;; buffer items (buffer lists only)
   ["----" nil :visible (eq sidebar-mode 'buffer)]
   ["Buffer Sort"
    (sidebar-negate-variable 'sidebar-buffers-sort)
    :visible (eq sidebar-mode 'buffer)
    :style toggle :selected sidebar-buffers-sort]
   ["Buffer Ignore *"
    (sidebar-negate-variable 'sidebar-buffer-ignore-asterisk)
    :visible (eq sidebar-mode 'buffer)
    :style toggle :selected sidebar-buffer-ignore-asterisk]
   ["Buffer Ignore Space"
    (sidebar-negate-variable 'sidebar-buffer-ignore-space)
    :visible (eq sidebar-mode 'buffer)
    :style toggle :selected sidebar-buffer-ignore-space]
   ;; match items (match lists only)
   ["----" nil :visible (eq sidebar-mode 'match)]
   ["Match Context"
    (sidebar-negate-variable 'sidebar-match-context)
    :visible (eq sidebar-mode 'match)
    :style toggle :selected sidebar-match-context]
   (list
    "Match Lines Before And After"
    :visible '(eq sidebar-mode 'match)
    ["0" (sidebar-match-adjust 0 0)
     :style radio :selected (and (= sidebar-match-lines-before 0)
                                 (= sidebar-match-lines-after 0))]
    ["1" (sidebar-match-adjust 1 1)
     :style radio :selected (and (= sidebar-match-lines-before 1)
                                 (= sidebar-match-lines-after 1))]
    ["2" (sidebar-match-adjust 2 2)
     :style radio :selected (and (= sidebar-match-lines-before 2)
                                 (= sidebar-match-lines-after 2))]
    ["3" (sidebar-match-adjust 3 3)
     :style radio :selected (and (= sidebar-match-lines-before 3)
                                 (= sidebar-match-lines-after 3))]
    ["4" (sidebar-match-adjust 4 4)
     :style radio :selected (and (= sidebar-match-lines-before 4)
                                 (= sidebar-match-lines-after 4))]
    ["5" (sidebar-match-adjust 5 5)
     :style radio :selected (and (= sidebar-match-lines-before 5)
                                 (= sidebar-match-lines-after 5))])
   (list
    "Match Lines Before"
    :visible '(eq sidebar-mode 'match)
    ["0" (sidebar-match-adjust 0)
     :style radio :selected (= sidebar-match-lines-before 0)]
    ["1" (sidebar-match-adjust 1)
     :style radio :selected (= sidebar-match-lines-before 1)]
    ["2" (sidebar-match-adjust 2)
     :style radio :selected (= sidebar-match-lines-before 2)]
    ["3" (sidebar-match-adjust 3)
     :style radio :selected (= sidebar-match-lines-before 3)]
    ["4" (sidebar-match-adjust 4)
     :style radio :selected (= sidebar-match-lines-before 4)]
    ["5" (sidebar-match-adjust 5)
     :style radio :selected (= sidebar-match-lines-before 5)])
   (list
    "Match Lines After"
    :visible '(eq sidebar-mode 'match)
    ["0" (sidebar-match-adjust nil 0)
     :style radio :selected (= sidebar-match-lines-after 0)]
    ["1" (sidebar-match-adjust nil 1)
     :style radio :selected (= sidebar-match-lines-after 1)]
    ["2" (sidebar-match-adjust nil 2)
     :style radio :selected (= sidebar-match-lines-after 2)]
    ["3" (sidebar-match-adjust nil 3)
     :style radio :selected (= sidebar-match-lines-after 3)]
    ["4" (sidebar-match-adjust nil 4)
     :style radio :selected (= sidebar-match-lines-after 4)]
    ["5" (sidebar-match-adjust nil 5)
     :style radio :selected (= sidebar-match-lines-after 5)])
   ;; customize items
   "----"
   (list
    "Customize"
    ["Filter List" (customize-group 'sidebar-filter)]
    ["Faces" (customize-group 'sidebar-faces)]
    ["Directory" (customize-group 'sidebar-directory)]
    ["Tag" (customize-group 'sidebar-tags)]
    ["Bookmark" (customize-group 'sidebar-bookmarks)]
    ["Buffer" (customize-group 'sidebar-buffers)]
    ["Match" (customize-group 'sidebar-match)]
    ["All" (customize-group 'sidebar)])))

;; _____________________________________________________________________________
;;                                                                              
;;;                       mode, hooks, timers, and advice                       
;; _____________________________________________________________________________
;;                                                                              
(defun sidebar-delete-sidebar ()
  "Delete window of sidebar or window of sidebar associated with selected window.
Kill buffer of sidebar and any associated buffer in `sidebar-reconfigure'."
  (interactive)
  (if (eq major-mode 'sidebar-mode)
      (sidebar-delete-window (get-buffer-window (current-buffer)))
    (let ((selected-window (selected-window)))
      (dolist (buffer sidebar-list)
        (with-current-buffer buffer
          (when (eq sidebar-client-window selected-window)
            (sidebar-delete-window (get-buffer-window buffer))))))))

(defun sidebar-synchronize (buffer)
  "In tag or bookmark mode synchronize sidebar-client-window unless we had an input
event within sidebar-client-window or \"scroll-other-window\" was part of the last
command."
  (when (timerp sidebar-synchronize-timer)
    (cancel-timer sidebar-synchronize-timer))
  (with-current-buffer buffer
    (when (and (buffer-live-p buffer)
               (eq major-mode 'sidebar-mode)
               (or (eq sidebar-mode 'tag)
                   (eq (sidebar-line-face) 'sidebar-bookmark-name))
               ;; may add scroll-other-window and the like here
               (not (string-match "scroll-other-window"
                                  (symbol-name last-command)))
               (or (not (listp last-command-event))
                   ;; it's not clear whether the following is what we really want
                   (and (not (eq (posn-window (event-start last-command-event))
                                 sidebar-client-window))
                        (not (eq (posn-window (event-end last-command-event))
                                 sidebar-client-window)))))
      (sidebar-view))))

(defun sidebar-prettify (size-string pretty-string)
  "Prettify numbers by inserting dots."
  (if (> (length size-string) 3)
      (sidebar-prettify (substring size-string 0 -3)
                          (concat "." (substring size-string -3) pretty-string))
    (concat size-string pretty-string)))

(defun sidebar-attributes (buffer)
  "Display attributes and other informative messages in echo area."
  (with-current-buffer buffer
    (when (eq major-mode 'sidebar-mode)
      (when (timerp sidebar-attributes-timer)
        (cancel-timer sidebar-attributes-timer))
      (let ((line-face (sidebar-line-face))
            ;; for symbolic links select target
            (line-file (sidebar-line-file))
            message-log-max)
        (when line-file (setq line-file (file-truename line-file)))
        (cond
         ((memq line-face
                '(sidebar-directory-name
                  sidebar-bookmark-directory-name))
          (let ((attributes (file-attributes line-file)))
            (message
             (propertize
              (concat
               (nth 8 attributes) "  "
               (format-time-string "%d.%m.%Y %T" (nth 5 attributes)) "  "
               line-file)
              'face 'sidebar-directory-name))))
         ((memq line-face
                '(sidebar-file-name sidebar-bookmark-file-name
                  sidebar-grep-name))
          (let ((attributes (file-attributes line-file)))
            (message
             (propertize
              (concat
               (nth 8 attributes) "  "
               (sidebar-prettify
                (format "%i" (nth 7 (file-attributes line-file))) "  ")
               (format-time-string "%d.%m.%Y %T" (nth 5 attributes)) "  "
               line-file)
              'face 'sidebar-file-name))))
         ((eq line-face 'sidebar-bookmark-name)
          (let ((bookmark (sidebar-property)))
            (message                    ; no properties here
             (concat
              (car bookmark) "    "
              (sidebar-bookmark-file-name bookmark)))))
         ((eq line-face 'sidebar-buffer-directory-name)
          (let ((name (with-current-buffer (sidebar-property)
                        (directory-file-name default-directory))))
            (message
             (propertize name 'face 'sidebar-directory-name))))
         ((eq line-face 'sidebar-buffer-name)
          (let ((name (sidebar-line-file)))
            (if name
                (message (propertize name 'face 'sidebar-file-name))
              (message (buffer-substring-no-properties
                        (line-beginning-position) (line-end-position)))))))))))

(defun sidebar-mode ()
  "Major mode for traversing directories in separate window."
  (kill-all-local-variables)
  (use-local-map sidebar-mode-map)
  (set-syntax-table sidebar-mode-syntax-table)
  (setq major-mode 'sidebar-mode)
  (setq mode-name "sidebar")
  (put 'sidebar-mode 'mode-class 'special)
  (auto-save-mode 0)
  (buffer-disable-undo)
  (setq buffer-read-only t)
  (setq sidebar-list (sidebar-addq (current-buffer) sidebar-list))
  ;; set some local variables, overriding these might break sidebar
  (set (make-local-variable 'revert-buffer-function) 'sidebar-revert)
  ;; we need some horizontal space for displaying tooltips, likely we have to
  ;; make this customizable.
  (set (make-local-variable 'x-max-tooltip-size) '(100 . 2))
  ;; no buffer boundaries so far
  (set (make-local-variable 'indicate-buffer-boundaries) nil)
  (set (make-local-variable 'indicate-empty-lines) nil) ; don't show empty lines
  (set (make-local-variable 'next-line-add-newlines) nil) ; no newline adding
  (set (make-local-variable 'truncate-lines) t) ; always truncate lines
  (set (make-local-variable 'cursor-type) nil)  ; suppress cursor
  (set (make-local-variable 'goal-column) 0) ; always stay at column zero
  (set (make-local-variable 'redisplay-dont-pause) t) ; redisplay-dont-pause
  (make-local-variable 'frame-title-format)
  (setq sidebar-rename nil)
  (sidebar-filter-list-regexp)        ; update filter list
  (sidebar-shrink-window)
  ;; sidebar overlays (buffer-local)
  (setq sidebar-overlay (make-overlay 1 1))
  (setq sidebar-before-overlay (make-overlay 1 1))
  (overlay-put sidebar-before-overlay 'priority 1000)
  (setq sidebar-after-overlay (make-overlay 1 1))
  (overlay-put sidebar-after-overlay 'priority 1000)
  (setq sidebar-after-string
        (make-string (/ (display-pixel-height) (frame-char-height)) ?\n))
  (setq sidebar-line-overlay (make-overlay 1 1))
  (setq sidebar-context-overlay (make-overlay 1 1))
  (overlay-put sidebar-context-overlay 'face 'sidebar-context)
  (overlay-put sidebar-context-overlay 'priority 500)
  ;; echo overlays (sidebar local)
  (setq sidebar-line-echo-overlay (make-overlay 1 1))
  (setq sidebar-context-echo-overlay (make-overlay 1 1))
  (overlay-put sidebar-context-echo-overlay 'face 'sidebar-context)
  ;; cancel timers
  (when sidebar-attributes-timer (cancel-timer sidebar-attributes-timer))
  (when sidebar-synchronize-timer (cancel-timer sidebar-synchronize-timer))
  ;; add menu
  (easy-menu-add sidebar-menu sidebar-mode-map)
  ;; add and run hooks
  (add-hook 'pre-command-hook 'sidebar-pre-command nil t)
  (add-hook 'post-command-hook 'sidebar-post-command t t)
  (sidebar-add-hooks)
  (run-hooks 'sidebar-mode-hook))

(defun sidebar-reposition ()
  "Assert that point is not within before- or after-overlay."
  (remove-hook 'window-scroll-functions 'sidebar-window-scroll-timer t)
  ;; Handle header-line in associated buffer
  (if (and (buffer-live-p sidebar-client-buffer)
	   (with-current-buffer sidebar-client-buffer
	     header-line-format))
      (setq header-line-format "")
    (setq header-line-format nil))
  (cond
   ((save-excursion (skip-syntax-backward " ") (bobp))
    (skip-syntax-forward " ")
    ;; for sidebars containing whitespace only go to bob, observe we don't
    ;; consider invisible text here
    (when (eobp) (goto-char (point-min)))
    (beginning-of-line))
   ((save-excursion (skip-syntax-forward " ") (eobp))
    (skip-syntax-backward " ")
    (beginning-of-line)
    (when (get-char-property (point) 'invisible)
      (goto-char (or (previous-single-char-property-change (point) 'invisible)
		     (point-min)))
      (forward-line -1)))
   (t (beginning-of-line)))
  (add-hook 'window-scroll-functions 'sidebar-window-scroll-timer t t))

;; Added on 2006-05-05, see below.
(dolist (cmd
 '(forward-char backward-char
   left-char right-char		
   next-line previous-line
   forward-word backward-word
   end-of-line beginning-of-line
   move-end-of-line move-beginning-of-line
   end-of-buffer beginning-of-buffer
   scroll-up scroll-down
   up-list down-list backward-up-list
   end-of-defun beginning-of-defun
   m&d-forward-sexp m&d-backward-sexp
   m&d-begin-of-defun m&d-end-of-defun
   m&d-forward-up m&d-backward-up
   m&d-home m&d-end
   forward-sexp backward-sexp
   forward-list backward-list
   forward-sentence backward-sentence
   forward-paragraph backward-paragraph))
  (put cmd 'sidebar 'move))

(defun sidebar-pre-command ()
  "Remove affects on sidebar and sidebar-client-window.  Cancel synchronization timer
and remove any overlays but window overlay."
  (if sidebar-rename
      (let ((inhibit-read-only t))
        (when (and (eq (overlay-buffer sidebar-rename-highlight-overlay)
                       (current-buffer))
                   (<= sidebar-rename-from (point))
                   (<= sidebar-rename-from (mark))
                   (<= (point) sidebar-rename-to)
                   (<= (mark) sidebar-rename-to))
	  ;; Changed this on 2006-05-05 to deactivate mark on a move command in
	  ;; order to handle m&d, however, the following is not state of the
	  ;; art, see CUA for handling this more elegantly.
	  (cond
	   ((or (memq this-command
		      '(self-insert-command
			self-insert-iso
			yank clipboard-yank
			insert-register insert-parentheses))
		(and (memq this-command
			   '(delete-backward-char backward-delete-char
						  backward-delete-char-untabify delete-char))
		     (setq this-command 'ignore)))
	    (setq mark-active nil)
	    (delete-region (min (point) (mark)) (max (point) (mark)))
	    (delete-overlay sidebar-rename-highlight-overlay))
	   ((eq (get this-command 'sidebar) 'move)
	    (setq mark-active nil))))
	(sidebar-rename-reposition))
    (let ((command-name (symbol-name this-command))
          window-line)
      ;; In match-mode try to preserve the position in `sidebar-client-window' clicked
      ;; with the mouse.  `window-line' saves the window line of `posn-point',
      ;; below we recenter `sidebar-client-window' on that line.
      (when (and (eq sidebar-mode 'match)
                 (listp last-input-event)
		 ;; Avoid errors when switching frames.
		 (not (eq (car last-input-event) 'switch-frame))
                 (not (string-match
                       "scroll-bar-toolkit-scroll\\|vertical-scroll-bar\\|mouse-wheel\\|mwheel" command-name))
                 (eq (posn-window (event-end last-input-event)) sidebar-client-window)
                 (numberp (posn-point (event-end last-input-event))))
        ;; in match lists record current line
        (with-saved-window-and-buffer
	 (sidebar-select-window sidebar-client-window)
	 (goto-char (posn-point (event-end last-input-event)))
	 (setq window-line (sidebar-window-line))))
      ;; cancel timers
      (when sidebar-synchronize-timer (cancel-timer sidebar-synchronize-timer))
      ;; remove line and echo overlays
      (unless (string-match "sidebar-\\(increase\\|decrease\\)" command-name)
        ;; update line overlay
        (when sidebar-line-overlay
          (overlay-put sidebar-line-overlay 'face
                       'sidebar-current-line-non-selected-window))
        ;; remove echo overlay
        (when sidebar-line-echo-overlay
          (delete-overlay sidebar-line-echo-overlay)))
      ;; Handle marks
      (when (and (eq sidebar-mode 'directory) (string-match "sidebar" command-name))
	(cond
	 ((not (eq (get this-command 'sidebar) 'mark))
	  ;; Unmark all.
	  (sidebar-unmark-all))
	 ((memq 'shift (event-modifiers (aref (this-single-command-raw-keys) 0)))
	  ;; Mark line unless not already marked.
	  (sidebar-mark-current))))
      ;; make invisibility overlays visible (this handles also scrolling of sidebar-client-window)
      (when (and (eq sidebar-mode 'match)
                 (buffer-live-p sidebar-client-buffer)
                 (not (string-match
                       ;; match/replace commands handle overlays in match-highlight,
                       "sidebar-\\(down\\|up\\|next\\|prior\\|home\\|end\\|match\\|replace\\|increase\\|decrease\\)"
                       command-name)))
        (when sidebar-context-echo-overlay
          (delete-overlay sidebar-context-echo-overlay))
        (when sidebar-context-overlay
          (delete-overlay sidebar-context-overlay))
        (with-current-buffer sidebar-client-buffer
          (remove-from-invisibility-spec '(sidebar-invisible . t))
	  ;;           (setq mark-active nil) ; deactivate mark here ???
          ;; *** TODO: The following must be done in a different way ***
          (dolist (overlay (overlays-in (point-min) (point-max)))
            (when (overlay-get overlay 'sidebar-match)
              (overlay-put overlay 'face nil)))))
      (unless (or (and (listp last-input-event)
		       (or (eq (car last-input-event) 'switch-frame)
			   (eq (posn-window (event-end last-input-event))
			       (selected-window))))
		  (eq sidebar-mode 'match)
		  (string-match "sidebar\\|customize\\|scroll-bar-toolkit-scroll\\|vertical-scroll-bar\\|mouse-wheel\\|mwheel"
				command-name))
	;; shrink window to handle next-window, previous-window, other-window, ...
	    (sidebar-shrink-window))
      (beginning-of-line)
      ;; Recenter after leaving sidebar with mouse in match-mode.
      (when window-line
        (with-saved-window-and-buffer
	 (sidebar-select-window sidebar-client-window)
	 (recenter window-line))))))

(defun sidebar-window-scroll-timer (window display-start)
  "Hooks into `window-scroll-functions' and should catch _all_ extraneous commands
moving the window-start-position of a sidebar including `scroll-other-window'
and mouse-wheel scrolls."
  (when sidebar-window-scroll-timer (cancel-timer sidebar-window-scroll-timer))
  (when (string-match "scroll-bar-toolkit-scroll\\|vertical-scroll-bar\\|mouse-wheel\\|mwheel"              
                      (symbol-name last-command))
    (setq sidebar-window-scroll-timer
          ;; we can't run `sidebar-window-scroll' immediately here since it
          ;; might not stop during scrolling
          (run-with-idle-timer
           sidebar-scroll-delay nil 'sidebar-window-scroll window))))

(defun sidebar-window-scroll (window)  
  "Reposition sidebar after scrolling.  Invoked by `sidebar-window-scroll-timer'."
  (remove-hook 'window-scroll-functions 'sidebar-window-scroll-timer t)
  (condition-case nil
      (with-saved-window-and-buffer
        (let ((selected-window (selected-window)))
          (sidebar-select-window window)
          ;; window displayed sidebar when scroll-timer was initialized, but
          ;; this might have changed in the meantime
          (when (eq major-mode 'sidebar-mode)
            (if sidebar-rename
                (sidebar-rename-reposition)
              (sidebar-reposition)
              (move-overlay
               sidebar-line-overlay (line-beginning-position) (line-beginning-position 2))
              (overlay-put
               sidebar-line-overlay 'face
               (if (and (eq selected-window window)
                        ;; the following conjunct might become customizable
                        (not (string-match "scroll-bar-toolkit-scroll\\|vertical-scroll-bar\\|mouse-wheel\\|mwheel"
                                           (symbol-name last-command)))
                        (or (not (eq sidebar-mode 'bookmark))
                            (not (eq (sidebar-line-face)
                                     'sidebar-bookmark-name))
                            (not (eq last-command 'sidebar-delete))))
                   'sidebar-current-line-selected-window
                 ;; limbo state
                 'sidebar-current-line-non-selected-window))))))
    (error nil))
  (add-hook 'window-scroll-functions 'sidebar-window-scroll-timer t t))

(defun sidebar-mouse-leave-buffer ()
  "Run `sidebar-pre-command' on all sidebars when switching window with the mouse.
Essential in order to remove echo highlighting."
  (with-saved-window-and-buffer
    (dolist (buffer sidebar-list)
      (when (get-buffer-window buffer)
        ;; till further notice consider sidebars on this frame only
        (sidebar-select-window (get-buffer-window buffer))
	(sidebar-pre-command)))))

(defun sidebar-add-hooks ()
  (add-hook 'window-scroll-functions 'sidebar-window-scroll-timer t t)
  (add-hook 'find-file-hook 'sidebar-bookmark-find-file)
  (add-hook 'write-file-functions 'sidebar-bookmark-write-file)
  (add-hook 'window-configuration-change-hook 'sidebar-reconfigure-post)
  (add-hook 'kill-buffer-hook 'sidebar-reconfigure-post)
  (add-hook 'after-make-frame-functions 'sidebar-reconfigure-post)
  (add-hook 'delete-frame-functions 'sidebar-reconfigure-post)
  (add-hook 'mouse-leave-buffer-hook 'sidebar-mouse-leave-buffer))

(defun sidebar-remove-hooks ()
  (remove-hook 'window-scroll-functions 'sidebar-window-scroll-timer t)
  (remove-hook 'find-file-hook 'sidebar-bookmark-find-file)
  (remove-hook 'write-file-functions 'sidebar-bookmark-write-file)
  (remove-hook 'window-configuration-change-hook 'sidebar-reconfigure-post)
  (remove-hook 'kill-buffer-hook 'sidebar-reconfigure-post)
  (remove-hook 'after-make-frame-functions 'sidebar-reconfigure-post)
  (remove-hook 'delete-frame-functions 'sidebar-reconfigure-post)
  (remove-hook 'mouse-leave-buffer-hook 'sidebar-mouse-leave-buffer))

(defun sidebar-modeline-prefix (symbol default reverse &optional nospace)
  "Propertize first character of SYMBOL unless it equals DEFAULT and append `^'
if REVERSE is non-nil."
  (concat
   (if (eq symbol default)
       (when reverse (propertize "^" 'face 'sidebar-modeline-prefix))
     (propertize
      (concat
       (upcase (substring (symbol-name symbol) 0 1))
       (cond
        (reverse "^")
        (nospace "")
        (t " ")))
      'face 'sidebar-modeline-prefix))))

(defun sidebar-update-mode-line ()
  "Set sidebar-title and update-mode-line."
  (let ((title-prefix
         (cond
          ((eq sidebar-mode 'directory)
           (concat
            " "
            (sidebar-modeline-prefix
             sidebar-files-sort 'name sidebar-reverse)))
          ((eq sidebar-mode 'bookmark)
           (let ((pre-slash
                  (sidebar-modeline-prefix
                   sidebar-files-sort 'name
                   sidebar-reverse 'nospace))
                 (post-slash
                  (sidebar-modeline-prefix
                   sidebar-bookmarks-sort
                   'name sidebar-bookmarks-reverse)))
             (if (string-equal pre-slash "")
                 (if (string-equal post-slash "")
                     " "
                   (concat (propertize " -/" 'face 'sidebar-modeline-prefix)
                           post-slash))
               (if (string-equal post-slash "")
                   (concat  " " pre-slash
                           (propertize "/- " 'face 'sidebar-modeline-prefix))
                 (concat " " pre-slash
                         (propertize "/" 'face 'sidebar-modeline-prefix)
                         post-slash))))))))
    ;; set sidebar-title
    (setq sidebar-title
          (cond
           ((eq sidebar-mode 'directory)
            (concat
             (when sidebar-directory-brackets "[")
             (let ((name
                    ;;; (string-make-unibyte
                     (file-name-nondirectory
                      (directory-file-name
                       (file-truename default-directory)))));)
               (if (string-equal name "")
                   ;; handle drive names on W32
                   (if (memq system-type '(ms-dos windows-nt))
                       (substring
                        (directory-file-name default-directory) 0
                        (string-match
                         "\\\\\\|/" (directory-file-name default-directory)))
                     (directory-file-name (file-truename default-directory)))
                 name))
             (when sidebar-directory-brackets "]")))
           ((eq sidebar-mode 'match)
            ;; the following must become customizable <---------
            (concat
             (format
              (if (= sidebar-match-count 1)
                  "%s    %s match for %s`%s'"
                "%s    %s matches for %s`%s'")
              ;; the following might be irritating
              ;;                    (buffer-name (buffer-base-buffer sidebar-client-buffer))
              (buffer-name sidebar-client-buffer)
              sidebar-match-count
              (if sidebar-match-type
                  (concat (symbol-name sidebar-match-type) " ")
                "")
              sidebar-match-string)))
           ;; do something for grep here
           (t (buffer-name sidebar-client-buffer))))
    ;; update modeline
    (setq mode-line-format
          (list
           (cond
            ((eq sidebar-mode 'tag)
             (concat " " (file-name-nondirectory sidebar-file)))
            ((eq sidebar-mode 'bookmark)
             (if sidebar-file
                 (concat title-prefix (file-name-nondirectory sidebar-file))
               " *BOOKMARKS*"))
            ((eq sidebar-mode 'meta) " *META*")
            ((eq sidebar-mode 'buffer)
	     (if (or (fboundp 'window-group) (fboundp 'group-root-window))
		 (concat ; WINDOW
		  (replace-regexp-in-string
		   " on.+" ""
		   (replace-regexp-in-string
		    "#<window " "<#" (format "   %s" (selected-window))))
		  (cond
		   ((and (fboundp 'group-root-window)
			 (group-root-window (selected-window)))
		    (format "/%s>" (group-root-window (selected-window))))
		   ((and (fboundp 'window-group)
			 (window-group (selected-window)))
		    (format "/%s>" (window-group (selected-window))))
		   (t ">")))
	       " *BUFFERS*"))
            ((eq sidebar-mode 'match)
             ;; display match count, other things won't fit anyway
             (format (format "%%%dd" (1+ sidebar-width)) ; 1+ is left fringe
                     sidebar-match-count))
            ((eq sidebar-mode 'grep) " *GREP*")
            (t (concat title-prefix sidebar-title)))))
    (force-mode-line-update)))

(defvar sidebar-mark-list nil)
(make-variable-buffer-local 'sidebar-mark-list)

;; *** TODO: Add something here ... In particular we must not destroy
;; sidebar-mark-list if the command needs it ***
(dolist (cmd '(sidebar-down sidebar-up sidebar-space sidebar-backspace sidebar-mark-list))
  (put cmd 'sidebar 'mark))

;; *** TODO: A complete mess the following ***
(defun sidebar-post-command ()
  "Higlighting and related stuff."
  (with-valid-sidebar
   (let ((command-name (symbol-name this-command)))
     ;; *** TODO: Remove the following, it's for testing only ***
;;;      (when (not (eq (get-buffer-window (current-buffer)) (selected-window)))
;;;        (message ".... selected is %s but sidebar's is %s"
;;; 		(selected-window)
;;; 		(get-buffer-window (current-buffer)))
;;;        (ding) (sit-for 2))
     (cond
      (sidebar-rename
       (sidebar-rename-reposition))
      ((string-match "scroll-bar-toolkit-scroll\\|vertical-scroll-bar\\|mouse-wheel\\|mwheel" command-name)
       ;; this clause might get customizably removed
       (sidebar-reposition))
      (t
       (cond
	((and (eq sidebar-mode 'bookmark)
              (not (string-match "sidebar" command-name)))
         ;; maybe refresh bookmarks
         (sidebar-bookmark-load)
         (when (or (/= sidebar-bookmark-alist-modification-count
                       bookmark-alist-modification-count)
                   (and (file-exists-p bookmark-default-file)
                        (or (/= (car sidebar-bookmark-default-file-modtime)
                                (car (nth 5 (file-attributes
                                             bookmark-default-file))))
                            (/= (cadr sidebar-bookmark-default-file-modtime)
                                (cadr (nth 5 (file-attributes
                                              bookmark-default-file)))))))
           ;; the following closes all opened bookmarks but those pertaining
           ;; to sidebar-file, we do _not_ change sidebar-file here 
           (sidebar-bookmarks)))
        ((and (eq sidebar-mode 'buffer)
              (not (string-match "sidebar" command-name)))
         ;; revert sidebar, may change sidebar-client-buffer and sidebar-file,
         ;; somehow this is not consistent with our philosophy of retaining
         ;; the original sidebar-file/buffer
         (setq sidebar-client-buffer (window-buffer sidebar-client-window))
         (setq sidebar-file (buffer-file-name sidebar-client-buffer))
         (sidebar-buffers (buffer-list)))
	((and (eq sidebar-mode 'match)
	      (not (eq (window-buffer sidebar-client-window) sidebar-client-buffer)))
	 ;; Assert that `sidebar-client-window' displays `sidebar-client-buffer'.
	 (sidebar-reconfigure)))
       ;; cancel timers
       (when sidebar-attributes-timer
         (cancel-timer sidebar-attributes-timer))
       (when sidebar-synchronize-timer
         (cancel-timer sidebar-synchronize-timer))
       (cond
        ((and (memq (sidebar-line-face)
                    '(sidebar-tag-name sidebar-bookmark-name))
              (not (string-match
                    "sidebar-\\(up\\|down\\)-drag" command-name)))
         (let ((window-line (sidebar-window-line))
               (window-height-min (sidebar-window-height-min))
               (view (or (not (memq (sidebar-line-face)
                                    '(sidebar-tag-name
                                      sidebar-bookmark-name)))
                         (and (string-match "sidebar" command-name)
                              (not (string-match
                                    "sidebar-\\(tag\\|bookmark\\)"
                                    command-name))))))
           (unless (string-match "sidebar-\\(down\\|up\\)" command-name)
             ;; readjust window of sidebar
             (sidebar-overlays)
             (sidebar-reposition)
             (sidebar-adjust-window))
           ;; put sidebar line overlay
           (overlay-put
            sidebar-line-overlay 'face
	    'sidebar-current-line-selected-window)
	    ;; The following was merely disconcerting.
;;;             (if view
;;;                 'sidebar-current-line-selected-window
;;;               'sidebar-current-line-non-selected-window))
           (move-overlay sidebar-line-overlay (line-beginning-position) (line-beginning-position 2))
	   ;; recenter here to avoid flashes in `sidebar-synchronize'
           (cond
            ((and view (> window-line (- window-height-min 7)))
             (recenter (max 0 (- window-height-min 7))))
            (t (recenter window-line)))
           (cond
            ((eq sidebar-mode 'bookmark)
             (sidebar-attributes (current-buffer)))
            ((eq (sidebar-line-face) 'sidebar-tag-name)
             ;; for tags write full name 
             (let (message-log-max)
               (message "%s" (get-text-property (point) 'help-echo)))))
           (when view
	     ;; For some strange reason the following was never assigned (i.e.,
	     ;; `sidebar-synchronize-timer' was always nil).
             (setq sidebar-synchronize-timer
		   (run-with-idle-timer
		    (if (string-match
			 ;; delay holds only for sidebar-up(-jump) and
			 ;; sidebar-down(-jump) commands (-preserve is handled
			 ;; by sidebar-view already) !
			 "sidebar-\\(up\\|down\\)\\($\\|-preserve\\|-jump\\)"
			 command-name)
			sidebar-synchronize-delay 0)
		    nil 'sidebar-synchronize (current-buffer))))))
        ((eq sidebar-mode 'match)
         (unless (string-match "sidebar-\\(down\\|up\\)" command-name)
           ;; readjust window of sidebar
           (sidebar-overlays)
           (sidebar-reposition))
         ;; put sidebar line overlay
         (overlay-put
          sidebar-line-overlay 'face
          'sidebar-current-line-selected-window)
         (move-overlay sidebar-line-overlay (line-beginning-position) (line-beginning-position 2))
         (sidebar-match-highlight))
        (t
         (unless (string-match "sidebar-\\(down\\|up\\)" command-name)
           ;; readjust window of sidebar
           (sidebar-overlays)
           (sidebar-reposition)
           (sidebar-adjust-window))
         ;; put sidebar line overlay
         (overlay-put
          sidebar-line-overlay 'face
          'sidebar-current-line-selected-window)
         (move-overlay sidebar-line-overlay (line-beginning-position) (line-beginning-position 2))
         (setq sidebar-attributes-timer
               (run-with-idle-timer
                sidebar-attributes-delay nil
                'sidebar-attributes (current-buffer)))))
       ;; for some reason this seems necessary
       (setq mark-active nil)
       ;; update modeline
       (sidebar-update-mode-line)
       ;; frame title, customize this eventually <---------------
       (setq frame-title-format sidebar-title)
       (when sidebar-redisplay-discard-input (discard-input)))))))

;; Note: When the window-configuration changes we don't run `sidebar-reconfigure'
;; immediately.  We rather add `sidebar-reconfigure' to `post-command-hook'.  As a
;; consequence, `post-command-hook' may run both `sidebar-post-command' and
;; `sidebar-reconfigure' (the latter hopefully first).  We cannot rely on
;; `sidebar-post-command' to run `sidebar-reconfigure' since the former is a local
;; hook only and would be executed iff a sidebar were the current buffer.  This
;; explains the weird use of `sidebar-reconfigure-post' below.
(defun sidebar-reconfigure-post (&rest args)
  "Add `sidebar-reconfigure' to `post-command-hook'.
Ignore ARGS \(likely some frame\).  Adding is suppressed if
`sidebar-inhibit-reconfigure' is non-nil."
  (unless sidebar-inhibit-reconfigure
    (add-hook 'post-command-hook 'sidebar-reconfigure)))

;; _____________________________________________________________________________
;;                                                                              
;;;                                 provide us                                  
;; _____________________________________________________________________________
;;                                                                              
(provide 'sidebar)

(defun sidebar-next-visible-line (&optional arg)
  (let ((count (or arg 1)))
    (while (and (not (zerop count)) (not (eobp)))
      (forward-line)
      (when (get-char-property (point) 'invisible)
	(goto-char (or (next-single-char-property-change (point) 'invisible)
		       (point-max))))
      (setq count (1- count)))))

(defun sidebar-prev-visible-line (&optional arg)
  (let ((count (or arg 1)))
    (while (and (not (zerop count)) (not (bobp)))
      (forward-line -1)
      (when (get-char-property (point) 'invisible)
	(goto-char (or (previous-single-char-property-change (point) 'invisible)
		       (point-min)))
	(forward-line -1))
      (setq count (1- count)))))

;; 2. Copy relative file name to kill-ring.
;; 3. Copy absolute file name to kill-ring.

;; backup files only.

(defun sidebar-rename-with-date ()
  (interactive)
  (sidebar-backup-with-date t))

(defun sidebar-backup-with-date (&optional rename)
  (interactive)
  (when (eq (sidebar-line-face) 'sidebar-file-name)
    (let* ((old-name (sidebar-line-file))
	   (old-name-sans (file-name-sans-extension old-name))
	   (old-name-extension (file-name-extension old-name t))
	   (directory default-directory)
	   (current-time (current-time))
	   (date (format-time-string "(%Y-%0m-%0d-" current-time))
	   (number 0)
	   new-name)
      (when (and (or (file-exists-p old-name)
		     (sidebar-message "File %s does not exist" old-name))
		 (or (not (file-directory-p old-name))
		     ;; We could backup directories but ...
		     (sidebar-message "File %s is a directory" old-name)))
	(while (file-exists-p
		(concat old-name-sans date (number-to-string (1+ number)) ")" old-name-extension))
	  (setq number (1+ number)))
	(if (and rename (not (zerop number)))
	    (progn
	      (setq new-name
		    (concat old-name-sans date (number-to-string number) ")" old-name-extension))
	      (when (y-or-n-p (format "OVERWRITE %s with copy of %s " new-name old-name))
		(rename-file old-name new-name t)))
	  (setq new-name
		(concat old-name-sans date (number-to-string (1+ number)) ")" old-name-extension))
	  (when (y-or-n-p (format "Copy %s to %s " old-name new-name))
	    (copy-file old-name new-name 0 t)))
	(sidebar-directory-list
	 default-directory (file-name-nondirectory old-name))))))

;; 1. Copy relative file name to kill-ring.
;; 2. Copy absolute file name to kill-ring.

(defun sidebar-mark-current ()
  "Assert that current line has a sidebar mark overlay."
  (unless (or (looking-at "[ \t]*\n") (eobp)
	      (catch 'marked
		(dolist (overlay (overlays-at (line-beginning-position)))
		  (when (eq (overlay-get overlay 'face) 'sidebar-mark)
		    (throw 'marked t)))))
    (let ((overlay (make-overlay
		    (line-beginning-position) (line-beginning-position 2))))
      (setq sidebar-mark-list (cons overlay sidebar-mark-list))
      (overlay-put overlay 'face 'sidebar-mark))))

(defun sidebar-unmark-current ()
  "Remove any sidebar mark overlay from current line."
  (dolist (overlay (overlays-at (line-beginning-position)))
    (when (eq (overlay-get overlay 'face) 'sidebar-mark)
      (delete-overlay overlay)
      (setq sidebar-mark-list (delq overlay sidebar-mark-list  )))))

(defun sidebar-mark-all (&optional from)
  "Assert that every line has a sidebar mark overlay.
With optional argument FROM a negative number mark all lines before
current line.  With optional argument FROM a positive number mark all
lines after and including current line."
  (interactive)
  (save-excursion
    (save-restriction
      (cond
       ((or (not (numberp from)) (zerop from)))
       ((< from 0)
	(narrow-to-region (point-min) (line-beginning-position)))
       ((> from 0)
	(narrow-to-region (line-beginning-position) (point-max))))
      (goto-char (point-min))
      (skip-syntax-forward " ")
      (while (not (looking-at "[ \t]*\\(?:\n\\|\\'\\)"))
	;; This loop stops at an empty line or at eob.
	(sidebar-mark-current)
	(forward-line)))))

(defun sidebar-unmark-all (&optional from)
  "Remove all sidebar mark overlays.
With optional argument FROM a negative number unmark all lines before
current line.  With optional argument FROM a positive number unmark all
lines after and including current line."
  (interactive)
  (if (or (not (numberp from)) (zerop from))
      (progn
	(dolist (overlay sidebar-mark-list)
	  (delete-overlay overlay))
	(setq sidebar-mark-list nil))
    (save-excursion
      (save-restriction
	(cond
	 ((< from 0)
	  (narrow-to-region (point-min) (line-beginning-position)))
	 ((> from 0)
	  (narrow-to-region (line-beginning-position) (point-max))))
	(goto-char (point-min))
	(skip-syntax-forward " ")
	(while (not (looking-at "[ \t]*\\(?:\n\\|\\'\\)"))
	  ;; This loop stops at an empty line or at eob.
	  (sidebar-unmark-current)
	  (forward-line))))))

(defun sidebar-mark-toggle ()
  "Toggle sidebar mark overlay for current line."
  (interactive)
  (unless (catch 'marked
	    (dolist (overlay (overlays-at (line-beginning-position)))
	      (when (eq (overlay-get overlay 'face) 'sidebar-mark)
		;; Delete overlay by side-effect.
		(delete-overlay overlay)
		(setq sidebar-mark-list (delq overlay sidebar-mark-list))
		(throw 'marked t))))
    ;; Mark.
    (sidebar-mark-current)))

(defun sidebar-mark-list (&optional relative print)
  ;; Sorts sidebar-mark-list as a side-effect
  (setq sidebar-mark-list
	(sort sidebar-mark-list
	      (lambda (o1 o2)
		(> (overlay-start o1) (overlay-start o2)))))
  (let (files)
    (save-excursion
      (dolist (overlay sidebar-mark-list)
	(goto-char (overlay-start overlay))
	(setq files (cons
		     (if relative
			 (buffer-substring-no-properties
			  (line-beginning-position) (line-end-position))
		       (expand-file-name
			(buffer-substring-no-properties
			 (line-beginning-position) (line-end-position))))
		     files))))
    (when print (message "........ %s" files) (sit-for 2))
    files))

(defun sidebar-mark-list-absolute ()
  (interactive)
  (sidebar-mark-list nil t))

(defun sidebar-mark-list-relative ()
  (interactive)
  (sidebar-mark-list t t))

(defun sidebar-ediff-base ()
  (interactive)
  (let ((files (sidebar-mark-list t))
	(directory
	 (file-name-nondirectory 
	  (directory-file-name default-directory))))
    (ediff-directories
     (if (string-equal directory "src")
	 "c:/emacs-base/src/"
       "c:/emacs-base/lisp/")
     default-directory (concat "^" (regexp-opt files 'words)))))

(defun sidebar-ediff-trunk ()
  (interactive)
  (let ((files (sidebar-mark-list t))
	(directory
	 (file-name-nondirectory 
	  (directory-file-name default-directory))))
    (ediff-directories
     (if (string-equal directory "src")
	 "c:/emacs-trunk/src/"
       "c:/emacs-trunk/lisp/")
     default-directory (concat "^" (regexp-opt files 'words)))))

;;; (defun sidebar-ediff-revision ()
;;;   "Run Ediff on a pair of buffers, BUFFER-A and BUFFER-B."
;;;   (interactive)
;;;   (let* ((backup-file
;;; 	  (vc-version-backup-file-name (buffer-file-name sidebar-client-buffer))))
;;;     (when backup-file
;;;       (ediff

;;; BUG: the second time I call the local tags, not all variables get
;;; listed any more ???????????
